// -*- C++ -*-

// DataPipe.h
//
// Classes for sources and sinks of data
// - schedulable using Scheduler
// - transferring data through CircularBuffers
// for Audio Pattern Recognition Language (aprl)
// dpwe@media.mit.edu 1995feb02
//

#include <stdio.h>
#include <assert.h>

#include "DataPipe.h"

// --------------------------- DataSinkList ----------------------------

DataSinkList:: ~DataSinkList(void)
{   /* destroy all the nodes in the list */
    DataSinkNode *n, *dpn = head;

    while(dpn) {
	n = dpn->next;
	delete dpn;
	dpn = n;
    }
}

void DataSinkList::add(DataSink *proc)
{   /* Add new DataSink* at top of list with a new node. */
    DataSinkNode *n;

    n = new DataSinkNode(proc, head);
    head = n;
}

int DataSinkList::remove(DataSink *proc)
{   /* Search down the list, find and delete the node whose contents 
       is <proc>, close up the list.  Report error to stderr and return 0
       if node not found, else return 1. */
    DataSinkNode *n, **pdpn = &head;
    int done = 0;

    while(!done && *pdpn)  { 
	if( proc == ((*pdpn)->data) )  {
	    n = *pdpn;
	    *pdpn = (*pdpn)->next;
	    n->next = 0;
	    delete n;
	    done  = 1;
	} else {
	    pdpn = &((*pdpn)->next);
	}
    }
    if(!done)  {
	fprintf(stderr, "DataSinkList::remove: not found\n");
	return 0;
    }
    return 1;
}

void DataSinkList::apply(DataSinkFn *dpf, void *dpfdata)
{   /* Map the function to the contents of every node in the list.
       <dpfdata> is a private pointer passed to the client function. */
    DataSinkNode *dpn = head;

    while(dpn)  {
	(*dpf)(dpn->data, dpfdata);
	dpn = dpn->next;
    }
}

DataSinkList *DataSinkList::Duplicate(void)
{   /* Make a copy of the top-level elements.  Needed for ~DataSource */
    /* At the moment, reverses the list, which is OK. */
    DataSinkList *dsl = new DataSinkList();
    DataSinkNode *dpn = head;

    while(dpn)  {
	dsl->add(dpn->data);
	dpn = dpn->next;
    }
    return dsl;
}
    

static void SmallestSeq(DataSink *d, void *dpfdata)
{   /* MAPpable fn finds min in list */
    double *mnsq = (double *)dpfdata;
    double sq = d->Sequence();
    if(sq < *mnsq)	*mnsq = sq;
}

double DataSinkList::earliestSequence(void)
{   /* find the smallest sequence val in list (source must be smaller) */
    double rslt = 999999.0;	// assumed bigger than any
    apply(SmallestSeq, (void *)&rslt);
    return rslt;
}

// ------------------------- DataSource -------------------------------
static void ClearSinkInput(DataSink *sink, void *)
{   /* lambda-like fn to call on DataSource->outputs in ~DataSource */
    sink->SetInput(0);
}

DataSource::~DataSource(void)
{   /* make sure all Sinks know we're deserting them */
    /* have to wokr with a *copy* of outputs, since calling SetInput(0) 
       on outputs has the additional effect of deleting that element 
       from the 'outputs' list of its input i.e. this list! */
    DataSinkList *t = outputs->Duplicate();
    t->apply(ClearSinkInput, 0);
    /* outputs will have had all its elements deleted by this */
    delete outputs; 
    /* get rid of our copy too */
    delete t;
}

#ifdef MAIN	// ------------------------- TEST CODE ----------------------

#include <math.h>

class RandSrc : public DataSource {
private:
    int blksize;
public:
    RandSrc(double srate=0, int bufsz = 0) : DataSource(srate, bufsz)
        { blksize = bufsz-1; };
    ~RandSrc(void)
        { };
    int Update(double /* time */)	// generate some random samples
        {  int i, n, todo = blksize, totalDone = 0;
	   DATA *d;

	   fprintf(stderr, "RandSrc(%d):", todo);
           PushHead();
	   while(todo) {
	       assert(n = GetHeadPtr(&d));
	       n = min(n, todo);
	       for(i = 0; i<n; ++i)  {
		   d[i] = drand48();
		   fprintf(stderr, "%.3f ", d[i]);
	       }
	       UpdateHead(n);
	       todo -= n;
	       totalDone += n;
	   }
	   fprintf(stderr, "\n");
	   CircularBuffer::print("RSrc");
	   return totalDone;
       };
};

class SquareProc : public DATAprocess {
public:
    SquareProc(void)	{ };
    ~SquareProc(void)	{ };
    int process(DATA *input, int istride, int niItems, 
		DATA *output, int ostride, int noItems, 
		int *poutdone)
	{  int i,todo = min(niItems, noItems);
	   DATA d;
	   for(i=0; i<todo; ++i) 
	   	{ d = *input; *output = d*d;
		  input += istride; output += ostride; }
	   *poutdone = todo; return todo; }
};

class SquarePipe : public DataPipeProcess {	// simple pipe copies i to o
private:
public:
    SquarePipe(DataSource *in = 0) : DataPipeProcess(in)
        { procFn = new SquareProc(); }
    ~SquarePipe(void)
        { delete procFn; }
    int Update(double t)
    { DataPipeProcess::Update(t); CircularBuffer::print("SqPipe"); }
};

class StraightPipe : public DataPipeProcess {	// simple pipe copies i to o
private:
public:
    StraightPipe(DataSource *in = 0) : DataPipeProcess(in)
        { procFn = new DATAcopyClass(); }
    ~StraightPipe(void)
        { delete procFn; }
    int Update(double t)
    { DataPipeProcess::Update(t); CircularBuffer::print("StPipe"); }
};

class DummySink : public DataSink {	// dummy sink prints data
public:
    DummySink(DataSource *in=0) : DataSink(in)	{ };
    ~DummySink(void)				{ };
    int Update(double /* time */)
        { int totalDone = 0;
	  if(Input())  {
	    if(Input()->PopTail()<0)  {
		fprintf(stderr, "DummySink: error popping tail\n");
		return 0; 
	    }
	    int avai = Input()->Avai();
	    DATA *d;
	    int i, n;
	    fprintf(stderr, "DummySink(%d):", avai);
	    while(totalDone < avai)  {
		assert(n = Input()->GetTailPtr(&d));
		n = min(n, avai-totalDone);
		for(i = 0; i<n; ++i)  {
		    fprintf(stderr, "%.3f ", d[i]);
		}
		Input()->UpdateTail(n);
		// inputCount += n;
		totalDone += n;
	    }
	    fprintf(stderr, "\n");
	}
	return totalDone;
      };
};

void testDPipe(void)
{
    RandSrc      rs(10000, 10);
    DummySink    ds;
    SquarePipe sq;
    DummySink    ds2;
    StraightPipe sp(&rs);

    ds.SetInput(&sq);
    ds2.SetInput(&sp);

    sq.SetInput(&sp);

    Scheduler sc;

    sc.Enschedule(&rs);
    sc.Enschedule(&sp);
    sc.Enschedule(&ds);
    sc.Enschedule(&ds2);

    fprintf(stderr, "Update 0\n");
    sc.Update(0);
    fprintf(stderr, "Update 1\n");
    sc.Update(1);
    fprintf(stderr, "Update 2\n");
    sc.Update(2);
}

void main(/* int argc, char **argv */)
{
    testDPipe();
}

#endif /* MAIN */
