/*******************************************************************************

	chart.C:
		
		Contains class:	 Arc, ArcCollection, ArcIterator, Chart

		Author:		Dan Jurafsky
		Date:		September, 1992
		History:	DSJ	7-23-93	Added header

******************************************************************************/

#include "grammar.h"
#include "parser.h"
#include "semant.h"
#include <assert.h>
#ifdef __GNUC__
#include <pfstream.h>
#else /* !__GNUC__ */
#include <iostream.h>
#endif /* __GNUC__ */
extern int ArcNews;
extern int ArcDeletes;
int Unify;
int onlineUnify;
int do_cover_sem;

/******************************************************************************

	class ArcCollection 

		Data:
			int	iArcs;
			Arc	*Arcs[MAXARCS];

		Constructor:
			ArcCollection();

		Methods:
			void	AddArc(Arc*);
			void	Reset();
			void 	print();

******************************************************************************/

ArcCollection::ArcCollection()
{
	iArcs = 0;
}

void ArcCollection::print()
{
	int i;
	for (i=0;i< iArcs; i++) {
		Arcs[i]->print();
	}
}

void ArcCollection::AddArc(Arc *a)
{
	int i;
	if (iArcs == MAXARCS) {
		for (i=0;i< iArcs; i++) 
			Arcs[i]->print();
		error("Too many arcs");
	}
	Arcs[iArcs++] = a;
}

void ArcCollection::Reset()
{
	int i;
	for (i = 0; i < iArcs; i++ ) {
		//Arcs[i]->Reset();
		delete Arcs[i];
		Arcs[i] = NULL;
	}
	iArcs = 0;
}


/******************************************************************************

	NOTE: really oughta get rid of this confusing operator-overloading

	class ArcIterator 

	Data:
		int	current;
		ArcCollection *c;

	Constructor:
			 ArcIterator();
			 ArcIterator(ArcCollection *);

	Methods:
		Arc *	 operator()(Symbol*,int);
		Arc *	 operator()(Symbol*);
		Arc *	 operator()();
		Arc *	 operator()(Symbol*, int, int);
		Arc * 	NextStart(Symbol* Cat, int Start);
		void	 Reset();
};
******************************************************************************/
ArcIterator::ArcIterator()
{
}	

ArcIterator::ArcIterator(ArcCollection * a)
{
	c = a;
	current = -1;
}	

void ArcIterator::Reset()
{
	current = -1;
}

Arc * ArcIterator::operator()()
{	
	if (current == -1) 
		current = 0;
	if (current < c->iArcs)
		return c->Arcs[current++];
	else
		return 0;
}
	
Arc * ArcIterator::operator()(Symbol *Cat)
{
	uint found = FALSE;

	/* This is only somewhat sneaky.  If the previous guy
		was current, we start this search from the one
		past current.  The sneaky part is that the init
		state sets current to -1, and -1 + 1 = 0, how convenient.*/
			
	current++;

	for ( ; current < c->iArcs; current++) {
		if ( (c->Arcs[current]->IsComplete == 0) && 
		    (c->Arcs[current]->NextSym == Cat)) {
			found = TRUE;
			break;
		}
	}
	if (found)
		return c->Arcs[current];
	else
		return 0;
}
Arc * ArcIterator::operator()(Symbol *Cat, int Stop)
{
	uint found = FALSE;

	/* This is only somewhat sneaky.  If the previous guy
		was current, we start this search from the one
		past current.  The sneaky part is that the init
		state sets current to -1, and -1 + 1 = 0, how convenient.*/
			
	current++;

	for ( ; current < c->iArcs; current++) {
		if ((c->Arcs[current]->To == Stop) && 
// 		    (c->Arcs[current]->From == 0) && 
 		    (c->Arcs[current]->IsComplete == 0) && 
		    (c->Arcs[current]->NextSym == Cat)) {
			found = TRUE;
			break;
		}
	}
	if (found)
		return c->Arcs[current];
	else
		return 0;
}

Arc * ArcIterator::NextStart(Symbol *Cat, int Start)
{
	uint found = FALSE;

	/* This is only somewhat sneaky.  If the previous guy
		was current, we start this search from the one
		past current.  The sneaky part is that the init
		state sets current to -1, and -1 + 1 = 0, how convenient.*/
			
	current++;

	for ( ; current < c->iArcs; current++) {
		if ((c->Arcs[current]->To == Start) && 
		    (c->Arcs[current]->lhs == Cat)) {
			found = TRUE;
			break;
		}
	}
	if (found)
		return c->Arcs[current];
	else
		return 0;
}

Arc * ArcIterator::operator()(Symbol *Cat, int Start, int Stop)
{
	
	uint found = FALSE;

	/* This is only somewhat sneaky.  If the previous guy
		was current, we start this search from the one
		past current.  The sneaky part is that the init
		state sets current to -1, and -1 + 1 = 0, how convenient.*/
			
	current++;

	for ( ; current < c->iArcs; current++) {
	    //c->Arcs[current]->print();
		if ((c->Arcs[current]->From == Start) && 
 		    (c->Arcs[current]->To == Stop) && 
		    (c->Arcs[current]->lhs == Cat)) {
			found = TRUE;
			break;
		}
	}
	if (found)
		return c->Arcs[current];
	else
		return 0;
}


/******************************************************************************

	Note:  Arc really should NOT be a subclass of Rule, each Arc should
			just have a pointer to a Rule, would be much more
			efficient.


	class Arc 

	Data:
		int	From;
		int	To;
		int	IsComplete;
		Symbol	*NextSym;
		int	dot;
		Tree *	Tree;

	Constructors:
	        Arc();
	        Arc(Rule*);
	        ~Arc();
	        Arc(Symbol*,int,int);

	Methods:

		Arc &	operator=(Arc&);
		void	AdvanceDot();
		void	print();
		void    print(FILE *);
		void	Reset();
		void	BuildSem(int, char*);
		int	Disjoint(int *);

******************************************************************************/

Arc::~Arc()
{
	ArcDeletes++;
	if (ArcTree)
		delete ArcTree;
}

void Arc::Reset()
{
  int i;

	if (ArcTree)
	  delete ArcTree;
	ArcTree = NULL;
	dot = 0;
	NextSym = NULL;
	IsComplete = 0;
	RuleCleanup();
	prefixprob=0;
	insideprob=0;
}

Arc::Arc()
{
	ArcNews++;
	dot = 0;
	IsComplete = 0;
	ArcTree = NULL;
	NextSym = NULL;
	insideprob = prefixprob = 0;
}

Arc::Arc(Symbol *s, int FromPoint, int ToPoint)
{
	ArcNews++;
        lhs = s;
	From = FromPoint;
	To = ToPoint;
	dot = 0;
	IsComplete = 1;
	ArcTree = NULL;

	NextSym = NULL;

	iNumChildren = 0;
	prob = 0;
	insideprob = prefixprob = 0;
	rhs = NULL;
}

Arc::Arc(Rule* r)
{
	uint i;

	ArcNews++;
	lhs = r->lhs;
	prob = r->prob;
	RuleNum = r->RuleNum;
	prefixprob=0;
	insideprob=0;

	dot = 0;
	From = 0;
	To = 0;
	IsComplete = 0;
	ArcTree = NULL;
	NextSym = NULL;

#ifndef NOPARSE
	if (r->Sem) {
		Sem = new char[strlen(r->Sem)+1];
		strcpy(Sem,r->Sem);
	} else
	  Sem = NULL;
#endif

	iNumChildren = r->iNumChildren;

	assert(iNumChildren < 25);
	rhs = new Symbol*[iNumChildren];
	for ( i = 0; i < iNumChildren; i++)
		rhs[i] = r->rhs[i];

	NextSym = rhs[0];

}

Arc & Arc::operator=(Arc &r)
{
	uint i;

	lhs = r.lhs;
	prob = r.prob;
	dot = r.dot;
	From = r.From;
	To = r.To;
	NextSym = r.NextSym;
	IsComplete = r.IsComplete;
	prefixprob=r.prefixprob;
	insideprob=r.insideprob;

	if (ArcTree) {
		delete ArcTree;
		ArcTree = NULL;
	}
	if (r.ArcTree) 
	  ArcTree = r.ArcTree->Copy();

#ifndef NOPARSE
	if (Sem)
	  delete [] Sem;
	Sem = NULL;
	if (r.Sem) {
		Sem = new char[strlen(r.Sem)+1];
		strcpy(Sem,r.Sem);
	}
#endif


	iNumChildren = r.iNumChildren;
	if (rhs)
		delete [] rhs;

	assert(iNumChildren < 25);
	rhs = new Symbol*[iNumChildren];
	for ( i = 0; i < iNumChildren; i++)
		rhs[i] = r.rhs[i];
	
	return *this;
}

char * varstr(int i)
{
	char *tmp;

	tmp = new char[3];

	tmp[0] = '$';
	tmp[1] = (char)(i + '0');
	tmp[2] = 0;
	return tmp;
}

char *dstrstr(char * text, char * pat)
{
	char *p = text;
	int x = strlen(pat);

	if (strlen(text) < x)
	  return NULL;
	
	while (*p) {
		if (!strncmp(pat,p,x))
		  return p;
		p++;
	}
	return NULL;
}

void Arc::BuildSem(int dot, char * s) 
{
	if (Unify && do_cover_sem) {
	  if (IsComplete) {   /* do only on completed arcs */
	    if (Sem)                       /*  (this routine called before  */
					   /*  AdvanceDot)  */
	      delete Sem;
	    Sem = GetSemFromDag();
	  }
	} else {
		char *tmp;
		char * pos;
		char *x = varstr(dot);
		int slen = (s)?strlen(s):0;	

		// if incorporating null semantics, just return
		if ((!s) || (s[0] == '\0')) {
			  delete [] x;
			  return;
		}

		if (!Sem || (Sem[0] == '\0')) {
			// if sem of parent is null, just copy in sem of child
			if (Sem)
				delete [] Sem;
			Sem = new char[slen+1];
			strcpy(Sem,s);
	
		} else {
			/* search for variable in parent semantics indicating
			  child position, and copy it in */
			if ((pos = dstrstr(Sem,x)) == NULL) {
			    delete [] x;
			    return;
			}
			tmp = new char[strlen(Sem) - 2 + slen + 1];
			strncpy(tmp,Sem,pos-Sem);
			strncpy(tmp+(int)(pos-Sem),s+1,slen-2);	// strip brackets
			strcpy(tmp+(int)(pos-Sem)+slen-2,pos+2);
			delete [] Sem;
			Sem = tmp;
		}
		delete [] x;
	}
}


void Arc::print(ostream &os)
{
	int i;
	os << "(" << From << " " << To << ")";
	if (IsComplete) {
		os << "# ";
	}
	os << "{" << ArcTree->GetProb() << "} ";
	lhs->print(os);
	os << " -> ";
	for ( i = 0; i < iNumChildren; i++) {
		if ((dot == i) && !IsComplete)
		  os << " . ";
		rhs[i]->print(os);
		os << " ";
	}
	//os << Sem;
	os << "[" << prob << "]";
	os << "<I:" << insideprob << ">";
	os << "<P:" << prefixprob << ">";
	if (state == STATE_SCAN)
		os << "(SCAN)";
	if (state == STATE_COMPLETION)
		os << "(COMPLETION)";
	if (state == STATE_PREDICTION)
		os << "(PREDICTION)";
	//if (ArcTree)
		//os << "<T:" << ArcTree->GetProb() << ">";
	os << "\n" << flush;
	if (IsComplete && ArcTree && ArcTree->ParentDag && os==cout)
	  ArcTree->ParentDag->write(os,1);
}


void Arc::AdvanceDot()
{
	if (++dot < iNumChildren)
	        NextSym = rhs[dot];
	else
	        IsComplete = 1;
}


int Arc::Disjoint(int * intervallist)
{
	int i;
#ifdef NOTDEF
	printf("Checking arc from %d to %d\n",From,To);
	printf("Intervallist is:\n");
	for (i=0;intervallist[i] != -1;i++)
		printf("%d ", intervallist[i]);
	printf("\n");
	fflush(stdout);
#endif
	for (i = 0; intervallist[i] != -1; i+=2) {
		if ((From == intervallist[i]) 
			|| (To == intervallist[i+1])
			|| ((From > intervallist[i]) && (From < intervallist[i+1]))
			|| ((To > intervallist[i]) && (To < intervallist[i+1]))
			|| ((To < intervallist[i]) && (From > intervallist[i]))
			|| ((To < intervallist[i+1]) && (From > intervallist[i+1])))
			return FALSE;
	}
	return TRUE;
}

// return a string built up from the arcs DAG.  to be called
//  on the arc spanning the entire sentence
char *Arc::GetSemFromDag() {
  char *retval;

  const int BufSiz = 1024;
  char *ret = new char[BufSiz];
  *ret = '\0';
  retval=formulate_query(ArcTree->ParentDag, ret, BufSiz);
  if(retval) {
    return retval;
  } else {
    delete[] ret;
    return NULL;
  }
}
    



void Arc::print(FILE *fp)
{
  int i;
  char *x=lhs->Stringify();

  fprintf(fp,"(%d %d){%e} %s ->",From,To,ArcTree->GetProb(),x);
  free(x);
  for ( i = 0; i < iNumChildren; i++) {
    if ((dot == i) && !IsComplete)
      fprintf(fp," .");
    x=rhs[i]->Stringify();
    fprintf(fp," %s",x);
    free(x);
  }
  fprintf(fp," [%f]<I:%f><P:%f>\n",prob,insideprob,prefixprob);
  cout.sync_with_stdio();
  fflush(fp);
  cout.sync_with_stdio();
}


AgendaList::AgendaList()
{
	head = tail = 0;
}
void AgendaList::Add(Arc * ap)
{
	//cout << "Adding arc to agenda:\n";
	//ap->print();
	//cout << flush;
	// increment tail
	int oldtail = tail;
	//if (tail == 299) 
	if (tail == MAXAGENDA)
		tail = 0;
	 else
		tail++;
	if (tail == head)
		error("Overflow in Agenda\n");
	q[oldtail] = ap;
}

Arc *AgendaList::Get()
{
	int oldhead = head;
	if (head == tail)
		return NULL;
	//if ((head == (tail - 1)) || ((head == 300) && (tail == 0))) 
	//if (head == 299) 
	if (head == MAXAGENDA) 
		head = 0;
	 else
		head++;
	
	return q[oldhead];
}

KeyStack::KeyStack()
{
	index = 0;
	TheStack[index] = NULL;
}

void KeyStack::Push(Arc * el)
{
	if (index == MAXSTACK - 1)
		error("stack overflow");
	TheStack[++index] = el;
}

Arc *KeyStack::Pop()
{
	if (index == 0)
		return 0;
	else
		return TheStack[index--];
}

Arc *KeyStack::Peek()
{
	return TheStack[index];
}

void Chart::Reset()
{
	Arcs.Reset();
	Constits.Reset();
	arccount = 0;
	Tcl_DeleteHashTable(&HashTable);
	Tcl_InitHashTable(&HashTable,TCL_STRING_KEYS);
}

void Chart::ExtendArc(Arc* a, int To, Tree * t, char *s)
{
	Arc * ap = new Arc();
	ap->state=STATE_NONE;

	// make a copy of the previous arc
	*ap = *a;

#ifndef NOTREE
	// build a new parse tree to embed the lower one
	ap->ArcTree->ConjoinChild(ap->dot,t);
	ap->ArcTree->TreeProb = ap->ArcTree->TreeProb * t->TreeProb;

#endif
	
	// Advance dot
	ap->AdvanceDot();
	ap->To = To;

	if (onlineUnify && !ap->ArcTree->checkConstraints(ap->dot,ap->RuleNum)) {
	  delete ap;
	  return;
	}

	// If arc is complete, but doesn't unify, delete it
	if (ap->IsComplete && Unify && !(ap->ArcTree->unifyChildren(ap->RuleNum))) {
	  delete ap;
	  return;
	}
#ifndef NOPARSE
	ap->BuildSem(ap->dot,s);
	//printf("multiplying %g by %g\n",ap->ArcTree->TreeProb , t->TreeProb);
#endif

	// Insert the new arc
	Arcs.AddArc(ap);
	
	// If arc is complete, put onto key list
	if (ap->IsComplete) {
	  Keys.Push(ap);
	}
//	 else
//	        delete ap;
}

void Chart::NewArc(Rule *r, int From, int To, Tree * t, char *s)
{
	Arc * ap = new Arc(r);
	ap->state = STATE_NONE;
	ap->From = From;
#ifndef NOTREE
	ap->ArcTree = new Tree(ap->lhs,t,ap->iNumChildren,r->RuleNum,r->Constraints);
	if (t)
		ap->ArcTree->TreeProb = r->prob * t->TreeProb;
	else
		ap->ArcTree->TreeProb = r->prob;

#endif
	
	// Advance dot
	ap->AdvanceDot();
	ap->To = To;

	if (onlineUnify && !ap->ArcTree->checkConstraints(ap->dot,ap->RuleNum)) {
	  delete ap;
	  return;
	}
	
	// If arc is complete, but doesn't unify, delete it
	if (ap->IsComplete && Unify && !(ap->ArcTree->unifyChildren(ap->RuleNum))) {
	  delete ap;
	  return;
	}
#ifndef NOPARSE
	ap->BuildSem(ap->dot,s);
	//printf("setting prob to %g * %g = %g\n",r->prob,t->TreeProb,ap->ArcTree->TreeProb );
#endif

	// Insert the new arc
	Arcs.AddArc(ap);
	
	// If arc is complete, put onto key list
	if (ap->IsComplete) {
	  Keys.Push(ap);
	}
	//else
         //       delete ap;
}
	 
Arc * Chart::PopKey()
{
	return Keys.Pop();
}

#ifdef NOTDEF
Arc * Chart::GetEdge(Symbol *s, int from)
{
	ArcIterator 	it = &Arcs;
	Arc * ap;

	if (ap = it.NextStart(s,from))
		return ap;
	return 0;
}

Arc * Chart::GetEdge(Symbol *s, int from)
{
	char tmp[MAXSYMLEN+5];
	char tmp2[5];
	int newptr = 0;
	Tcl_HashEntry * ptr;
	char *x;

	x = s->Stringify();
	strcpy(tmp,x);
	free(x);
	sprintf(tmp2,"%d",from);
	strcat(tmp,tmp2);

	ptr = Tcl_FindHashEntry(&HashTable, tmp);
	
	if (ptr == NULL) {
		return NULL;
	} else
		return (Arc *)Tcl_GetHashValue(ptr);
}
#endif
void Chart::AddConstit(Symbol *s, int From, int To)  
{
        Arc * p = new Arc(s,From,To);
	p->IsComplete = 1;
        Constits.AddArc(p);
//	delete p;
}

void Chart::AddConstit(Arc * ap)
{
	Arc * p = new Arc();

	// make a copy of the previous arc
	*p = *ap;

	p->IsComplete = 1;
	Constits.AddArc(p);
}


Tree *Chart::GetParseTree(Symbol *start, int Len)
{
	static ArcIterator 	next = &Constits;
	Arc 		*ap;

	ap = next(start,0,Len);

	if (ap == NULL) {
//		cout << "No parses for this sentence" << flush;
	        next = &Constits;
		return NULL;
	}

	return ap->ArcTree;
}

char *Chart::GetSem(Symbol *start, int Len, double *f)
{
	ArcIterator 	next = &Constits;
	Arc 		*ap;

	ap = next(start,0,Len);

	if (ap == NULL) {
//		cout << "No parses for this sentence" << flush;
		return NULL;
	}

	if (Parsetype==TOPDOWN)
	  *f = GetSentenceProb();
	else
	  *f = GetBottomUpSentenceProb();

	if (Unify)
	  return ap->GetSemFromDag();
	else
	  return ap->Sem;

}

int NotBlank(char * in)
{
	int i;
	for (i=0;i<strlen(in);i++) {
		if (in[i] != ' ')
			return TRUE;
	}
	return FALSE;
}


char *Chart::GetCoverSem(int *guesslist)
{
	Arc 		*ap;
	int 		i = 0;
	int		j = 0;
	char		*ret;
	Arc		*ArcArray[MAXGUESS];
	int		size = 0;
	int		intervallist[MAXGUESS];


	if (!do_cover_sem)
	  return NULL;

	ArcArray[0] = NULL;
	intervallist[0] = -1;
	while ((ap = GetBestSem(intervallist)) != NULL)
		ArcArray[i++] = ap;
	
	if (ArcArray[0] == NULL) {
		return NULL;
	}

       if (i >= MAXGUESS)
               error("no more room in guesslist");
 

	/* Construct the return semantics by glueing together
		the semantics of all the pieces with "and"s */

	for (j=0;j<i;j++)  {
		strchr(ArcArray[j]->Sem,'{')[0] = ' ';
		strchr(ArcArray[j]->Sem,'}')[0] = ' ';
		size += strlen(ArcArray[j]->Sem);
	}
	ret = (char *)malloc(size + (i * strlen(" and "))+1);
	strcpy(ret,"{");
	strcat(ret,ArcArray[0]->Sem);
	for (j=1;j<i;j++)  {
		strcat(ret," and ");
		strcat(ret,ArcArray[j]->Sem);
	}
	strcat(ret,"}");

	/* Flatten the collection of intervals into a binary list */

	for (i = 0; intervallist[i] != -1; i+=2) {
		for (j=intervallist[i];j<intervallist[i+1];j++)
			guesslist[j] = 1;
	}
	return ret;
}
Arc *Chart::GetBestSem(int *intervallist)
{
	ArcIterator 	next = &Constits;
	Arc 		*ap;
	Arc		*BestArc = NULL;
	Arc		*NextBestArc = NULL;
	int 		i,j = 0;
//	char		*ret;

	while (ap = next()) {
		if (!BestArc) {
			if (ap->Sem && NotBlank(ap->Sem) && ap->Disjoint(intervallist))
				BestArc = ap;
		} else if (ap->Sem && NotBlank(ap->Sem) && ap->Disjoint(intervallist)) {
				if ((ap->To - ap->From) > (BestArc->To - BestArc->From))  
					BestArc = ap;
				else if (((ap->To - ap->From) == (BestArc->To - BestArc->From)) &&
					 (strlen(ap->Sem) > strlen(BestArc->Sem)))
					BestArc = ap;
		}
	}

	if (BestArc == NULL) {
		return NULL;
	}

	
	// find end of list
	for (i=0; intervallist[i] != -1; i++);

       if (i >= MAXGUESS)
               error ("Overflow in guess list");
 

	// add new interval to list
	intervallist[i] = BestArc->From;
	intervallist[i+1] = BestArc->To;
	intervallist[i+2] = -1;
	return BestArc;
}

double Chart::GetBottomUpSentenceProb()
{
  Tree *tp;
  double prob=0.0;

  tp = GetParseTree(Start,Len);
  while(tp) {
    prob+=tp->TreeProb;
    tp = GetParseTree(Start,Len);
  }
  return prob;
}
  
double Chart::GetSentenceProb()
{
	ArcIterator next = &Arcs;
	Arc *ap;
	double psum = 0.0;

	//printf("Len is %d\n",Len);
	while (ap = next()) {
		// For each arc in chart which is still active at current position
	
		//ap->print();

		if ((ap->To == Len) && (ap->state == STATE_SCAN))  {
		//if ((ap->To == Len) && (!ap->IsComplete) && (ap->state == STATE_SCAN))  {
			/* let p be the probability of the prefix parse tree
				associated with this arc */
				
			psum += ap->prefixprob;

			//cout << "Computing Sent Prob with arc: ";
			//ap->print();
			//cout << flush;
		}
	}
	return psum;
}

#ifdef NOTDEF
double Chart::GetSentenceProb()
{
	//ArcIterator 	next = &Constits;
	ArcIterator 	next = &Arcs;
	Arc 		*ap;
	double		prob = 0;

	while (TRUE) {
		ap = next(Start,0,Len);

		if (ap == NULL) {
			//printf("got no arcs from next, start %d, len is %d\n",Start,Len);
			return prob;
		}

		prob += ap->prefixprob;
		//printf("GetProb: Adding prob &g for arc: ",prob);
		//ap->print();
		if (ap->prefixprob != ap->insideprob) {
			cout << "wrong prob\n" << flush;
		}
		//prob += ap->ArcTree->TreeProb;
	}
}
#endif


int Chart::GetNumberOfParses(Symbol *start, int Len)
{
	ArcIterator 	next = &Constits;
	Arc 		*ap;
	int		num = 0;

	while (TRUE) {
		ap = next(start,0,Len);

		if (ap == NULL)
			return num;
		num++;
	}
}

#ifdef TATEYAMA
void Chart::sendAllParses2ptw(char *instr)
{
  ArcIterator 	next = &Constits;
  Arc 		*ap;
  int		num = 0;
  extern FILE *s4ptw;
  
  fprintf(s4ptw, "S:IS %s\n", instr);
  while (TRUE) {
    ap = next(Start,0,Len);
    
    if (ap == NULL) {
      // number of parse tree is ...
      fprintf(s4ptw, "7:NT %d\n", num);
      // end parses
      fprintf(s4ptw, "9:EP ==============================\n");
      return;
    }
    num++;
    ap->ArcTree->output2ptw(s4ptw);
  }
}
#endif

void Chart::print()
{
	uint i;
	cout << "printing the chart:\n";
	for (i=0;i< Arcs.iArcs; i++) {
		Arcs.Arcs[i]->print();
	}
//	cout << "printing the constits:\n";
//	for (i=0;i< Constits.iArcs; i++) {
//		Constits.Arcs[i]->print();
//	}
}

Chart::Chart()
{
	Tcl_InitHashTable(&HashTable,TCL_STRING_KEYS);
	arccount = 0;
}

ParseTreeIterator::ParseTreeIterator(Chart *c)
{
	start = c->Start;
	next = &c->Constits;
	len = c->Len;
}

void ParseTreeIterator::Reset()
{
	next.Reset();
}

Tree *ParseTreeIterator::Next()
{
	Arc *ap;

	ap = next(start,0,len);

	if (ap == NULL) 
		return NULL;

	return ap->ArcTree;
}

void Chart::TopDownExtendArcUp(Arc* a, int To, Tree * t, char *s, double xinsideprob,int isterminal)
{
	Arc * ap = new Arc();

	// make a copy of the previous arc
	*ap = *a;
#ifdef DEBUG
	printf("UpExtending arc:\n");
	ap->print();
	printf("which has tree:\n");
	ap->ArcTree->print();
	printf("with tree:\n");
	if (t)
			t->print();
	fflush(stdout);
#endif

#ifndef NOTREE
	// build a new parse tree to embed the lower one
	ap->ArcTree->ConjoinChild(ap->dot,t);
	ap->ArcTree->TreeProb = ap->ArcTree->TreeProb * t->TreeProb;
#endif
	ap->insideprob = ap->insideprob * xinsideprob;
	ap->prefixprob = ap->prefixprob * xinsideprob;
#ifdef DEBUG
	printf("Built new tree:\n");
	ap->ArcTree->print();
	fflush(stdout);
#endif
	//printf("multiplying %g by %g\n",ap->ArcTree->TreeProb , t->TreeProb);

	if (isterminal)
		ap->state = STATE_SCAN;
	else
		ap->state = STATE_COMPLETION;
	
	// Advance dot
	ap->AdvanceDot();
	ap->To = To;
	if (onlineUnify && !ap->ArcTree->checkConstraints(ap->dot,ap->RuleNum)) {
	  delete ap;
	  return;
	}
	// If arc is complete, but doesn't unify, delete it
	if (ap->IsComplete && Unify && !(ap->ArcTree->unifyChildren(ap->RuleNum))) {
	  delete ap;
	  return;
	}
#ifndef NOPARSE
	ap->BuildSem(ap->dot,s);
#endif

	Register(ap);


	// Insert the new arc
	// don't forget to delete it when we pop it.
	Agenda.Add(ap);
	//printf("TopDownExtendArcUp extending arc to\n");
	//ap->print();
	//printf("with passed prob %f, new inside prob %f, new prefixprob = %f\n",xinsideprob,ap->insideprob,ap->prefixprob);
}
		
#ifdef NOTDEF


void Chart::TopDownExtendArcDown(Arc* Top, int To, Tree * t, char *s)
{
	Arc * ap = new Arc();

	// make a copy of the previous arc
	*ap = *Top;
#ifdef DEBUG
	printf("DownExtending arc:\n");
	ap->print();
	printf("with tree:\n");
	t->print();
#endif
	
#ifndef NOTREE
	// build a new parse tree to embed the lower one
	ap->ArcTree->ConjoinChild(ap->dot,t);
	ap->ArcTree->TreeProb = ap->ArcTree->TreeProb * t->TreeProb;
#endif
	//printf("multiplying %g by %g\n",ap->ArcTree->TreeProb , t->TreeProb);
#ifdef DEBUG
	printf("Built new tree:\n");
	ap->ArcTree->print();
	fflush(stdout);
#endif
	
	// Advance dot
	ap->AdvanceDot();
	ap->To = To;
#ifndef NOPARSE
	ap->BuildSem(ap->dot,s);
#endif

	Register(ap);

	// Insert the new arc
	Keys.Push(ap);
}
#endif
void Chart::TopDownNewArc(Rule *r, int From, int To, Tree * t, char *s,int iscomplete,double iprefixprob,int isscanning)
{
	Arc * ap = new Arc(r);
	ap->From = From;
	ap->prefixprob = iprefixprob * r->prob;
	ap->insideprob = r->prob;
	if (isscanning)
			ap->state = STATE_SCAN;
	else
			ap->state = STATE_PREDICTION;
#ifndef NOTREE
	ap->ArcTree = new Tree(ap->lhs,t,ap->iNumChildren,r->RuleNum,r->Constraints);
	if (t)
		ap->ArcTree->TreeProb = r->prob * t->TreeProb;
	else
		ap->ArcTree->TreeProb = r->prob;
#endif
#ifdef DEBUG2
	printf("TopDownNewArc adding arc based on rule\n");
	r->print();
	printf("with passed prob %f, rule prob %f, prefixprob = %f\n",iprefixprob,r->prob,ap->prefixprob);
	fflush(stdout);
	cout << flush;
	//printf("Built new tree:\n");
	//ap->ArcTree->print();
	//fflush(stdout);
#endif

	//printf("setting prob to %g * %g = %g\n",r->prob,t->TreeProb,ap->ArcTree->TreeProb );
	ap->IsComplete = 0;

	
	// Advance dot
	if (iscomplete)
		ap->AdvanceDot();
	ap->To = To;
	if (onlineUnify && !ap->ArcTree->checkConstraints(ap->dot,ap->RuleNum)) {
	  delete ap;
	  return;
	}

	// If arc is complete, but doesn't unify, delete it
	if (ap->IsComplete && Unify && !(ap->ArcTree->unifyChildren(ap->RuleNum))) {
	  delete ap;
	  return;
	}
#ifndef NOPARSE
	ap->BuildSem(ap->dot,s);
	//printf("Built new sem\n");
	//fflush(stdout);
#endif
	Register(ap);

	// Insert the new arc
	Agenda.Add(ap);
}
void Chart::UpdateArcProb(Rule *r, int From, int To, double prob)
{
	Arc * ap;
	if ((ap = GetArc(r->RuleID(),From,To)) == NULL) {
		printf("Missing rule: %d %d %d\n",r->RuleID(),From,To);
		r->print();
		exit(1);
	} else {
		ap->prefixprob += prob  * r->prob;
	}
}

Arc * Chart::GetArc(int rulenum,int from, int to)
{
	int newptr = 0;
	Tcl_HashEntry * ptr;
	char tmp[50];

	sprintf(tmp,"%d %d %d",rulenum,from,to);
	ptr = Tcl_FindHashEntry(&HashTable, tmp);
	if (ptr == NULL) {
		return NULL;
	} else
		return (Arc *)Tcl_GetHashValue(ptr);
}
void Chart::Register(Arc *ap)
{
	int newptr = 0;
	Tcl_HashEntry * ptr;
	char tmp[50];

	sprintf(tmp,"%d %d %d",ap->RuleID(),ap->From,ap->To);
	ptr = Tcl_CreateHashEntry(&HashTable, tmp, &newptr);
	Tcl_SetHashValue(ptr,ap);
}


#ifdef NOTDEF
void Chart::Register(Arc *ap)
{
	char * tmp;
	int newptr = 0;
	Tcl_HashEntry * ptr;
	char tmp2[5];
	char *x;

	x = ap->lhs->Stringify();
	strcpy(ArcIndex[arccount],x);
	free(x);
	sprintf(tmp2,"%d",ap->To);
	strcat(ArcIndex[arccount],tmp2);
	ptr = Tcl_CreateHashEntry(&HashTable, ArcIndex[arccount], &newptr);
	Tcl_SetHashValue(ptr,ap);
	arccount++;
}
#endif

