/* ----------------------------------------------------------------
 *   FILE
 *	libpq.h
 *
 *   DESCRIPTION
 *	POSTGRES LIBPQ buffer structure definitions.
 *
 *   NOTES
 *	This file contains definitions for structures and
 *	externs for functions used by both frontend applications
 *	and the POSTGRES backend.  See the files libpq-fe.h and
 *	libpq-be.h for frontend/backend specific information
 *
 *   IDENTIFICATION
 *	$Header: /u/jurafsky/CVSROOT/berp/tightc/libpq.h,v 1.2 1994/11/30 17:53:01 fosler Exp $
 * ----------------------------------------------------------------
 */

#ifndef LibpqIncluded 	/* include this file only once. */
#define LibpqIncluded 	1

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include "utils/exc.h"

/* ----------------
 * PQArgBlock --
 *	Information (pointer to array of this structure) required
 *	for the PQfn() call.
 * ----------------
 */
typedef struct {
    int len;
    int isint;
    union {
        int *ptr;	/* can't use void (dec compiler barfs)	*/
	int integer;
    } u;
} PQArgBlock;

/* ----------------
 * TypeBlock --
 * 	Information about an attribute.
 * ----------------
 */
#define NameLength 16

typedef struct TypeBlock {
    char name[NameLength];	/* name of the attribute */
    int adtid;			/* adtid of the type */
    int adtsize;		/* adtsize of the type */
} TypeBlock;

/* ----------------
 * TupleBlock --
 *	Data of a tuple.
 * ----------------
 */
#define TupleBlockSize 100

typedef struct TupleBlock {
    char **values[TupleBlockSize];	/* an array of tuples */
    struct TupleBlock *next;		/* next tuple block */
    int    tuple_index;			/* current tuple index */
} TupleBlock;

/* ----------------
 * GroupBuffer --
 * 	A group of tuples with the same attributes.
 * ----------------
 */
typedef struct GroupBuffer {
    int no_tuples;		/* number of tuples in this group */
    int no_fields;		/* number of attributes */
    TypeBlock *types;  		/* types of the attributes */
    TupleBlock *tuples;		/* tuples in this group */
    struct GroupBuffer *next;	/* next group */
} GroupBuffer;

/* ----------------
 * PortalBuffer --
 *	Data structure of a portal buffer.  
 * ----------------
 */
typedef struct PortalBuffer {
    int rule_p;			/* 1 if this is an asynchronized portal. */
    int no_tuples;		/* number of tuples in this portal buffer */
    int no_groups;		/* number of tuple groups */
    GroupBuffer *groups;	/* tuple groups */
} PortalBuffer;

/* ----------------
 * PortalEntry --
 *	an entry in the global portal table
 *
 * Note: the portalcxt is only meaningful for PQcalls made from
 *       within a postgres backend.  frontend apps should ignore it.
 * ----------------
 */
#define PortalNameLength 32

typedef struct PortalEntry {
    char 	  name[PortalNameLength]; /* name of this portal */
    PortalBuffer  *portal;	          /* tuples contained in this portal */
    Pointer	  portalcxt;	          /* memory context (for backend) */
    Pointer	  result;	          /* result for PQexec */
} PortalEntry;

#define MAXPORTALS 10

extern PortalEntry *portals[];

/*
 * Exceptions.
 */

#define libpq_raise(X, Y) ExcRaise((Exception *)(X), (ExcDetail) (Y),\
				   (ExcData)0, (ExcMessage) 0)

extern Exception MemoryError, PortalError, PostquelError, ProtocolError;

/* 
 * POSTGRES backend dependent Constants. 
 */

#define initstr_length 256
#define error_msg_length 4096
#define command_length 20
#define remark_length 80
#define portal_name_length 16

/*
 * External functions.
 */
extern "C" void pqdebug (char *target, char *msg);
extern "C" void pqdebug2 (char *target, char *msg1, char *msg2);
extern "C" void PQtrace ();
extern "C" void PQuntrace ();
extern "C" int PQnportals (int rule_p);
extern "C" void PQpnames (char **pnames, int rule_p);
extern "C" PortalBuffer *PQparray (char *pname);
extern "C" int PQrulep (PortalBuffer *portal);
extern "C" int PQntuples (PortalBuffer *portal);
extern "C" int PQninstances (PortalBuffer *portal);
extern "C" int PQngroups (PortalBuffer *portal);
extern "C" int PQntuplesGroup (PortalBuffer *portal, int group_index);
extern "C" int PQninstancesGroup (PortalBuffer *portal, int group_index);
extern "C" int PQnfieldsGroup (PortalBuffer *portal, int group_index);
extern "C" int PQfnumberGroup (PortalBuffer *portal, int group_index, char *field_name);
extern "C" char *PQfnameGroup (PortalBuffer *portal, int group_index, int field_number);
extern "C" GroupBuffer *PQgroup (PortalBuffer *portal, int tuple_index);
extern "C" int PQgetgroup (PortalBuffer *portal, int tuple_index);
extern "C" int PQnfields (PortalBuffer *portal, int tuple_index);
extern "C" int PQfnumber (PortalBuffer *portal, int tuple_index, char *field_name);
extern "C" char *PQfname (PortalBuffer *portal, int tuple_index, int field_number);
extern "C" int PQftype (PortalBuffer *portal, int tuple_index, int field_number);
extern "C" int PQsametype (PortalBuffer *portal, int tuple_index1, int tuple_index2);
extern "C" char *PQgetvalue (PortalBuffer *portal, int tuple_index, int field_number);
extern "C" void PQclear (char *pname);
extern "C" caddr_t pbuf_alloc (size_t size);
extern "C" void pbuf_free (caddr_t pointer);
extern "C" PortalBuffer *pbuf_addPortal ();
extern "C" GroupBuffer *pbuf_addGroup (PortalBuffer *portal);
extern "C" TypeBlock *pbuf_addTypes (int n);
extern "C" TupleBlock *pbuf_addTuples ();
extern "C" char **pbuf_addTuple (int n);
extern "C" char *pbuf_addValues (int n);
extern "C" PortalEntry *pbuf_addEntry ();
extern "C" void pbuf_freeEntry (int i);
extern "C" void pbuf_freeTypes (TypeBlock *types);
extern "C" void pbuf_freeTuples (TupleBlock *tuples, int no_tuples, int no_fields);
extern "C" void pbuf_freeGroup (GroupBuffer *group);
extern "C" void pbuf_freePortal (PortalBuffer *portal);
extern "C" int pbuf_getIndex (char *pname);
extern "C" PortalEntry *pbuf_setup (char *pname);
extern "C" void pbuf_close (char *pname);
extern "C" GroupBuffer *pbuf_findGroup (PortalBuffer *portal, int group_index);
extern "C" pbuf_findFnumber (GroupBuffer *group, char *field_name);
extern "C" void pbuf_checkFnumber (GroupBuffer *group, int field_number);
extern "C" char *pbuf_findFname (GroupBuffer *group, int field_number);
extern "C" void pq_init (int fd);
extern "C" void pq_gettty (int tp);
extern "C" int pq_getport ();
extern "C" void pq_close ();
extern "C" void pq_flush ();
extern "C" void pq_getstr (char *s, int maxlen);
extern "C" void pq_getnchar (char *s, int off, int maxlen);
extern "C" int pq_getint (int b);
extern "C" void pq_putstr (char *s);
extern "C" void pq_putnchar (char *s, int n);
extern "C" void pq_putint (int i, int b);
extern "C" int pq_getinaddr (struct sockaddr_in *sin, char *host, int port);
extern "C" int pq_getinserv (struct sockaddr_in *sin, char *host, char *serv);
int pq_connect (char *dbname , char *user , char *args , char *hostName , char *debugTty , char *execFile , int portName );
extern "C" int pq_accept ();
extern "C"  void PQsetdb (char *dbname);
extern "C"  void PQreset ();
extern "C"  void PQfinish ();
extern "C"  char *PQexec (char *query);

    
#endif LibpqIncluded
