/*******************************************************************************

	parsemain.C:

		The main routine for the parser.

		Author:		Dan Jurafsky
		Date:		October 1992
		History:	DSJ	6-4-93	Added header


	Depending on the command line arguments, the parser does different things.
	In all of them, it opens INPUTFILENAME, reads each line, and parses it bottom-up.
	In the basic version, called as follow:

		parser INPUTFILENAME
	it does this, and for each sentence, if it fails to parse, writes it to "parse.err".
	If it parses, it may call XlateQuery to clean up the semantics, then
	prints out information on the sentence, its probability, its semantics,

	In addition, if there is a "print" argument,
		parser INPUTFILENAME print 
	it prints out various parts of the chart when it is done.

	With the "count" argument, the parser uses the INPUTFILE to recompute 
	the rule probabilities for the grammar. In this case, nothing gets printed
	out for each individual parse, but at the very end, the parser writes out
	to stdout all the grammar rules with their new probabilities.  Of course,
	this should really be done iteratively, but one iteration is usually sufficient
	to 3 significant figures.

		parser INPUTFILENAME count

******************************************************************************/

#include "grammar.h"
#include "parser.h"
#include "db.h"
#include "bigram.h"
#include "unify_hash.h"
#include "unify_list.h"
#include <icsiargs.h>
#include <iostream.h>

/********************************* GLOBAL VARIABLES ************************/

static Grammar gr("ss",0);
static ChartParser p(&gr);

/*  These are for finding memory leaks */
int News = 0;
int Deletes = 0;
int NodeDeletes = 0;
int NodeCreates = 0;
int ArcNews = 0;
int ArcDeletes = 0;
int RuleNews = 0;
int RuleDeletes = 0;
int SymbolNews = 0;
int SymbolDeletes = 0;
int UnifyYes=0;
int UnifyNo=0;
extern int ConstraintListNews;
extern int ConstraintListDeletes;
extern int ConstraintNews;
extern int ConstraintDeletes;

char *wp_cont="./wp_middle.b.nfp";
extern int do_occurs_check;
extern int Unify;
extern int onlineUnify;
extern int do_cover_sem;
int topDown;

// stream for parse tree window (ptw)
extern FILE *s4ptw;
extern FILE *open_ptw(char *);
int pt=0;

/********************************* Main ************************/

main(int argc, char *argv[])
{
	double pr;
	char buf[BUFSIZ];
	char inbuf[BUFSIZ];
	char *s;
	int flags = 0;
	char s1[BUFSIZ];
	char *grammarfile = "./berp.grammar";
	char *inputfile = NULL;
	int dotrain = 0;
	int doprint = 0;
	int guessflag = 0;
	int guesslist[MAX_SENTENCE_LEN];
	float totalp = 0;
	float totalunp  = 0;
	int sentencecount = 0;
	int i;
	int parsed = 0, unparsed = 0;
	extern HashTable *stringTable;
    icsiargtab_t args[] = {
      { 0, "Parser will parse input strings, return their probabilities, "
          " and can also train scfg probabilitieson a corpus",  ARG_DESC},
      { "input", ": the sentences file as input", ARG_STR, &inputfile, ARG_OPT},
      { "grammar", ": the SCFG file (defaults to ./berp.grammar)", ARG_STR, &grammarfile, ARG_OPT},
      { "train", "true if the program should train scfg probabilities on this corpus",
	    ARG_BOOL, &dotrain },
      { "wp_cont", "the backoff bigram file (defaults to ./wp_middle.b.nfp)",
	ARG_STR, &wp_cont, ARG_OPT },
      { "occurs", "do the occurs check? (defaults to FALSE)", ARG_BOOL, &do_occurs_check, ARG_OPT },
      { "unify", "do unification? (defaults to TRUE)", ARG_BOOL, &Unify, ARG_OPT },
      { "online_unify", "do online unification? (defaults to FALSE, TRUE implies unify=TRUE)", ARG_BOOL, &onlineUnify, ARG_OPT },
      { "cover_sem", "guess semantics for failed parses? (defaults to TRUE)", ARG_BOOL, &do_cover_sem, ARG_OPT },
      { "topdown", "use top down parser rather than bottom up (default: FALSE)", ARG_BOOL, &topDown, ARG_OPT},
      { "debug", ": true to print out lots of debug info", ARG_INT, &doprint, ARG_OPT },
          { "parsetree", "print out parse trees", ARG_BOOL, &pt, ARG_OPT},

      { 0,0  }
    };
    char *appname;
	FILE *fpIn;
	do_occurs_check=0;
	do_cover_sem=1;
 
	Unify=1;
	onlineUnify=0;
	topDown=0;
	icsiargs(args,&argc, &argv, &appname);

	if (onlineUnify) Unify=1;

	if (inputfile == NULL) 
		fpIn = stdin;
	else
		fpIn = fopen(inputfile,"r");

	FILE *errorfp = fopen("parse.err","w");  // log of unparsed sentences
	FILE *goodfp = fopen("parse.good","w");  // log of parsed sentences

	s1[0] = 0;

	// Read in the grammar

	gr.ReadFile(grammarfile);
	
	if (dotrain)
		p.CountRulesInit();
	if (pt) {
#ifdef OLD_STUFF
#define PTWDIR "."
#define PTWNAME "ptbrowser"
	  /* open should be write-only, I think */
	  s4ptw = popen(PTWDIR "/" PTWNAME, "w");
	  if (s4ptw == NULL) {
	    perror(PTWDIR "/" PTWNAME);
	    /* exit(1); */
	  }
#else
	  s4ptw=open_ptw((char *)NULL);
#endif
	}

	/**** The main loop: read in lines and parse them ******/

	while (fgets(inbuf,BUFSIZ,fpIn) != NULL) {
		if (inbuf[strlen(inbuf)-1] == '\n')
			inbuf[strlen(inbuf)-1] = '\0';

		strcpy(buf,"BEG ");
		strcat(buf,inbuf);
		strcat(buf," END");
		
		if (doprint)
		  cout << ":" << buf << "\n";

		// clear out the guess list 
		for (i=0;i<30;i++)
			guesslist[i] = 0;

		// parse
		pr=0;
		if (topDown)
		  s = p.TopDownParser(buf,&pr,dotrain,doprint);
		else
		  s = p.Parser(buf,&pr,dotrain,doprint,&guessflag,guesslist);

		/* if we're just learning the rule probabilities we don't need
			any of the rest of this stuff, but otherwise we need
			to tell what the guesslist was, do the contextmodule xlations,
			and of course tell the user if this sentence parsed or not */

		if (!dotrain) {
			cout << inbuf << flush;
			cout << " Prob: " <<  pr << " , Sem: " << s << "\n" << flush;
			if (pr > 0) {
				fprintf(goodfp,"%s\n",inbuf);
				fflush(goodfp);
			} else {
				fprintf(errorfp,"%s\n",inbuf);
				fflush(errorfp);
			}

			if (s && s[0]) {

				/* The parser returns the semantics of the sentence
					in the string 's' - if s is non-NULL, the sentence
					has a legitimate semantics */

				if (guessflag) {

					/* The guesslist lists the words which were used
						in building the semantics; if the guessflag is
						TRUE, it means that the full sentence was not
						parsed, and that the fragment algorithm built the
						semantics */

					cout << "\nGuessed:";
					for (i=0;i<30;i++) {
						if (guesslist[i])
							cout << " " << i;
					}
					cout << endl;
				}
#ifdef NOTDEF
				// this happens to be commented out for now, since 
				// i've been testing the syntax-only lately
				flags = XlateQuery(s,s1,1);
				cout << s1 << "\n";
#endif
				s1[0] = 0;
				//free(s);
			}
		} else {
			if (pr > 0)
				parsed++;
			else
				unparsed++;
			fprintf(stderr,".");
		}
		/* Reset the parser by freeing all the arcs in the chart, etc */
		p.Reset();
		ios::sync_with_stdio();
		cout.flush();
		ios::sync_with_stdio();

		// free up all dags in existence - should be fixed
		Dag::flush_the_toilet();
		if (s)
		  delete[] s;  // allocated in Arc::GetSemFromDag()
	}

	/*  If we're learning rule probabilities, normalize'm so they all sum to one,
		and dump'em to stdout */

	if (dotrain) {
		cout << "\n" << flush;
		p.NormalizeRules();
		p.PrintRuleCount();
		fprintf(stderr,"\nParsed: %d, Unparsed: %d\n",parsed,unparsed);
	}

	gr.Reset();
	delete stringTable;
	Link_free_list.clobber();
	Dag::clobber_free_list();
	
	// checking for memory leaks - I've still got a couple
	if (doprint) {
		printf("\n%d calls to operator new\n%d calls to operator delete\n",News,Deletes);
		printf("\n%d calls to operator Nodenew\n%d calls to operator Nodedelete\n",NodeCreates,NodeDeletes);
		printf("\n%d calls to operator Symbolnew\n%d calls to operator Symboldelete\n",SymbolNews,SymbolDeletes);
		printf("\n%d calls to operator Arcnew\n%d calls to operator Arcdelete\n",ArcNews,ArcDeletes);
		printf("\n%d calls to operator Rulenew\n%d calls to operator Ruledelete\n",RuleNews,RuleDeletes);
		printf("\n%d calls to operator ConstraintListnew\n%d calls to operator ConstraintListDelete\n",ConstraintListNews,ConstraintListDeletes);
		printf("\n%d calls to operator Constraintnew\n%d calls to operator ConstraintDelete\n",ConstraintNews,ConstraintDeletes);
		
	}
	printf("Unify yes: %d\n",UnifyYes);
	printf("Unify no: %d\n",UnifyNo);
	
	exit(0);
}

void *operator new(size_t sz)
{
	++News;
	return malloc(sz);
}

void operator delete(void *p)
{
	++Deletes;
	free((char *)p);
}

	
