/*******************************************************************************

	parser.C:
		
		Contains class:	 ChartParser

		Author:		Dan Jurafsky
		Date:		September, 1992
		History:	DSJ	6-7-93	Added header

******************************************************************************/

#include "grammar.h"
#include "parser.h"
#include "bigram.h"

/*******************************************************************************

	class ChartParser:	
			
	Data:
		Chart	c;
		Grammar	*grammar;
		Bigram	*bg;
		double	*RuleCountArray;
		float 	termarray[MAXSYMBOLS];

	Constructor:
		ChartParser(Grammar *);

	Methods:
			Symbol  *GetWord(char*&);
		void	Reset();
		void	CountRulesInit();
		void	SetBigram(Bigram *);
		void 	CountRules();
		char	*Parser( char *,double *,int,int, int*,int,int*,float*,float*);
		void 	PrintRuleCount();
		void 	NormalizeRules();
		char *TopDownParser(char *,double*, int , int , int ,float *pfollowsize, float *unpfollowsize);
		void TopDownInit(char * , int *, int*,int *,int *);
		void Propose(Symbol*,int,double );
		int ComputeFollowSet(int)
		int PrintTopFollowSet(int,int, int* ,int *, int *);

*******************************************************************************/

ChartParser::ChartParser(Grammar *g)
{
	grammar = g;
	UnkWord = grammar->Symbols.AddSymbol("'???'");
	EndWord = grammar->Symbols.AddSymbol("'END'");
	Pause = grammar->Symbols.AddSymbol("'.'");
}

void ChartParser::Reset()
{
	c.Reset();
}

char *ChartParser::Parser(char *words, double*prob, int countflag, int printlevel, int *guessflag,int *guesslist)
{
	char *p = words;
	Symbol *s;
	int icount,i;
	int index = 0;
	int from = 0,to = 0;
	Arc *kp;
	Tree *tp;
	ArcIterator next = &c.Arcs;
	Arc *ap;
	Tree *t = NULL;
	int len = 0;
	char sem[300];
	int followcount = 0;
	int numparses = 0;
	int wordindex[70];
	int unprunedfollowcount = 0;
	Rule * rules[MAXRHSRULES];
#ifdef TATEYAMA
	char *wordssave;
	extern FILE *s4ptw;
#endif


	if (words == NULL)
		return NULL;

#ifdef TATEYAMA
	wordssave = strdup(words);
#endif

	c.Start = grammar->GetStart();
	c.Parsetype = BOTTOMUP;

	while (TRUE) {
		
		/* Pop the next finished constit off the
		  stack.  If the stack is empty, get a word instead */

		if ((kp = c.PopKey()) != NULL) {
			s = kp->lhs;
			from = kp->From;
			to = kp->To;
			if (kp->ArcTree)
				t = kp->ArcTree->Copy();
#ifndef NOPARSE
			if (kp->Sem)
				strcpy(sem,kp->Sem);
			else
#endif
				sem[0] = '\0';


			// Add the completed rule to the chart

			c.AddConstit(kp);

		} else {

			// Nothing on stack, get next word instead

			if (s = GetWord(p)) {
				len++;
				//cout << "Got Token " << s << "\n" << flush;
				//printf("Got Token %s\n",s);
				//fflush(stdout);
				wordindex[index] = s->ID();
			} else	{
				c.Len = len;
				if (countflag) {
					CountRules();
 					// *prob = c.GetSentenceProb();
					return c.GetSem(c.Start,len,prob);
				} else {
					numparses = c.GetNumberOfParses(c.Start,len);
					if (printlevel >= PRINTCHART) {
						c.print();
						tp = c.GetParseTree(c.Start,len);
						while (tp) {
							cout << "Parse tree is: ";
							tp->print();
							tp = c.GetParseTree(c.Start,len);
						}
					}
					if (printlevel >= PRINTPROB) {
						cout << "Number of parses for this sentence is "  << numparses << "\n";
						cout << "Sentence Prob is " << c.GetBottomUpSentenceProb() << "\n" << flush;
					}
#ifdef TATEYAMA
					if (NULL != s4ptw) {
					  c.sendAllParses2ptw(wordssave);
					  fflush(s4ptw);
					}
#endif
					free(wordssave);
					if (numparses)  {
						*guessflag = 0;
#ifdef NOPARSE
						return NULL;
#else
						return c.GetSem(c.Start,len,prob);
#endif
					} else {
						*guessflag = 1;
#ifdef NOPARSE
						return NULL;
#else
						return c.GetCoverSem(guesslist);
						// null if !do_cover_sem
#endif
					}
				}
			}
			from = index;
			to = ++index;
			if (index == 70) 
				error("sentence is too long\n");

		// this out??? - no, it blows up later
			t = new Tree(s);

#ifndef NOPARSE
			sem[0] = '\0';
#endif

			// Add the new word to the chart

			c.AddConstit(s,from,to);
		}
		

		/* Now get all the rules that this word begins,
		and add them to the chart */

		icount = grammar->GetRulesWithRhs(s,rules);

#ifdef DEBUG
		for (i=0;i<icount;i++)
		  rules[i].print();
#endif

		for (i=0;i<icount;i++)  {
			c.NewArc(rules[i],from,to,t,sem);
		}

		/* Now check all the previous active arcs which end at 
		  our starting point. If our word can continue any such arc A,
		  then add a new arc from the start of the sentence to
		  our ending point, with the rule from A properly updated */

		next.Reset();
		while (ap = next(s,from)) {
			c.ExtendArc(ap,to,t,sem);
		}
		delete t;
	}
		



}
/*******************************************************************************

CountRules:  For a given sentence, iterator over all parse trees 'treep',
				and for each parse tree, count the number of times each grammar
				rule is used, weighting the count by the parse-tree
				probability.  This will compute an (unnormalized) rule
				count for each rule in the parser, based on the sentence
				that was just parsed.
				
******************************************************************************/

void ChartParser::CountRules()
{
	double SentProb;
	RuleIterator next(grammar);
	ParseTreeIterator pt(&c);
	Rule *rulep;
	Tree * treep;
	int num;
	double prob;
	int i;

	SentProb = c.GetSentenceProb();

	//cout << "sent prob is " << SentProb << "\n" << flush;

	while (treep = pt.Next()) {
		treep->CountRulesUsed(RuleCountWork);
		prob = treep->GetProb();
		for (i=0;i < grammar->NRules(); i++) {
			if (RuleCountWork[i]) {
				RuleCountArray[i] += (double)( prob / SentProb ) * (double)RuleCountWork[i];
				RuleCountWork[i] = 0;
			}
		}
	}
}
#ifdef NOTDEF
void ChartParser::CountRules()
{
	double SentProb;
	RuleIterator next(grammar);
	ParseTreeIterator pt(&c);
	Rule *rulep;
	Tree * treep;
	int num;
	double prob;

	SentProb = c.GetSentenceProb();

	while (rulep = next()) {
		while (treep = pt.Next()) {
			num = treep->CountRulesUsed(rulep->RuleNum);
			prob = treep->GetProb();
			RuleCountArray[rulep->RuleNum] += (double)( prob / SentProb ) * (double)num;
		}
		pt.Reset();
	}
}
#endif


/*******************************************************************************

		CountRulesInit:  Allocate and zero out the rule count array
				
******************************************************************************/

void ChartParser::CountRulesInit()
{
	int i;
	int nrules = grammar->NRules();

	RuleCountArray = (double *)malloc(sizeof(double) * nrules);
	RuleCountWork = (int *)malloc(sizeof(int) * nrules);
	for (i=0;i<nrules;i++) {
		RuleCountArray[i] = (double)0;
		RuleCountWork[i] = 0;
	}
}

/*******************************************************************************

	NormalizeRules:  Turn all the counts into probabilities by normalizing
				the counts for each particular rule 'r' so that
				they sum to '1'.  Right now we smooth the counts
				by just assuming all rules which never occurred
				have actually occurred 0.001 times.  Got to fix 
				this with Good-Turing or Deleted Interp.
			
******************************************************************************/

void ChartParser::NormalizeRules()
{
	int i;
	Rule *p;
	double prob = 0;
	RuleIterator next = &grammar->Productions;
	NonTermIterator nonterm = &grammar->Symbols;
	Symbol *symp;
	float totalcount;
	int nrules = grammar->NRules();

	/*  Right now we use this dumb smoothing algorith, */
	for (i=0;i<nrules;i++) {
		//printf("rulecount[%d]=%g\n",i,RuleCountArray[i]);
		if (RuleCountArray[i] == 0)
			RuleCountArray[i] = 0.1;
			//RuleCountArray[i] = 0.001;
	}

	while (symp = nonterm.Next()) {
		totalcount = 0;
		next.Reset(symp);
		while (p = next(symp)) {
			totalcount += RuleCountArray[p->RuleNum];
		}
		next.Reset(symp);
		while (p = next(symp)) {
			if (totalcount)
				RuleCountArray[p->RuleNum] = RuleCountArray[p->RuleNum] / totalcount;
			p->prob = RuleCountArray[p->RuleNum];
		}
	}
}


/*******************************************************************************

		PrintRuleCount:  Dump out the grammar with the new probabilities.
				
******************************************************************************/

void ChartParser::PrintRuleCount()
{
	int i;
	int nrules = grammar->NRules();

	for (i=0;i<nrules;i++) {
		//cout << RuleCountArray[i] << " [" << i << "] ";
		grammar->GetRule(i)->print();
	}
	cout << flush;
	free((char *)RuleCountArray);
}

/*******************************************************************************

	ComputeFollowSet:   Given a chart position 'to', compute all the
				words which could come next.  We do this
				by going through every arc in the chart which
				could end at 'to', look at the next symbol
				for this arc, and asking the grammar what are
				the possible single-terminal-prefixs
				of the expansions of this symbol.
				
******************************************************************************/

int ChartParser::ComputeFollowSet(int to)
{
	int i;
	ArcIterator next = &c.Arcs;
	Arc * ap;
	int count = 0;
	int fsize = 0;
	//float probarray[grammar->Symbols.NTerms()];
	double p;
	double psum = 0;


	/* If we have not read in the LeftCorner
		matrix, we can't possible really want
		to compute follow probs, so just return */

	if (grammar->LeftCorner == NULL)  
		return 0;

	//for (i=0;i < grammar->Symbols.NTerms();i++) {
		//probarray[i] = 0;
	//}

	/*  termarray is an array which is used to pass follow-set information 
		around.  It is a data member of the class ChartParser which is 
		non-zero for every element which is in the follow-set */

	for (i=0;i<MAXSYMBOLS;i++)
		termarray[i] = 0;

	/* The visit list is used to speed up the prefix computation */

	grammar->ResetVisitList();

	//cout << "Current Follow set at point " << to << ":\n" << flush;
	while (ap = next()) {

			//cout << "Looking at arc: ";
			//ap->print();
			//cout << flush;

		// For each arc in chart which is still active at current position

		if ((ap->To == to) && (!ap->IsComplete) && 
				((ap->state == STATE_COMPLETION)
				|| (ap->state == STATE_SCAN)))  {

			/* let p be the probability of the prefix parse tree
				associated with this arc */
				
			p = ap->prefixprob;
			psum += ap->prefixprob;

			//cout << "Adding in arc: ";
			//ap->print();
			//cout << flush;

			/* the probarray has an entry for each terminal. For each one,
				we need to add in the probability that this arc
				will left-expand to this terminal, times the probability
				of the parse of the prefix, (p). */

			TermIterator termit = &(grammar->Symbols);
			Symbol * symp = NULL;
			while( (symp = termit.Next()) != NULL) {
				i = symp->ID();
				//if (grammar->Symbols.GetSymbol(i)->IsTerminal()) {
					termarray[i] += (p * ((double **)(grammar->LeftCorner->ptr))[ap->NextSym->ID()][i]);
					if (((double **)grammar->LeftCorner->ptr)[ap->NextSym->ID()][i]) {
						//cout << "for terminal symbol ";
						//grammar->Symbols.GetSymbol(i)->print();
						//cout << "LeftCorner prob is " << ((double **)grammar->LeftCorner->ptr)[ap->NextSym->ID()][i] << "\n";
						//cout << flush;
					}
				//}
			}
		}
	}

	//c.print();
	//cout << flush;
	//float testsum = 0.0;

	// Compute the follow-set size
	for (i=0;i<MAXSYMBOLS;i++) {
		if (termarray[i] != 0.0)  {
			//cout << "termarray is " << termarray[i] << "; testsum: " << testsum << "\n" << flush;
		    //testsum += termarray[i];
			termarray[i] = termarray[i] / psum;
			fsize++;
		}
	}
	//cout << "testsum is " << testsum << ",psum is " << psum << "\n" << flush;
	//cout << "prefix prob is " << c.GetSentenceProb() << "\n" << flush;

	/* All this BIGRAM stuff is so we can compute how much the CFG is filtering
		the bigram follow-set size. I only ifdef it when making these measurements. */

#ifdef NOTDEF
	for (i=0;i<MAXSYMBOLS;i++) {
		if (termarray[i]) {
			count++;
			temps = new Symbol(i);
			cout << i << " " << temps->stringify() << "\n";
			delete temps;
		}
	}
#endif
	//printf("Follow set size of %d\n",fsize);
	//fflush(stdout);
	return fsize;
}
/*******************************************************************************

		CopyOutTermArray: used by follow.C

*******************************************************************************/

void ChartParser::CopyOutTermArray(float * out)
{
	int i;
	for (i=0;i<MAXSYMBOLS;i++) 
		out[i] = termarray[i];
}

/*******************************************************************************

  Some utilities:
	NotWord - true if the string passed it is a filled pause ("[um]")
		a comment ("/"), a pause ("."), the unknown word ("???"),
		a deleted word ("<asdf>"), a sentence id ("cfg01.009"),
		or a fragment ("th-").
	RemoveStars - take off the stars that surround mispronounced words
	GetWord	- pull the next word from the input buffer by repeatedly
			grabbing a token until it gets one that is
			a legal word (!NotWord())
		
******************************************************************************/


int NotWord(char* s)
{
	int slen = strlen(s);
	return ((s[0] == '/') || 
		(s[0] == '.') || 
		(s[0] == '[') || 
		(s[0] == '?') || 
		(s[0] == '-') || 
		(s[0] == '<') || 
		((slen == 12) &&
			(s[8] == '.')) ||
        ((slen == 8) && (s[0]=='0'||s[0]=='1'||s[0]=='2') && (s[6]=='s')) ||
		((slen  == 9) &&
			(s[5] == '.')) ||
		(s[slen-1] == '-'));
}


void RemoveStars(char * s)
{
	int i,j;
	int len;
	char buffer[MAXSYMLEN];

	len = strlen(s);
	for (i = 0,j = 0; i < len; i++,j++) {
		if ((i == 0) && (s[i] == '-'))
			buffer[j] = s[i];
		else if ((s[i] != '*') && (s[i] != '!') && (s[i] != ':')) {
			buffer[j] = s[i];
		} else
			j--;
	}
	buffer[j] = 0;
	strcpy(s,buffer);
}

inline int IsGarbage(char *s)
{
	return !strcmp(s,"???");
}

Symbol * ChartParser::Tokenizer(char*& inp)
{
	int i;
	char buffer[MAXSYMLEN];
	Symbol * LastToken = NULL;
	char *p = inp;
	
	ntokens = 0;

	for (i=0; i< 70; i++) {
		buffer[0] = '\'';
		while (TRUE) {
			if (GetNextToken(buffer+1,p)) {
				//printf("got Token (%s)\n",buffer);
				fflush(stdout);
				 
				if (IsGarbage(&buffer[1])) {
					LastToken = UnkWord;
					continue;
				}
				if (buffer[1] == '.')  {
					LastToken = Pause;
					continue;
				}
				if (NotWord(&buffer[1])) {
					LastToken = NULL;
					continue;
				} else  {
					LastToken = NULL;
					break;
				}
			} else
				return LastToken;
		}
		RemoveStars(buffer);
		buffer[strlen(buffer)+1] = '\0';
		buffer[strlen(buffer)] = '\'';
		//printf("Tokenizer: buffer is [%s]\n",buffer);
		tokens[i] = grammar->Symbols.AddSymbol(buffer);
		LastToken = tokens[i];
		ntokens++;
	}
	return LastToken;
}


Symbol * ChartParser::GetWord(char*& p)
{
	char buffer[MAXSYMLEN];
	buffer[0] = '\'';
	while (TRUE) {
		if (GetNextToken(buffer+1,p)) {
			if (NotWord(&buffer[1]))
				continue;
			else 
				break;
		} else
			return FALSE;
	}
	RemoveStars(buffer);
	buffer[strlen(buffer)+1] = '\0';
	buffer[strlen(buffer)] = '\'';
	//printf("GetWord: buffer is [%s]\n",buffer);
	return grammar->Symbols.AddSymbol(buffer);
}

/*******************************************************************************

		TopDownParser:  
				
******************************************************************************/

char *ChartParser::TopDownParser(char *words,double*pr, int countflag, int printlevel)
{
	Arc *kp;
	ArcIterator next = &c.Arcs;
	Arc *ap;
	Tree *t;
	int len = 0;
	char sem[300];
	Symbol *s;
	int icount,i;
	int from,to;
	int complete;
	int index = 0;
	char *p = words;
	Tree *tp;
	int wordindex[40];
	Rule * rules[MAXRHSRULES];
	double prob = 0;
#define BIGRAM
#ifdef BIGRAM
	int begindex = grammar->Symbols.IsSymbol("BEG");
#endif

	c.Parsetype = TOPDOWN;
	//printf("made it into parser for '%s'\n",words);
	fflush(stdout);
	if (words == NULL)
		return NULL;


	c.Start = grammar->GetStart();

	TopDownInit();
	last_token = Tokenizer(p);
	//printf("last token is %s\n",last_token->NoQuotesStringify());

	//printf("ntokens is %d\n",ntokens);
	if (ntokens > 65)
		error("sentence too long");

	while (TRUE) {

		/* Pop the next Pending Arc from the Agenda */

		if ((kp = c.Agenda.Get()) != NULL) {
			c.Arcs.AddArc(kp);
			s = kp->lhs;
			from = kp->From;
			to = kp->To;
			if (kp->ArcTree) 
				t = kp->ArcTree->Copy();
			else
				t = NULL;
			if (kp->Sem)
				strcpy(sem,kp->Sem);
			else
				sem[0] = '\0';
			complete = kp->IsComplete;

			if (complete) {
				next.Reset();
				// Earley completion 
				while (ap = next(s,from)) {
					c.TopDownExtendArcUp(ap,to,t,sem,kp->insideprob,s->IsTerminal());
				}
				c.AddConstit(kp);
			} else {
				// Earley prediction
				if (kp->iNumChildren) {

				    if ((to >= ntokens) || ((grammar->LeftCorner != NULL) &&	
								(((double **)(grammar->LeftCorner->ptr))[kp->NextSym->ID()][tokens[to]->ID()] == (double) 0.0))) {
							/*  Look in the LeftCorner matrix (if there is 
								one) at the prob of the next nonterminal 
								generating the next terminal (if there
								is one).  If this prob is zero, don't
								predict this nonterminal, else do */
						} else {
							if (printlevel) {
								cout << "in arc";
								kp->print();
								cout << "Proposing " << kp->NextSym->NoQuotesStringify() << "with prob " << kp->prefixprob << " at position " << to << " with next word " << tokens[to]->NoQuotesStringify() << "\n" << flush;
							}
					    	Propose(kp->NextSym,to,kp->prefixprob);
						}
				}
			}
		    delete t;
		} else  {

			//if (s = GetWord(p)) {
			if (index < ntokens) {
				s = tokens[index];
				//cout << "Got Token " << s->NoQuotesStringify() << "\n" << flush;
				wordindex[index] = s->ID();
				len++;
				from = index;
				to = ++index;
				if (index == 65) 
					error("sentence is too long\n");

				/* Now get all the rules that this word begins,
				and add them to the chart */

				icount = grammar->GetRulesWithRhs(s,rules);

				//printf("TopParser: got %d rules from word",icount);
				//s->print();
				//fflush(stdout);

				for (i=0;i<icount;i++)  {
					next.Reset();
					while (ap = next(rules[i]->lhs,from)) 
						prob += ap->prefixprob;
#ifdef BIGRAM
					//printf("checking if symbol %s is BEG",s->NoQuotesStringify());
					if (s->ID() == begindex) 
						prob = 1;
#endif
					t = new Tree(s);
					c.TopDownNewArc(rules[i],from,to,t,NULL,1,prob,TRUE);
					delete t;  // I think this is ok, because trees get copied everywhere.
					prob=0;
				}
				ap = new Arc(s,from,to);
				ap->state=STATE_NONE;
				ap->ArcTree = new Tree(s);
				ap->insideprob = 1;
				c.Agenda.Add(ap);
	
			} else	{
				c.Len = len;
				(void)ComputeFollowSet(index);
				if (countflag)
					CountRules();
				if (printlevel >= PRINTCHART) {
					c.print();
					//tp = c.GetParseTree(c.Start,len);
					//if (tp) {
						//cout << "Parse tree is: ";
						//tp->print();
					//}
				}
				*pr = c.GetSentenceProb();
				//fprintf(stderr,"pr is %g\n",*pr);
			    //fflush(stderr);
				return NULL;
#ifdef NOTDEF
				cout << "Average Follow size is " << (float) followcount / (float) (len+1) << "\n" << flush;
				cout << "Average UnFiltered Follow size is " << (float) followcount / (float) (len+1) << "\n" << flush;
//				cout << "Number of parses for this sentence is "  << c.GetNumberOfParses(c.Start,len) << "\n" << flush;
//				cout << "Sentence Prob is " << c.GetSentenceProb() << "\n";
//				return c.GetParseTree(c.Start,len);
//				return c.GetSem(c.Start,len,prob);
//				return c.MakeParseTree(c.Start);
#endif
			}
		}
	}
}

/*******************************************************************************

	Propose:  Propose symbol 's' at position 'where', which means
			grab all rules 'r' which could be direct expansions 
			of 's', and stick them on the Agenda to be pulled off
			and added to the chart later.
		If the rules are already there (we just check the first one to
			make sure) then just update their probabilities instead.
			
******************************************************************************/

void ChartParser::Propose(Symbol *s,int where,double prob)
{
	int count, i;
	Rule * rules[MAXRHSRULES];

	count = grammar->GetRulesWithLhs(s,rules);
	//printf("about to add %d rules\n",count);
	if (!count)
		return;
	if (c.GetArc(rules[0]->RuleID(),where,where)) {
#ifdef DEBUG
	  printf("...already proposed.\n");
	  fflush(stdout);
#endif
		for (i=0;i<count;i++)  
			c.UpdateArcProb(rules[i],where,where,prob);
	} else {
#ifdef DEBUG
	  printf("...a new arc.\n");
	  fflush(stdout);
#endif
		for (i=0;i<count;i++)  
			c.TopDownNewArc(rules[i],where,where,NULL,NULL,0,prob,FALSE);
	}
}

/*******************************************************************************

	TopDownInit:  Start off the whole shebang by proposing the start
			symbol.
				
******************************************************************************/

void ChartParser::TopDownInit()
{
	Propose(c.Start,0,1);
}

/*******************************************************************************

	ComputeLCArray:  Compute or read in the Left Corner Array
				
******************************************************************************/


void ChartParser::ComputeLCArray()
{
	RuleIterator next(grammar);
	Rule *rulep;
	MATRIX *leftcorner, *identity, *temp;
	float **l,**id;
	int size;
	int i;
	DSP_FILE *aout;
	FILE *fp = NULL;
	char * trailer;
        char *read_trailer();

	fp = fopen("matrix.dat","r");
	if (fp != NULL) {
		fclose(fp);
		//printf("Reading Left-Corner Matrix\n");
		//fflush(stdout);
                if (!(grammar->LeftCorner = matrix_trailer_read("matrix.dat",trailer)))
                {
                  fprintf(stderr,"Couldn't read matrix.dat!\n");
                  exit(-1);
                }
		if (grammar->Header[0] && strcmp(trailer,grammar->Header)) {
			fprintf(stderr,"Left-Corner Matrix doesn't match grammar:\n Left corner matrix '%s'\nGrammar '%s'\n",trailer,grammar->Header);
			exit(0);
		}
		free (trailer);
		//printf("Done Reading Left-Corner Matrix\n");
		//fflush(stdout);
		return;
	}
	
	size = grammar->Symbols.NSymbols();
	printf("size is %d\n",size);

	// allocate our matrices 

	leftcorner = matrix_allocate(size,size,sizeof(float));
	l = (float **)leftcorner->ptr;
	identity = matrix_allocate(size,size,sizeof(float));
	id = (float **)identity->ptr;

	// create the identity matrix I

	for(i=0;i<size;i++)
			id[i][i] = 1;

	while (rulep = next()) {
		l[rulep->lhs->ID()][rulep->rhs[0]->ID()] += rulep->prob;
		//cout << "analyzing";
		//rulep->print();
		//cout << "setting l[" << rulep->lhs->ID() << rulep->lhs->NoQuotesStringify() << "][" << rulep->rhs[0]->ID() << rulep->rhs[0]->NoQuotesStringify() << "]";
		//cout << " to " << l[rulep->lhs->ID()][rulep->rhs[0]->ID()] << "\n";
	}

	/* don't set the diagonals for the terminals to 1.0 */

	//for (i=0;i<size;i++) {
		//if (Symbol(i).IsTerminal())
			//l[i][i]=1.0;
	//}

	temp = matrix_sub(identity,leftcorner);

	//printf("\nInitial Left-corner matrix:");
	//matrix_print(temp);
	printf("\nAbout to invert\n");
	fflush(stdout);

	grammar->LeftCorner = matrix_invert(temp);
	printf("\nDone with invert\n");
	fflush(stdout);

	//printf("\n\nLeftcorner matrix inverse (lcinvert):");
	////matrix_print(LeftCorner);

	//fp = fopen("matrix.dat","w");
	aout = matrix_write(grammar->LeftCorner,"matrix.dat");
	if (aout == NULL)
		error("error writing matrix\n");
	else {
		if (grammar->Header[0])
			write_trailer(grammar->Header,aout);
		else
			write_trailer("No Header",aout);
	}

	matrix_free(leftcorner);
	matrix_free(identity);
}

