/*******************************************************************************

	rules.C:
		
		Contains classes:
			Symbol
			SymbolCollection
			NonTermIterator
			Rule
			RuleCollection

		Author:		Dan Jurafsky
		Date:		September, 1992
		History:	DSJ	5-27-93	Added header

******************************************************************************/
#include "grammar.h"
#include <stdio.h>
#include <assert.h>
extern int RuleNews;
extern int RuleDeletes;
extern int SymbolNews;
extern int SymbolDeletes;
extern int Unify;

//#ifndef __GNUC__
//extern "C" {
//	extern int      vfprintf(FILE *, const char *, void *);
//}
//#endif

extern SymbolCollection *sc;

inline int isspace(char c)
{
	return ((c == ' ') || (c == '\t') || (c == '\n'));
}

void error(const char* fmt ...) 
{
	va_list args;

	va_start (args,fmt);

	fprintf(stderr,"Fatal error: ");
	vfprintf(stderr,fmt,args);
	fprintf(stderr,"\n");
	exit(1);
}

/************************************************************************

	NOTE: Need to get rid of this class and replace it and SymbolCollection
		with a nice efficient symbol table.

	class Symbol: Each symbol has the following private data:
	
		int	id;			// id = its index in the symbol collection
		char 	Name[MAXSYMLEN];	// a null-terminated string, fixed-length
							just for simplicity
		int	terminal;		// boolean TRUE = terminal, FALSE = non-

	Plus four constructors: 
			Symbol(char*);
			Symbol(int);
			Symbol();
			Symbol(char*,int);
			~Symbol(char*);

	And these useful functions:
		int	IsTerminal()   {return terminal;}
		int	IsNonTerminal()  {return (!terminal);}
		int	ID()	{return id;}
		char 	*Stringify();
		char 	*NoQuotesStringify();
		void	print();
		void	Reset();
		friend	int operator !=(Symbol s1,Symbol s2)
			{return (s1.terminal!=s2.terminal)||(s1.id != s2.id);}
		friend	int operator ==(Symbol s1,Symbol s2)
			{return ((s1.terminal==s2.terminal)&&((s1.id != -1) && 
					(s1.id == s2.id)));}
	};

************************************************************************/

Symbol::~Symbol()
{
	SymbolDeletes++;
}

Symbol::Symbol()
{
  SymbolNews++;
  Name[0] = '\0';
}

Symbol::Symbol(char *symname,int ind)
{
  SymbolNews++;
	strcpy(Name,symname);
	id = ind;
	terminal = FALSE;
}

Symbol::Symbol(int ind)
{
  SymbolNews++;
	strcpy(Name,sc->Symbols[ind].Name);
	id = ind;
	terminal = TRUE;
}

char *Symbol::Stringify()
{
	char *ret = (char *)malloc(strlen(Name)+3);
	if (!terminal)  {
		strcpy(ret,Name);
	} else {
		ret[0] = '\'';
		strcpy(&ret[1],Name);
		ret[strlen(ret)+1] = 0;
		ret[strlen(ret)] = '\'';
	}
	return ret;
}
char *Symbol::NoQuotesStringify()
{
//	char *ret = malloc(strlen(Name)+1);
//	strcpy(ret,Name);
//	return ret;
	return Name;
}

void Symbol::print(ostream &os)
{
	char * x = Stringify();
	//cout << " " << x << " ";
	os << x;
	free(x);
	
}

//ostream& operator<<(ostream&s, Symbol S)
//{
//   return s << S.stringify();
//}
//ostream& operator<<(ostream&s, Symbol *S)
//{
//   return s << S->stringify();
//}


inline int hasquotes(char * s)
{
	return (s[0] == '\'');
}

void Symbol::Reset()
{
	Name[0] = '\0';
	id = 0;
	terminal = FALSE;
}


/**************************************************************************

	class SymbolCollection:   maintains the lexicon for a grammar.
				It consists of an array of size MAXSYMBOLS
				of Symbols.  The array is indexed by a hash table;
				the hash function is passed a string, and returns
				the array index.
		
	Data:

		Symbol          Symbols[MAXSYMBOLS];
		int             iTerms;
		int             iNonTerms;
		int             iSymbols;
		int		iLastTerm;
		Tcl_HashTable           HashTable;

	Methods:
				SymbolCollection();
		Symbol *        AddSymbol(char *);  
		int             NSymbols();
		int             NTerms();
		int             IsSymbol(char *);
		Symbol*         GetSymbol(int num);

***************************************************************************/


SymbolCollection::SymbolCollection()
{
	iSymbols = 0;
	Tcl_InitHashTable(&HashTable,TCL_STRING_KEYS);
}
void SymbolCollection::SetLastSym(int sym)
{
	LastSym = sym;
}

int SymbolCollection::IsSymbol(char *symname)
{
	int newptr = 0;
	Tcl_HashEntry * ptr;

	ptr = Tcl_FindHashEntry(&HashTable, symname);
	
	if (ptr == NULL) {
		return -1;
	} else
		return (int)Tcl_GetHashValue(ptr);
}

Symbol * SymbolCollection::AddSymbol(char *name)
{
	int index;
	int newptr = 0;
	Tcl_HashEntry * ptr;
	int terminal = FALSE;
	char  symname[MAXSYMLEN];

	if (hasquotes(name)) {
		// strip off the quotes
		strcpy(symname,name+1);
		symname[strlen(symname) - 1] = 0;
		terminal = TRUE;
	} else
		strcpy(symname,name);

	/*if (hasquotes(name)) {
		// strip off the quotes
		symname = name+1;
		symname[strlen(symname) - 1] = 0;
		terminal = TRUE;
	} else
		symname = name;*/
	
	ptr = Tcl_CreateHashEntry(&HashTable, symname, &newptr);
	if (newptr == 0) {
		// this entry already there;
		return &Symbols[(int)Tcl_GetHashValue(ptr)];
	} else {
		index = iSymbols++;
		strcpy(Symbols[index].Name,symname);
		Symbols[index].id = index;
		if (terminal) {
			iTerms++;
			Symbols[index].terminal = TRUE;
		} else {
			iNonTerms++;
			Symbols[index].terminal = FALSE;
		}
		Tcl_SetHashValue(ptr,index);
		return &Symbols[index];
	}
}

int SymbolCollection::NSymbols()
{
	return iSymbols;
}
int SymbolCollection::NTerms()
{
	return iTerms;
}

Symbol *SymbolCollection::GetSymbol(int num)
{
	return &Symbols[num];
}

/************************************************************************

	class NonTermIterator:   An iterator over nonterminals in a 
					SymbolCollection
	Data:
			int	current;	// index of current position
			SymbolCollection *c;	// symbolcollection to iterate over

	Methods:
			NonTermIterator(SymbolCollection * );
		void 	NonTermIterator::Reset();
		Symbol *NonTermIterator::Next();

*************************************************************************/

NonTermIterator::NonTermIterator(SymbolCollection * l)
{
	c = l;
	current = -1;
}	

void NonTermIterator::Reset()
{
	current = -1;
}
Symbol *NonTermIterator::Next()
{	
	Symbol * ret;

	if (current == -1) 
		current = 0;

	// new code
	if (current >= c->iSymbols)
	   return 0;
	
	while ((current < c->iSymbols) && (c->Symbols[current].IsTerminal()))
	   current++;

	if (current == c->iSymbols)
	   return 0;

	if (!c->Symbols[current].IsTerminal()) {
		ret = &(c->Symbols[current]);
		current++;
		return ret;
	} else
		return 0;
}

/************************************************************************

	class TermIterator:   An iterator over nonterminals in a 
					SymbolCollection
	Data:
			int	current;	// index of current position
			SymbolCollection *c;	// symbolcollection to iterate over

	Methods:
			TermIterator(SymbolCollection * );
		void 	TermIterator::Reset();
		Symbol *TermIterator::Next();

*************************************************************************/

TermIterator::TermIterator(SymbolCollection * l)
{
	c = l;
	current = -1;
	LastSym = c->LastSym;
}	

void TermIterator::Reset()
{
	current = -1;
}
Symbol *TermIterator::Next()
{	
	Symbol * ret;

	if (current == -1) 
		current = 0;

	// new code
	if (current >= c->iSymbols)
	   return 0;
	
	while ((current < c->iSymbols) && (!c->Symbols[current].IsTerminal()))
	   current++;

	if ((current == c->iSymbols) || current > LastSym)
	   return 0;

	if (c->Symbols[current].IsTerminal()) {
		ret = &(c->Symbols[current]);
		current++;
		return ret;
	} else
		return 0;
}


/************************************************************************

	class Rule:   A grammar rule.  Each rule has:

		Symbol** rhs; 		// the right hand side, an array of Symbols
		double	prob;		// the probability
		int	iNumChildren; 	// number of constituents
		char *  Sem;		// the semantics, currently just a string
		int	RuleNum;	// the rule number
		Symbol 	*lhs;		// the left-hand side, a symbol

			Rule(char *);
			Rule();
			~Rule();
		Rule & operator=(Rule&);
		double	RuleProb()  {return prob;}
		void	RuleCleanup();
		void	print();
		int	IsLexical();
	};

************************************************************************/


Rule::Rule()
{
	RuleNews++;
	rhs = NULL;
	iNumChildren = 0;
	Sem = NULL;
	RuleNum = 0;
	Constraints = NULL;
}

/***************************

	Rule(char *): Constucts a rule by parsing a 
			string of the form:
			"1.0 s -> np vp { sem }"


***************************/

Rule::Rule(char * string)
{
	char tmpbuf[BUFSIZ];
	char *p = string;
	Symbol *tmpsym[MAXRH];
	uint isymindex = 0;
	uint i;
	int mask;
	extern int Constraint_compare(Constraint *,Constraint *);
	RuleNews++;

	Constraints=NULL;
	Sem = NULL;
	RuleNum = 0;

	/* Parse a string of the form "1.0 s -> np vp { sem }" */

	/* First comes the probability */

	if (GetNextToken(tmpbuf,p))
		prob = atof(tmpbuf);
	else
		error("no prob in rule");

	/* next the constitute */

	if (GetNextToken(tmpbuf,p))
		lhs = sc->AddSymbol(tmpbuf);
	else {
	  fprintf(stderr, "Current string: %s\n",string);
	  error ("no lhs in rule");
	}

	if (GetNextToken(tmpbuf,p)) {
		if (strcmp(tmpbuf,"->"))
			error("no arrow in rule");
	} else
		error ("rule with lhs %s ends before arrow",lhs->NoQuotesStringify());


	/* Now read all the right hand side symbols into
	   our temporary buffer */
	mask=0;
	for (;;) {
		if (GetNextToken(tmpbuf,p) == 0)
			break;
		if (tmpbuf[0] == '{')
		    break;
		if (tmpbuf[0] == '(')
		  break;
		tmpsym[isymindex] = sc->AddSymbol(tmpbuf);
		if (tmpsym[isymindex++]->IsNonTerminal()) {
		  // believe me, the isymindex++ was correct here
		  mask |= 1<<isymindex;
		}
	}
	if (tmpbuf[0] == '{') {
		Sem = new char[strlen(tmpbuf) + 1];
		strcpy(Sem,tmpbuf);
	} else if (tmpbuf[0] == '(') {
	  Constraints=new ConstraintList(Constraint_compare,tmpbuf,mask); //tmpbuf
	}
	    
	iNumChildren = isymindex;

	rhs = new Symbol*[iNumChildren];
	for ( i = 0; i < iNumChildren; i++)
		rhs[i] = tmpsym[i];
}

Rule::~Rule()
{
	RuleDeletes++;
	if (rhs) 
	  delete [] rhs;
	if (Sem)
	  delete [] Sem;
	if (Constraints) {
	  Constraints->clobber();
	  delete Constraints;
	}
	Sem = NULL;
	rhs = NULL;
	Constraints = NULL;
}

Rule & Rule::operator=(Rule & r)
{
	uint i;

	lhs = r.lhs;
	prob = r.prob;
	RuleNum = r.RuleNum;
	iNumChildren = r.iNumChildren;
	if (rhs)
	  delete [] rhs;

	if (Sem)  
		delete [] Sem;
	Sem = NULL;
	if (Constraints) {
	  Constraints->clobber();
	  delete Constraints;
	}
	Constraints=NULL;
#ifndef NOPARSE
	if (r.Sem) {
		Sem = new char[strlen(r.Sem) + 1];
		strcpy(Sem,r.Sem);
	}
#endif
	if (r.Constraints) {
	  Constraints=r.Constraints; //suspect, perhaps should copy
	}

	if (!iNumChildren)
		error ("zero kids");
	assert(iNumChildren < 20);
	rhs = new Symbol*[iNumChildren];
	for ( i = 0; i < iNumChildren; i++)
		rhs[i] = r.rhs[i];
	return *this;
}

void Rule::RuleCleanup()
{
	if (rhs)
	  delete [] rhs;
	iNumChildren = 0;
	rhs = NULL;
	if (Sem)
	  delete [] Sem;
	Sem = NULL;
	RuleNum = 0;
	lhs = NULL;
}

int Rule::IsLexical()
{
	if ((iNumChildren) && rhs[0]->IsTerminal())
			return TRUE;
	return FALSE;
}

void Rule::print()
{
  int i;

  cout << prob << " ";
  lhs->print();
  cout << " -> ";
  for ( i = 0; i < iNumChildren; i++) {
    rhs[i]->print();
    cout << " ";
  }
  if (Unify) {
    ConstraintIterator i(Constraints);
    
    cout << " (";
    for (;!i.end();i.next()) {
      cout << "\n\t";
      i.cur()->write();
    }
    cout << ")\n" << flush;
  } else {
    cout << "\t\t" << Sem;
    cout << "\n" << flush;
  }
}



int GetNextToken(char *buffer, char*& p)
{
  int parencount;
  char *pstart=p;

	char *x = buffer;
	/* ignore leading whitespace */
	while (isspace(*p))
		p++;
	if (*p == '\0')
		return FALSE;

	/* copy token until whitespace or NULL */

	if (*p == '{') {
		while ( (*p != '}') && (*p != '\0')) 
	                *buffer++ = *p++;
	        if (*p == '}')
	               *buffer++ = *p++;
	        else
		       error("No close sem bracket");
        } else if (*p == '(') {
	  parencount=1;
	  while((*p != ')' || parencount) && (*p)) {
	    *buffer++ = *p++;
	    if (*p == '(') parencount++;
	    else if (*p == ')') parencount--;
	  }
	  if (*p == ')') 
	    *buffer++=*p++;
	  else
	    {  
	      fprintf(stderr,"Current buffer: %s\n",pstart);
	      error("No close unif paren");
	    }

	  
        } else {
		while ( !isspace(*p) && (*p != '\0')) 
		  *buffer++ = *p++;
	}

	*buffer = '\0';

	if  ((x[0] == '\'') && (*(buffer-1) != '\'')) {
		error("bad token: [%s]\n",x);
	}
	return TRUE;

}


/************************************************************************

	class RuleCollection:  The list of rules which comprises a grammar.

	Data:
		int	iRules;			// The number of rules
		Rule	*Rules[MAXRULES];	// The rules (actually pointers to
							rules.)
	
	Methods:
			RuleCollection();
		void	AddRule(Rule*);
		int 	NRules();
		Rule * 	GetRule(int );
		void	Reset();


************************************************************************/

RuleCollection::RuleCollection()
{
	iRules = 0;
}

void RuleCollection::Reset()
{
	int i;
	for (i=0;i<iRules;i++)
		delete Rules[i];
	iRules = 0;
}

void RuleCollection::AddRule(Rule *r)
{
	Rules[iRules++] = r;
	r->RuleNum = iRules-1;
}

int RuleCollection::NRules()
{
	return iRules;
}

Rule* RuleCollection::GetRule(int num)
{
	return Rules[num];
}


