#ifndef _LIST_C_
#define _LIST_C_


#include <stdlib.h>
#include <iostream.h>

template<class Type>
class DStack {
 private:
  Type *array;
  int top;
  int size;
  void grow() { 
    int new_size = size *2;
    Type *new_array = new Type[new_size];
    for (int i = 0; i < top; i++)
      new_array[i] = array[i];
    delete [] array;
    array = new_array;
    size = new_size;
  }
 public:
  DStack(int init_size = 100) { 
    size = init_size; 
    array = new Type[size]; 
    top = 0;                /* index of next free slot */
  }
  ~DStack() { delete [] array; }
  void push(Type d) { if (top >= size) grow(); array[top++] = d; }
  Type pop() { return array[--top]; }
  Type peek() { return array[top - 1]; }
  int empty() { return top == 0; }
  void clear() { top = 0; }
  void clobber() { while (!empty()) delete pop(); }
  int count() { return top; }
};


#define bool int
#define TRUE 1
#define FALSE 0


/* #define DEBUG */
#ifdef DEBUG
enum Debug_tag { Link_code = 50, Link_unused = 51, 
		   List_code = 52, List_unused = 53 };

extern int debug_counter;
#endif


extern DStack<void *> Link_free_list;

template<class Type>
class Link {
public:
  Type data;
  Link *next;
#ifdef DEBUG
  Debug_tag tag;
  int list_id;
#endif
  Link(Type d, Link *n = NULL) { 
    data = d; 
    next = n; 
#ifdef DEBUG
    tag = Link_code;
#endif
  }
#ifdef DEBUG
  ~Link() { 
    cout << next << endl;
    if (tag != Link_code) {
      char c;
      cerr << "bogus Link::~Link\n";
      cin >> c;
    }
    tag = Link_unused; 
  }
#endif
  void clobberLink() { if (this && data) { delete data; /* next->clobber(); */ } }
  
#ifdef __GNUC__
  /* memory routines follow */  
  void *operator new(size_t n);
  void operator delete(void *s, size_t n);
#endif /* __GNUC__ */

};

// forward template declaration since it's a friend in List
template<class Type>
class Iterator;

template<class Type>
class List {
// protected:
protected:
  Link<Type> *first, *current, *last;
#ifdef DEBUG
  Debug_tag tag;
  int id;
#endif 
public:
  List() { 
    first = current = last = NULL; 
#ifdef DEBUG
    id = ++debug_counter;
    tag = List_code;
#endif
  }
  void clear() {                     /* delete links without deleting data */
    while (first) { 
      current = first->next;
      delete first;
      first = current;
    }
    first = current = last = NULL; 
  }
  ~List() {
    clear();
#ifdef DEBUG
    cout << "List::~List\n";
    if (tag != List_code) {
      char c;
      cerr << "bogus List::~List\n";
      cin >> c;
    }
    tag = Link_unused; 
#endif
  }
  void clobber() { Link<Type> *c, *n = first;
                   while(n) {
		       c = n; n = c->next;  c->clobberLink();
		   }
		   clear();
		 }
  void purge() { first = current = last = NULL; }
  void add_front(Type d) { first = new Link<Type>(d, first);
			   if (last == NULL) last = first; }
  bool empty() { return first == NULL; }
  int end() { return current == NULL; }
  inline void add(Type d);
  inline Type takeFirst();       /* return first item and remove from list */
  inline void remove(Type d);
  inline Type nth(int n);
  inline void reset();
  inline void next();
  inline Type cur();
  int count();
friend class Iterator<Type>;
};

#ifdef __GNUC__
#include <assert.h>

template<class Type>
void *Link<Type>::operator new(size_t n) {
  assert(n==1);
  if (Link_free_list.empty())
    return malloc(sizeof(Link<Type>));
  return Link_free_list.pop();
}

template<class Type>
void Link<Type>::operator delete(void *s, size_t n) {
  assert(n==1);
  if (s != NULL) {
    if (Link_free_list.count() < 10000)
      Link_free_list.push(s);
    else
      free(s);
  }
}
#endif /* __GNUC__ */

template<class Type>
class Iterator {
protected:
  Link<Type> *current;
public:
  void reset(List<Type> *l) { current = l->first; }
  void next() { current = current->next; }
  Type cur() { return current->data; }
  int end() { return current == NULL; }
};


template<class Type>
void List<Type>::add(Type d) {
  if (empty())
    first = last = new Link<Type>(d);
  else {
    last->next = new Link<Type>(d);
    last = last->next;
  }
#ifdef DEBUG
  last->list_id = id;
#endif
}

template<class Type>
Type List<Type>::takeFirst() {
  Link<Type> *temp = first;
  first = temp->next;
  if (empty())
    last = NULL;
  Type d = temp->data;
  temp->data = NULL;
  temp->next = NULL;
  delete temp;
  return d;
}

template<class Type>
void List<Type>::remove(Type d) {
  Link<Type> *prev = NULL, **cur = &first;
  while ((*cur) != NULL) {
    if ((*cur)->data == d) {
      if (last == *cur)
	last = prev;
      Link<Type> *temp = *cur;
      *cur = (*cur)->next;
      temp->next = NULL;
      delete temp;
      return;
    }
    prev = *cur;
    cur = &((*cur)->next);
  }
}

template<class Type>
Type List<Type>::nth(int n) {
  Link<Type> *cur = first;
  for (int i = 1; i < n; i++) {
    if (cur == NULL) {
      cerr << "Error: List does not have " << n << " elements\n";
      return (Type) NULL;
    }
    cur = cur->next;
  }
  return cur->data;
}

template<class Type>
void List<Type>::reset() {
  current = first;
}

template<class Type>
void List<Type>::next() {
  if (end()) {
    char c;
    cerr << "Error: List::next() called at end of list\n";
    cin >> c;
    return;
  }
  current = current->next;
}

template<class Type>
Type List<Type>::cur() {
  if (end()) {
    cerr << "List::cur() called after end of list\n";
    return (Type) 0;
  } else
    return current->data;
}

template<class Type>
int List<Type>::count() {
  int i = 0;
  for (Link<Type> *cur = first; cur != NULL; cur = cur->next) {
    i++;
#ifdef DEBUG
    char c;
    if (cur->tag != Link_code) {
      cerr << "Bogus link: " << i << '\t' << cur << endl;
      cin >> c;
    }
    if (cur->list_id != id) {
      cerr << "Link does not go with list " << cur->list_id
	<< '\t' << id << endl;
      cin >> c;
    }
#endif
  }
  return i;
}


template<class Type>
class SortedList : public List<Type> {
  int (*compare)(Type a, Type b);
 public:
  SortedList(int (*cfn)(Type a, Type b)) { compare = cfn; }
  inline bool add_test(Type d);
  inline void add(Type d);
};

template<class Type>
bool SortedList<Type>::add_test(Type d) {
  Link<Type> **cur = &first;
  int result = 1;           /* be sure to insert d if list is empty */
  while (*cur != NULL && (result = (*compare)((*cur)->data, d)) < 0) {
#ifdef DEBUG
    char c;
    if ((*cur)->tag != Link_code) {
      cerr << "Bogus link: " << '\t' << *cur << endl;
      cin >> c;
    }
    if ((*cur)->list_id != id) {
      cerr << "Link does not go with list " << (*cur)->list_id
	<< '\t' << id << endl;
      cin >> c;
    }
#endif
    cur = &((*cur)->next);
  }
  if (result == 0)   /* same as an item in the list already */
    return FALSE;
//  if (*cur == NULL)      /* this code would maintain last ptr */
//    last = *cur = new Link<Type>(d, *cur);
//  else
  *cur = new Link<Type>(d, *cur);/* warning: SortedList does not maintain last ptr */
#ifdef DEBUG
  (*cur)->list_id = id;
#endif
  return TRUE;
}

template<class Type>
void SortedList<Type>::add(Type d) {
  Link<Type> **cur = &first;
  int result = 1;           /* be sure to insert d if list is empty */
  while (*cur != NULL && (result = (*compare)((*cur)->data, d)) < 0) {
#ifdef DEBUG
    char c;
    if ((*cur)->tag != Link_code) {
      cerr << "Bogus link: " << '\t' << *cur << endl;
      cin >> c;
    }
    if ((*cur)->list_id != id) {
      cerr << "Link does not go with list " << (*cur)->list_id
	<< '\t' << id << endl;
      cin >> c;
    }
#endif
    cur = &((*cur)->next);
  }
//  if (*cur == NULL)      /* this code would maintain last ptr */
//    last = *cur = new Link<Type>(d, *cur);
//  else
  *cur = new Link<Type>(d, *cur);/* warning: SortedList does not maintain last ptr */
#ifdef DEBUG
  (*cur)->list_id = id;
#endif
}

#endif /* _LIST_C_ */
