/*
**
**  ff (fast forward): An optimized bunched-mode 3-layer forward-propagation MLP program.
**  The following are hard-coded assumptions:
**     o softmax output units
**     o bias units on input and hidden layer fixed as -1 valued.
**
**  $Header: /u/drspeech/src/ffwd/RCS/ff.c,v 1.4 1999/11/09 05:07:48 dpwe Exp $
**  Jeff Bilmes <bilmes@icsi.berkeley.edu>
**
*/


#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <ctype.h>

#include "ff.h"

extern double drand48();
extern unsigned short* seed48();

void *
lmalloc(int bytes)
{
  void *tmp;
  if ((tmp = (void*)malloc(bytes)) == NULL) {
    fprintf(stderr,"Can't allocate memory\n");
    exit(-1);
  }
  return tmp;
}


void
init_matrix(float *mat,int M,int N)
{   /* fill matrix with random values */
  float *endp = &mat[M*N];
  while (mat != endp) {
    *mat++ = drand48();
  }

}


void
bias_matrix(float *mat,int M,int N, float bval)
{  /* setup the 'bias' values in a matrix to the defined constant value.
      Bias units are are at the end of each row i.e. every N values, 
      where N is the number of columns. */
  float *p = &mat[N]-1;
  while (M--) {
    *p = bval;
    p += N;
  }
}


void
read_matrix(char *fname,float*mat,int M,int N,int add_bias, float bval)
{   /* Read an ascii file containing an MxN matrix as M rows of N cols.
       If <add_bias>, add the default bias val at the end of each 
       row, writing an Mx(N+1) output matrix. */
  int i,j;
  FILE *file;
  
  if ((file = fopen(fname,"r")) == NULL) {
    fprintf(stderr,"Can't open %s\n",fname);
    exit(-1);
  }
  for (i=0;i<M;i++) {
    for (j=0;j<N;j++) {
      fscanf(file,"%f",mat++);
    }
    if (add_bias)
      *mat++ = bval;
  }
  fclose(file);

}


void
write_matrix(char *fname,float*mat,int M,int N)
{   /* Write an MxN matrix to file as ascii, with line breaks between usage
       each row. */
  int i,j;
  FILE *file;

  if ((file = fopen(fname,"w")) == NULL) {
    fprintf(stderr,"Can't open %s\n",fname);
    exit(-1);
  }

  for (i=0;i<M;i++) {
    for (j=0;j<(N-1);j++) {
      fprintf(file,"%f ",*mat++);
    }
    fprintf(file,"%f\n",*mat++);
  }  
  fclose(file);
}

#ifdef MAIN

int ninp;
int nhid;
int nout;
int npat;
int bsize;
char *ifile;
char *ofile;
float bval = DEFBIASUNITVAL;
char *ih_file = "/dev/null";
char *ho_file = "/dev/null";

void usage(char *name)
{
  fprintf(stderr,"usage: %s [OPTIONS]\n",name);
  fprintf(stderr,"where [OPTIONS] include:\n");  
  fprintf(stderr,"  -ninp i                 : number of input units.\n");
  fprintf(stderr,"  -nhid h                 : number of hidden units.\n");
  fprintf(stderr,"  -nout o                 : number of output units.\n");
  fprintf(stderr,"  -ifile file             : input pattern file (must be pXi in size)\n");
  fprintf(stderr,"  -npat p                 : number of input patterns.\n");
  fprintf(stderr,"  -bsize b                : bunch size\n");
  fprintf(stderr,"  -ofile file             : output file\n");
  fprintf(stderr,"  -ih_file file           : file containing input->hidden weights\n");
  fprintf(stderr,"  -ho_file file           : file containing hidden->output weights\n");
}

void
parse_args(int argc,char*argv[])
{
  int i;
  if (argc == 1) {
    usage(argv[0]);
    exit(-1);
  }
  for (i=1;i<argc;i++) {
    if (argv[i][0] == '-') {
      if (!strcmp(&(argv[i][1]),"ninp")) {
	ninp = atoi(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"nhid")) {
	nhid = atoi(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"nout")) {
	nout = atoi(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"npat")) {
	npat = atoi(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"bsize")) {
	bsize = atoi(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"bval")) {
	bval = atof(argv[++i]);
      } else if (!strcmp(&(argv[i][1]),"ifile")) {
	ifile = argv[++i];
      } else if (!strcmp(&(argv[i][1]),"ofile")) {
	ofile = argv[++i];
      } else if (!strcmp(&(argv[i][1]),"ih_file")) {
	ih_file = argv[++i];
      } else if (!strcmp(&(argv[i][1]),"ho_file")) {
	ho_file = argv[++i];
      } else  {
        fprintf(stderr,"Error: Unknown or invalid argument (%s)\n",argv[i]);
	exit(-1);
      }
    }
  }
  if (ninp <= 0 || nhid <= 0 || nout <= 0 || npat <= 0 ||
      bsize <= 0 || ifile == NULL || ofile == NULL ||
      ih_file == NULL || ho_file == NULL) {
    fprintf(stderr,"Error with arguments\n");
    usage(argv[0]);
    exit(-1);
  }
}

int
main(int argc,char *argv[]) 
{
  /* weights */
  float *who;  /* hidden to output weights */
  float *wih;  /* hidden to output weights */
  
  /* patterns */
  float *ipats,*ipatsp; /* input patterns */

  /* scratch & misc. */
  float *hpats; /* hidden layer */
  float *opats; /* output layer */

  int iters, rem, i, ijump;
  
  parse_args(argc,argv);

  who = lmalloc(nout*(nhid+1)*sizeof(float));
  wih = lmalloc(nhid*(ninp+1)*sizeof(float));
  ipats = lmalloc(npat*(ninp+1)*sizeof(float));
  hpats = lmalloc(bsize*(nhid+1)*sizeof(float));  
  opats = lmalloc(bsize*nout*sizeof(float));  

  read_matrix(ifile,ipats,npat,ninp,TRUE,bval);
  read_matrix(ih_file,wih,nhid,ninp+1,FALSE,0);
  read_matrix(ho_file,who,nout,nhid+1,FALSE,0);

  bias_matrix(hpats,bsize,nhid+1,bval);

  iters = npat/bsize;
  rem = npat % bsize;
  ijump = bsize*(ninp+1);

  ipatsp = ipats;

  for (i=0;i<iters;i++) {
    bunched_mlp2_forwprop_sm(ninp,nhid,nout,bsize,
			     who,wih,
			     ipatsp,
			     hpats,opats);
    write_matrix(ofile,opats,bsize,nout);
    ipatsp += ijump;
  }
  if (rem != 0) {
    bunched_mlp2_forwprop_sm(ninp,nhid,nout,rem,
			     who,wih,
			     ipatsp,
			     hpats,opats);
    write_matrix(ofile,opats,rem,nout);
  }
  
  exit(0);
}
#endif /* MAIN */
