//
// Est.H
//
// Definitions of estimator objects
// Get fed a series of values and estimate their mean and sd
// or lowpass smooth them 
// or something.
//
// dpwe@media.mit.edu 1996feb14
//

#ifndef _EST_H
#define _EST_H

#include <iostream.h>
#include <math.h>
#include <assert.h>

class Estimator {
public:
    Estimator(void);
    virtual ~Estimator(void);
    // Default copy constructor etc will be ok - no allocated state

    virtual void Reset(double mean = 0, double sd = 0, int n = 0);
    // return to default condition

    virtual double NewVal(double x);	// Add a new value; return the smoothed vsn

    virtual void Print(ostream& os=cout, char *tag=0) const;

    // Access the current states of the estimates
    virtual double Mean(void) const;
    virtual double SD(void)  const;
    int    N(void)   const { return n; }

//protected:
    // rest of state is hidden
    double mean;
    double meansq;
    int n;
};

class RecursiveEst : public Estimator {
    // One-pole smoothing to calculate mean & meansq
public:
    RecursiveEst(double timeConst=1) { tc = timeConst; };
    ~RecursiveEst(void)            { }

    double NewVal(double x)        { assert(tc>=1.0);
	                             mean += (1/tc)*(x-mean);
                                     meansq += (1/tc)*(x*x-meansq);
				     ++n; return mean; }
    double Mean(void) const        { return mean; }
    double SD(void) const          { return sqrt(meansq - mean*mean); }

    double tc;
};

class BecomesRecEst : public RecursiveEst {
    // Eventually acts like a recursive est, but for the first few 
    // timesteps, forms an unweighted average of the inputs
public:
    BecomesRecEst(double timeConst=1) : RecursiveEst(timeConst)    { }
    ~BecomesRecEst(void) { }

    double NewVal(double x) { if(n >= tc) {
	                          RecursiveEst::NewVal(x);
                              } else {
				  ++n;
				  mean = ( (n-1)*mean + x )/n;
				  meansq = ( (n-1)*meansq + x*x )/n;
                              } return mean; }
};

class RapidRecEst : public RecursiveEst {
    // Special fast-startup recursive est actually runs *two* recursive 
    // estimators, one with a much shorter time constant, then cross-fades
    // between them over the first few steps so that initially unsmoothed 
    // data becomes more smooth
public:
    RapidRecEst(double timeConst=1, double fasterfact=10, double fadefact=2)
	: RecursiveEst(timeConst)
        { fastRE.tc = tc/fasterfact; tFade = tc*fadefact; }

    void SetTCs(double timeConst, double fasterFactor, double fadeFactor)
        { tc = timeConst; fastRE.tc = tc/fasterFactor; tFade = tc*fadeFactor; }

    double NewVal(double x) {  RecursiveEst::NewVal(x);
                               fastRE.NewVal(x);
			       return Mean(); }
    double Mean(void) const {  double k = exp(-n/tFade);
                               return k*fastRE.Mean() 
				   + (1-k)*RecursiveEst::Mean(); }
    double SD(void) const   {  double k = exp(-n/tFade);
                               return k*fastRE.SD() 
				   + (1-k)*RecursiveEst::SD(); }
    void Reset(double _mean = 0, double sd = 0, int _n = 0)
        { Estimator::Reset(_mean, sd, _n); fastRE.Reset(_mean, sd, _n); }
// private:
    RecursiveEst fastRE;
    double tFade;
};

inline ostream& operator << (ostream& os, const Estimator& es)
    { es.Print(os); return os; }

// Estimators for floatvecs

#include "floatArray.H"	// definitions of floatVecs

class fvEstimator {
public:
    fvEstimator(int _n=0)
        { Resize(_n); Reset(); }
    fvEstimator(const fvEstimator& fve)
        { operator= (fve); }
    virtual ~fvEstimator(void)
        { }
    fvEstimator& operator= (const fvEstimator& fve)
        { if(this != &fve)  {
	    Resize(fve.Size()); n = fve.n;
	    mean = fve.mean; meansq = fve.meansq;
	  } return *this; }

    virtual void Resize(int _n)
        { mean.Resize(_n); meansq.Resize(_n); }
    virtual void Reset(floatRef* _pmean = 0, floatRef* _psd = 0, int _n = 0)
        { if(_pmean) { assert(_pmean->Size()==mean.Size()); mean = *_pmean; } 
 	  else { mean.Clear(); }
          if(_psd) { assert(_psd->Size()==mean.Size()); 
                     meansq = mean*mean+(*_psd * *_psd); }
 	  else { meansq.Clear(); }
          n = _n; }

    virtual floatVec NewVal(floatRef& x)
        { ++n; mean = ((n-1)*mean + x)/n;
	  meansq = ((n-1)*meansq + x*x)/n;
	  return mean; }

    virtual void Print(ostream& os=cout, char *tag=0) const
    { os << "fvEstimator" << (void*)this;
      if(tag && strlen(tag)) os << " '" << tag << "'";
      os << ":" << endl;
      os << "  n=" << n << " mean=" << mean << " meansq=" << meansq << " mu=" 
	 << Mean() << " sd=" << SD() << " N=" << N() << endl; }

    // Access the current states of the estimates
    virtual floatVec Mean(void) const { return mean; }
    virtual floatVec SD(void)   const { floatVec tmp = meansq - mean*mean;
                                        return Sqrt(tmp); }
    int    N(void)    const { return n; }
    int    Size(void) const { return mean.Size(); }

//protected:
    // rest of state is hidden
    floatVec mean;
    floatVec meansq;
    int n;
};

class fvRecursiveEst : public fvEstimator {
    // One-pole smoothing to calculate mean & meansq
public:
    fvRecursiveEst(int _n=0, double timeConst=0) : fvEstimator(_n)
        { tc = timeConst; };
    ~fvRecursiveEst(void)            
        { }
    fvRecursiveEst(const fvRecursiveEst& fve)
        { operator= (fve); }
    fvRecursiveEst& operator= (const fvRecursiveEst& fve)
	{ if(this != &fve) {
	    fvEstimator::operator= (fve); tc = fve.tc;
	} return *this; }

    floatVec NewVal(floatRef& x)   { assert(tc>=1.0);
	                             mean += (1/tc)*(x-mean);
                                     meansq += (1/tc)*(x*x-meansq);
				     ++n; return mean; }
    double tc;
};

class fvBecomesRecEst : public fvRecursiveEst {
    // Eventually acts like a recursive est, but for the first few 
    // timesteps, forms an unweighted average of the inputs
public:
    fvBecomesRecEst(int _n=0, double timeConst=0) 
	: fvRecursiveEst(_n, timeConst)
        { }
    ~fvBecomesRecEst(void)
        { }
    fvBecomesRecEst(const fvBecomesRecEst& fve)
        { operator= (fve); }
    fvBecomesRecEst& operator= (const fvBecomesRecEst& fve)
	{ if(this != &fve) {
	    fvRecursiveEst::operator= (fve);
	} return *this; }
    fvBecomesRecEst& operator= (const fvRecursiveEst& fve)
	{ if((fvRecursiveEst*)this != &fve) {
	    fvRecursiveEst::operator= (fve);
	} return *this; }

    floatVec NewVal(floatRef& x) { if(n >= tc) {
                                     fvRecursiveEst::NewVal(x);
                                   } else {
	 			     fvEstimator::NewVal(x);
		 		   } return mean; }
};

class fvRapidRecEst : public fvRecursiveEst {
    // Special fast-startup recursive est actually runs *two* recursive 
    // estimators, one with a much shorter time constant, then cross-fades
    // between them over the first few steps so that initially unsmoothed 
    // data becomes more smooth
public:
    fvRapidRecEst(int _n=1, double timeConst=0, double fasterfact=10, 
		double fadefact=2)
	: fvRecursiveEst(_n, timeConst)
        { fastRE.Resize(_n);
	  fastRE.tc = tc/fasterfact; tFade = tc*fadefact; }
    ~fvRapidRecEst(void)
        { }
    fvRapidRecEst(const fvRapidRecEst& fve)
        { operator= (fve); }
    fvRapidRecEst& operator = (const fvRapidRecEst& fve)
        { if(this!=&fve)  {
	    fvRecursiveEst::operator= (fve); fastRE = fve.fastRE;
	  }  return *this; }

    void SetTCs(double timeConst, double fasterFactor, double fadeFactor)
        { tc = timeConst; fastRE.tc = tc/fasterFactor; tFade = tc*fadeFactor; }

    floatVec NewVal(floatRef& x) {  fvRecursiveEst::NewVal(x);
                                    fastRE.NewVal(x);
			            return Mean(); }
    floatVec Mean(void) const    {  double k = exp(-n/tFade);
                                    return k*fastRE.Mean() 
				             + (1-k)*fvRecursiveEst::Mean(); }
    floatVec SD(void) const      {  double k = exp(-n/tFade);
                                    return k*fastRE.SD() 
				            + (1-k)*fvRecursiveEst::SD(); }
    void Reset(floatRef *pmean = 0, floatRef* psd = 0, int _n = 0)
        { fvRecursiveEst::Reset(pmean, psd, _n); 
	  fastRE.Reset(pmean, psd, _n); }
    void Resize(int _n)
        { fvRecursiveEst::Resize(_n); fastRE.Resize(_n); }
// private:
    fvRecursiveEst fastRE;
    double tFade;
};

inline ostream& operator << (ostream& os, const fvEstimator& es)
    { es.Print(os); return os; }

#endif /* _EST_H */
