//
// ProbVec2.C
//
// A floatVec with three components - expectation and +/- sd - for each element
//
// dpwe@media.mit.edu 1995nov24
//

#include "ProbVec2.H"
#ifdef HAVE_LIBDAT
#include "myDAT.h"
#endif /* HAVE_LIBDAT */

#ifndef M_PI
#define M_PI 3.1415926535
#endif /* M_PI */

// This flag affects the return of AvgLogLikBits1dB - 
// if it's set, the averaging is done over *all* chans;
// otherwise, it only averages over nonzero chans (default behavior).
// By setting it, elements that do the same work in fewer channels 
// are favored.
int probRef::avgAllChans = 0;

double probRef::Mahal(const floatRef &fr)
{   /* return the mahalanobis distance between the expected vector & this.
       Only dimensions with nonzero SD are included. */
    double d; 
    floatVec diff = fr - *this;
    floatVec pdiff = ::Floor(diff,0); 
    floatVec ndiff = ::Ceil(diff,0);
    d = ::SumSq((pdiff*(psd!=0)) / (psd+(psd==0))) 
	+ ::SumSq((ndiff*(nsd!=0)) / (nsd+(nsd==0)));
    return sqrt(d); 
}
	
double probRef::Likelihood(const floatRef &fr)
{   /* score the likelihood that the passed floatref came from this pv */
    double d = 0, s = 0; 
    int i, n = 0; 
    for(i=0;i<len;++i) { 
	if(fr.Val(i) >= Val(i) && psd.El(i)>0) {
	    s = s+log(psd.El(i));
	    ++n;
	} else if(fr.Val(i) < Val(i) && nsd.El(i)>0) {
	    s = s+log(nsd.El(i));
	    ++n;  
	} 
    } 
    if(n>0) 
	d = exp(-0.5*pow(Mahal(fr),2.0) - s - (0.5*n)*log(2*M_PI));
//	d = exp(-0.5*pow(Mahal(fr),2.0))/(exp(s)*pow(2*M_PI,((double)n)/2));
    return d; 
}

double probRef::AvgLogLikBits(const floatRef &fr)
{   /* different likelihood fn - returns -logbase2 (i.e. bits) of *average* 
       gaussian probval over usable points (nonzero sd in given sense) */
    int i, n=0;
    floatVec diff = fr - *this;
    double logp = 0, logs = 0, d, _psd, _nsd;

    // Note that diff.El==0 will match either clause - just in case one 
    // of the sd polarities has been left at zero (which might otherwise 
    // cause that bin to be skipped when it wasn't meant to be).
    for(i = 0; i<diff.Size(); ++i)  {
	d = diff.El(i);
	_psd = this->psd.El(i);
	_nsd = this->nsd.El(i);
	if(_psd > 0 && d >= 0)  {
	    ++n;
	    logp += d*d/(_psd*_psd);
	    logs += log(_psd);
	} else if (_nsd > 0 && d <= 0)  {
	    ++n;
	    logp += d*d/(_nsd*_nsd);
	    logs += log(_nsd);
	}
    }
    assert(n>0);	// else in a funny state
    // We're taking the average normalized squared error per active channel
    logp /= n;
    // Now we have avg( ((x-u)/s)^2 )
    // Gaussian prob would be 1/(sqrt(2pi)s)*exp(-0.5 ((x-u)/s)^2 )
    // so logbaseeprob would be -0.5 ((x-u)/s)^2 - log s - 0.5 log 2pi
    // Since we're averaging the ((x-u)/s)^2, we should also average log s
    logs /= n;
    // We want logbase2 (to be in bits) so scale by ln(2)
    // (logbaseb(A) = logbasec(A)/logbasec(b)
    double rslt = -(-0.5 * logp - logs - 0.5 * log(2.0*M_PI))/log(2.0);
    // hmm, problem is that exp(-0.5logp - logs -0.5log(2pi)) is a 
    // prob density, not a true prob.  When s is quite small, the
    // prob density can be much bigger than one, giving negative bits. 
    // Not the simple meaurement I had hoped for...
    return rslt;
}

double probRef::AvgLogLikBits1dB(const floatRef &fr, 
				 floatRef* pbrkdown /* =0 */)
{   /* Slightly modified version of probRef::AvgLogLikBits, weights 
       each probability density as the *actual probability* that 
       the observed amplitude falls into a *1 dB* slot around its 
       actual value.
       If pbkrdown is nonnull, store the contribution of each cell 
       there. */
    int i, n=0;
    floatVec diff = fr - *this;
    double logp = 0, logs = 0, log1 = 0, d, val, onedb;
#define LOG2    0.69314718
#define LOG2PI  1.8378771
#define ONEDB   .12201845430196		// 10^(1/20) - 1
#define ONEONLOG2 1.44269504088896	// 1/ln(2) - log to bits
    // Note that diff.El==0 will match either clause - just in case one 
    // of the sd polarities has been left at zero (which might otherwise 
    // cause that bin to be skipped when it wasn't meant to be).
    floatRef psdr = PSD();
    floatRef nsdr = NSD();

    floatVec maxsd = ::Max(psdr, nsdr);
    floatVec pmask = diff>0;
    floatVec nmask = diff<0;
    floatVec exmask = (diff==0) * (fr!=0);	// exact matches here
    floatVec zmask = (diff==0) * (fr==0);	// simply absent channels
    floatVec nzmask = (zmask==0);
    floatVec scaledDiffs = pmask*SafeDiv(diff,psdr,0)
	                   +nmask*SafeDiv(diff,nsdr,0);
    floatVec badchans = pmask*(psdr==0)+nmask*(nsdr==0)+exmask*(maxsd==0);
    floatVec logPvec = 0.5 * (scaledDiffs^2);
    floatVec logSvec = ::Log(pmask*psdr + nmask*nsdr 
			     + exmask*maxsd + zmask + badchans) 
	                + 99.99*badchans;
    floatVec log1dBvec = ::Log(::Max(ONEDB*::Max(fr, diff*(pmask-nmask)), 
				     zmask));

    // trap situation where nonzero diff has zero sd i.e. inf logSvec
    if(::Sum(badchans)!=0) {
	// We have some channels with NZ diffce but Z sd
	cerr << "***AvBits1dB: " << ::Sum(badchans) << " chans were bad" << endl;
    }

    floatVec net = ONEONLOG2*(logPvec + logSvec - log1dBvec 
			      + 0.5*LOG2PI*nzmask);

    // Don't let any channel exceed the number of bits it would take 
    // to encode this signal at some small abs level
    double floorval = 0.000001;	// -100 dB, say
    floatVec bitsceil = ONEONLOG2*(::Log(::Max(fr,floorval)/floorval));
    net = ::Min(net, bitsceil);
    
    double nch  = ::Sum(nzmask);
    assert(nch>0);	// else in a funny state
    double bits;
    if(avgAllChans) {
	bits = ::Sum(net)/net.Size();
    } else {
	bits = ::Sum(net)/nch;
    }

    if(pbrkdown) {
	*pbrkdown = net;
    }

    return bits;
}

void probVec::CloneVec(const probRef& pr)
{ 
    delete [] data; 
    delete [] psd.Data(); 
    delete [] nsd.Data();
    len = pr.Size();
    _Init(len, new float[len], new float[len], new float[len]);
    floatRef::Copy(pr);
    psd.Copy(pr.PSD());
    nsd.Copy(pr.NSD());
}

// probRef dyadic ops

probVec operator + (const probRef& a, const probRef& b)
{ probVec f(a); f.Add(b); return f; }
probVec operator + (const probRef& a, double d)
{ probVec f(a); f.Offset(d); return f; }
probVec operator + (double d, const probRef& a)
{ probVec f(a); f.Offset(d); return f; }

probVec operator - (const probRef& a, const probRef& b)
{ probVec f(a); f.Sub(b); return f; }
probVec operator - (const probRef& a, double d)
{ probVec f(a); f.Offset(-d); return f; }
probVec operator - (double d, const probRef& a)
{ probVec f(a); f.Offset(-d); return f; }

probVec operator * (const probRef& a, const floatRef& b)
{ probVec f(a); f.Mul(b); return f; }
probVec operator * (const floatRef& b, const probRef& a)
{ probVec f(a); f.Mul(b); return f; }
probVec operator * (const probRef& a, double d)
{ probVec f(a); f.Scale(d); return f; }
probVec operator * (double d, const probRef& a)
{ probVec f(a); f.Scale(d); return f; }

probVec operator / (const probRef& a, const floatRef& b)
{ probVec f(a); f.Div(b); return f; }
// probVec operator / (const floatRef& a, const probRef& b)
// { probVec f(a); f.Div(b); return f; }
probVec operator / (const probRef& a, double d)
{ probVec f(a); f.Scale(d); return f; }
// probVec operator / (double d, const probRef& a)
// { probVec f(a); f = d/a; return f; }


probArray& probArray::operator = (const probArray &pa)
{ 
    probVec::CloneVec(pa); 
    rows = pa.rows; 
    cols = pa.cols;
    adata = data; 
    pfa = 0;  
    psdr._Init(rows, cols, psd.Data()); 
    nsdr._Init(rows, cols, nsd.Data()); 
    return *this;
}

int  probArray::Load(const char *filename, int dataset/*=0*/)
{   /* Load the array from a given dataset of a DAT file.
       Return the number of elements loaded, or -1 if bad read. */
    int ds, ch, rw, cl; double sr; char fm; char c = 'f', rc = -1;

#ifdef HAVE_LIBDAT    
    if(DATInfo((char *)filename, &ds, &ch, &rw, &cl, &sr, &fm)==0)  
#endif /* HAVE_LIBDAT */
    {
	return -1;  }
    if(dataset >= ds)  {  return -1; }
    /* cols are the fastest rotating dimension in a DAT file, 
       so make an implicit transposition on reading into the ucharArray */
    Resize(rw, cl);
#ifdef HAVE_LIBDAT
    rc = DATRead((char *)filename, dataset, 0, -1, -1, 
		 (void **)&data, c, &rw, &cl, &sr); 
#endif /* HAVE_LIBDAT */
    Transpose();
    return rc;
}

void probArray::Resize(int nr, int nc)
{ 
    float *ndata = new float[nr*nc];
    float *npsddata = new float[nr*nc];
    float *nnsddata = new float[nr*nc];
    float *a, *b;
    int i, j, dr = min(nr, rows), dc = min(nc, cols);
//cout << "Resize of " << (void *)this << " from " << rows << "x" << cols 
//    << " to " << nr << "x" << nc << endl;
    for(i=0;i<dr;++i)  {  
	for(j=0;j<dc;++j)  {
	    a = ndata+i*nc+j;
	    b = data+i*cols+j;
	    // *(ndata+i*nc+j) = *(data+i*cols+j);
	    *a = *b;
	    *(npsddata+i*nc+j) = *(psd.Data()+i*cols+j);
	    *(nnsddata+i*nc+j) = *(nsd.Data()+i*cols+j);
	}
	for(j=dc; j<nc; ++j) {
	    *(ndata+i*nc+j) = 0;
	    *(npsddata+i*nc+j) = 0;
	    *(nnsddata+i*nc+j) = 0;
	}
    }
    for(i=dr; i<nr; ++i)  {
	for(j=0; j<nc; ++j) {
	    *(ndata+i*nc+j) = 0;
	    *(npsddata+i*nc+j) = 0;
	    *(nnsddata+i*nc+j) = 0;
	}
    }
    delete [] data;
    delete [] psd.Data();
    delete [] nsd.Data();
    adata = data = ndata;
    len  = nr*nc;
    psd.SetData(npsddata);
    psd.SetLen(len);
    nsd.SetData(nnsddata);
    nsd.SetLen(len);
    rows = nr;
    cols = nc;
    psdr._Init(rows, cols, npsddata);
    nsdr._Init(rows, cols, nnsddata);
}

void probArray::Transpose(void)
{ 
    int nr = cols; int nc = rows;
    float *ndata = new float[nr*nc];
    float *npsddata = new float[nr*nc];
    float *nnsddata = new float[nr*nc];
    int i, j;
    for(i=0;i<nr;++i)  {  
	for(j=0;j<nc;++j)  {
	    *(ndata+i*nc+j) = *(data+j*cols+i); 
	    *(npsddata+i*nc+j) = *(psd.Data()+j*cols+i); 
	    *(nnsddata+i*nc+j) = *(nsd.Data()+j*cols+i); 
	} 
    }
    delete [] data;
    delete [] psd.Data();
    delete [] nsd.Data();
    adata = data = ndata;
    len = nr*nc;
    psd.SetData(npsddata);
    psd.SetLen(len);
    nsd.SetData(nnsddata);
    nsd.SetLen(len);
    rows = nr;
    cols = nc;
    psdr._Init(rows, cols, npsddata);
    nsdr._Init(rows, cols, nnsddata);
}

void probArray::Copy(const probArray& pa)
{
    int i,j;
    for(i=0; i<min(rows, pa.rows); ++i) {
	for(j=0; j<min(cols, pa.cols); ++j) { 
	    El(i,j) = pa.Val(i,j); 
	    psdr.El(i,j) = (pa.PSD()).Val(i,j); 
	    nsdr.El(i,j) = (pa.NSD()).Val(i,j); 
	} 
    } 
}

#ifdef MAIN

#include <iostream.h>

#define ___quote(ARG)      #ARG
#define ___QUOTE(ARG)	___quote(ARG)
#define TEST(num,clause)					\
        if(!(clause))	{ cerr << "FAIL Test " << num << ": "	\
			      << ___QUOTE(clause) << endl;  }

int Feq(float a, float b)
{   // loose version of operator==(float,float) - allow up to 0.001 diff
    if(fabs(a-b)<0.001)	return 1;
    else		return 0;
}

void testProbVec(void)
{
    // Just test the probability functions
    probVec v1;
    TEST("1.1", v1.Size()==0);
    v1.Resize(4);
    TEST("1.2", v1.Size()==4);
    TEST("1.3", v1[0]==0 && v1[1]==0 && v1[2]==0 && v1[3]==0);
    TEST("1.4", v1.PSDv()[0]==0 && v1.PSDv()[1]==0 && v1.PSDv()[2]==0 && v1.PSDv()[3]==0);
    TEST("1.4b", v1.NSDv()[0]==0 && v1.NSDv()[1]==0 && v1.NSDv()[2]==0 && v1.NSDv()[3]==0);
    // Test single value SD
    floatVec f(4);
    f.Fill(2);
    v1.Fill(1);
    v1.PSDv()[2] = 1;
    v1.NSDv()[2] = 2;
    TEST("1.5", Feq(v1.Likelihood(f), exp(-.5)/sqrt(2*M_PI)));
    // Test mahalanobis return too
    TEST("1.6", Feq(v1.Mahal(f), 1));
    // Test the -ve SD now
    f.Fill(0);
    TEST("1.5b", Feq(v1.Likelihood(f), exp(-.125)/(2*sqrt(2*M_PI))));
  cout << "v1.Lik(f)=" << v1.Likelihood(f) << ", exp..=" << exp(-.125)/(2*sqrt(2*M_PI)) << endl;
    // Test mahalanobis return too
    TEST("1.6b", Feq(v1.Mahal(f), 0.5));

    // Test with more vals
    f.Fill(2);
    v1.PSDv()[1] = 2;
    v1.PSDv()[3] = 3;
//    cout << "v1.Lik(f)=" << v1.Likelihood(f) << ", exp..=" << exp(-.5*(1+0.25+0.1111111111111))/(6*sqrt(pow(2*M_PI,3))) << ", mahal=" << v1.Mahal(f) << endl;
    TEST("1.7", Feq(v1.Likelihood(f), exp(-.5*(1+0.25+0.1111111111111))/(6*sqrt(pow(2*M_PI,3)))));
    TEST("1.8", Feq(v1.Mahal(f), sqrt(1+0.25+0.111111111111111)));
    // Adding
    probVec v2(4);
    v2.PSDv().Fill(2);
    v2.NSDv().Fill(1);
    v2[1] = 1;
    v1.Add(v2);
    TEST("1.9", v1[0] == 1 && v1[1] == 2 && v1.PSDv()[0] == 2 
	 && Feq(v1.PSDv()[1], 2*sqrt(2)) && Feq(v1.NSDv()[0], 1) 
	 && Feq(v1.NSDv()[2], sqrt(1+4)));
    // Adding a plain floatVec
    f[0] = -1;
    f[1] = 5;
    // v1.Add(f);
    // Initialization from a floatVec
    probVec v3(f);
    TEST("1.10", v3.Size()==4 && v3[0]==-1 && v3[1]==5 && v3.PSDv()[0]==0 
	 && v3.NSDv()[3]==0);
    // addition of probVecs
    v1.Add(v3);
    TEST("1.11", v1[0] == 0 && v1[1] == 7 && v1.PSDv()[0] == 2 
	 && Feq(v1.PSDv()[1], 2*sqrt(2)));
    // Cast to floatRefs
    floatRef *fr = (floatRef *)&v1;
    TEST("1.12", fr->Size()==4 && fr->Data()==v1.Data());

    float ex[46] = {0,0,0,0,0,0,0,0,0,0,0.0218362,0.0170904,0,0,0,0,0.00168066,0,0,0,0,0,0,0,0,0.00402637,0.00358045,0.0450559,0.124829,0.0830829,0.0796432,0.0828643,0.0704569,0.0436949,0.031316,0.019446,0.0113546,0.00727153,0.00520894,0.0040408,0.00316612,0.00257689,0.00211973,0.00173356,0.00139387,0.00105713};
    float psd[46] = {0,0,0,0,0,0,0,0,0,0,0.0112261,0.00878628,0,0,0,0,0.000864041,0,0,0,0,0,0,0,0,0.00206999,0.00184074,0.0231635,0.0641756,0.0427135,0.0409451,0.0426011,0.0362224,0.0224639,0.0160998,0.00999735,0.00583747,0.00373835,0.00267795,0.0020774,0.00162773,0.0013248,0.00108977,0.000891236,0.000716598,0.000543477};
    float nsd[46] = {0,0,0,0,0,0,0,0,0,0,0.0218362,0.0170904,0,0,0,0,0.00168066,0,0,0,0,0,0,0,0,0.00402637,0.00358045,0.0450559,0.124829,0.0830829,0.0796432,0.0828643,0.0704569,0.0436949,0.031316,0.019446,0.0113546,0.00727153,0.00520894,0.0040408,0.00316612,0.00257689,0.00211973,0.00173356,0.00139387,0.00105713};
    float err[46] = {0,0,0,0,0,0,0,0,0,0,0.0179897,0.0217889,0,0,0,0,-0.00168066,0,0,0,0,0,0,0,0,0.016453,0.0976814,0.100878,0.030561,0.0720192,0.132244,0.188669,0.142178,0.083338,0.0511092,0.0321552,0.020659,0.0143419,0.0107651,0.00819771,0.0066677,0.00539295,0.00441643,0.00363075,0.0028905,0.00217708};
    probRef pr1(46,ex,psd,nsd,1);
    floatRef f1(46,err,1);

    cout << "final test: Mahal=" << pr1.Mahal(f1) << " (52.1466), Lik=" << pr1.Likelihood(f1) << " (0), Bits=" << pr1.AvgLogLikBits(f1) << "(75.6963)"<< endl;
    cout << "            1dB Bits=" << pr1.AvgLogLikBits1dB(f1) << "(84.2919)"<< endl;

}

void testArray(void)
{   //  test the probArray class - just the Array parts
    // Default constructor
    probArray a1;
    TEST("2.1", a1.Rows()==0 && a1.Cols()==0);
    // Regular constructor
    probArray a2(2,3);
    TEST("2.2", a2.Rows()==2 && a2.Cols()==3);
    // El() and Val()
    a2.El(0,0) = 0; a2.El(1,0)=1; a2.El(0,1)=10; a2.El(1,1)=11;
    a2.El(0,2) = 20; a2.El(1,2)=21;
    TEST("2.3", a2.Val(0,0)==0 && a2.Val(1,0)==1 && a2.Val(0,1)==10 \
	 && a2.Val(1,1)==11 && a2.Val(0,2)==20 && a2.Val(1,2)==21);
    // El for access too
    TEST("2.4", a2.El(0,0)==0 && a2.El(1,0)==1 && a2.El(0,1)==10 \
	 && a2.El(1,1)==11 && a2.El(0,2)==20 && a2.El(1,2)==21);
    // Index accessing - indices same order - [row][col]
    TEST("2.5", a2[0][0]==0 && a2[1][0]==1 && a2[0][1]==10 \
	 && a2[1][1]==11 && a2[0][2]==20 && a2[1][2]==21);
    // Copy constructor
    probArray a3(a2);
    TEST("2.6", a3.Rows()==a2.Rows() && a3.Cols()==a2.Cols() \
	 && a3[0][0]==0 && a3[1][0]==1 && a3[0][1]==10 \
	 && a3[1][1]==11 && a3[0][2]==20 && a3[1][2]==21);
    // Comparison - same
    TEST("2.7", a3==a2);
    // different
    a3[0][2]=1;
    TEST("2.8", !(a3==a2));
    a3[0][2]=20;
    // different sizes
    TEST("2.9", !(a3==a1));
    // Copy assignment
    a1.Resize(a2.Rows(), a2.Cols());
//    a1 = a2; 
    a1.Copy(a2); 
    TEST("2.10", (a2==a1));
    a1[0][2]=200;
    TEST("2.11", !(a2==a1));
    // Assign() alias for copy assignment
    a1.Resize(0,0);
    TEST("2.11b", !(a1==a2));
    a1.CloneArray(a2); 
    TEST("2.11c", (a2==a1));
    a1[0][2]=200;
    // Get rows
    probRef r1(5, new float[5], new float[5], new float[5]);
    r1 = a1.RowRef(1);
    // doesn't modify the length of the dataref
    TEST("2.13", r1.Len() == 5);
    TEST("2.14", r1[0]==1 && r1[1]==11 & r1[2]==21);
    // undersize Ref to GetRow
    r1[2]=255; 
    r1.SetLen(2);
    r1 = a1.RowRef(1);
    TEST("2.16", r1[0]==1 && r1._El(2)==255);
    // Ref._El does no bounds checking
    // undersize Ref to SetRow
    a1.RowRef(0) = r1;
    TEST("2.17", a1[0][0]==1 && a1[0][1]==11 
	 && a1[0][2]==200);
    // oversize ref to SetRow
    r1.SetLen(5);
    a1.RowRef(0) = r1;
    TEST("2.18", a1[0][2]==255);
    // Transpose
    a1.CloneArray(a2);
    a1.Transpose();
    TEST("2.20", a1.Rows()==3 && a1.Cols()==2 && a1[0][0]==0 && a1[1][0]==10 
	 && a1[2][0]==20 && a1[0][1]==1 && a1[1][1]==11 && a1[2][1]==21);
    // Get cols
    r1 = a1.ColRef(1);
    // doesn't modify the length of the dataref
    TEST("2.22", r1.Len() == 5);
    TEST("2.23", r1[0]==1 && r1[1]==11 & r1[2]==21);
    // undersize Ref to GetCol
    r1[2]=255;
    r1.SetLen(2);
    r1 = a1.ColRef(1);
    TEST("2.25", r1[0]==1 && r1._El(2)==255);
    // undersize Ref to SetCol
    a1.ColRef(0) = r1;
    TEST("2.26", a1[0][0]==1 && a1[1][0]==11 
	 && a1[2][0]==20);
    // oversize ref to SetCol
    r1.SetLen(5);
    a1.ColRef(0) = r1;
    TEST("2.27", a1[2][0]==255);
    a1[2][0]=20;
    a1.Transpose();	// back to 3 cols, 2 rows
    a1[0][0]=0;
    a1[0][1]=10;
    a1[0][2]=20;
    // Sum
    a1.Resize(3,2);
    a1[0][0] = 48;	a1[0][1] = 51;
    a1[1][0] = 49;	a1[1][1] = 52;
    a1[2][0] = 50;	a1[2][1] = 53;    
    TEST("2.31a",  (a1.ColRef(0)).Sum()==147);
    TEST("2.31a",  (a1.ColRef(1)).Sum()==156);
    // Clear
    a1.Clear();
    TEST("2.32", a1.Sum()==0);
    // Resize
    a1.CloneArray(a3);
    a1.Resize(3,2);
    TEST("2.33", a1.Rows()==3 && a1.Cols()==2 && a1[0][1]==10 && a1[1][1]==11
	 && a1[2][1]==0);
    a1.CloneArray(a3);
    a1.Resize(1,4);
    TEST("2.34", a1.Rows()==1 && a1.Cols()==4 && a1[0][1]==10 && a1[0][2]==20
	 && a1[0][3]==0);
    // DAT file access
//    char tmpnam[] = "/usr/tmp/tmp.dat";
//    a3.saveDAT(tmpnam);
//    a1.loadDAT(tmpnam);
//    TEST("2.35", a1==a3);
//    cerr << "DAT tests a bit thin at the moment." << endl;
    // test array set
    a1.Resize(2,4);
    a1.Fill(99);
    TEST("2.36", a1.Rows()==2 && a1.Cols()==4 && a1[0][0]==99 && a1[1][0]==99 && a1[0][1]==99 && a1[1][1]==99);
    // Direct row-to-row copying via refs
    a1[0][0] = 101; a1[0][1] = 102; 
    a1[1][0] = 103; a1[1][1] = 104;
    //    a1[1] = a1[0];
    // You can't do it like that because a1 only has ONE ref, which is re-used.
    probRef re(a1[0]);
    a1[1] = re;
    TEST("2.37", a1[1][0]==101 && a1[1][1]==102);
    // Basic sanity check on ref : it's a row, so it has 4 columns
    // and the colums are spaced a row's length apart
    //    TEST("2.37b", re.Len()==4 && re.Step()==2);
    // No, columns are fast rotating, so step ==1
    TEST("2.37b", re.Len()==4 && (re.Step()==1 || re.Step()==2));
    if(re.Step()==1)	cout << "Columns are fastest rotating (a(0,0),a(0,1)..)" << endl;
    else		cout << "Rows are fastest rotating (a(0,0),a(1,0)..)" << endl;
    // Col-to-col copying via refs
    a2.Resize(4,2);
    a2[0][0] = 101; a2[0][1] = 102; 
    a2[1][0] = 103; a2[1][1] = 104;
    re.CloneRef(a2.ColRef(0));
    // Basic sanity check on ref : it's a col of a 4x2, so it has 4 rows
    // cols have their row elements close-packed
    //    TEST("2.38a", re.Len()==4 && re.Step()==1);
    // nope, row elements are spaced out
    TEST("2.38a", re.Len()==4 && (re.Step()==2 || re.Step()==1));
    if(re.Step()==2)	cout << "Columns are fastest rotating (a(0,0),a(0,1)..)" << endl;
    else		cout << "Rows are fastest rotating (a(0,0),a(1,0)..)" << endl;
    a2.ColRef(1) = re;
    TEST("2.38", a2[0][1]==101 && a2[1][1]==103);
    // Col-to-row copying
    a1.Fill(0);
    a1[1] = a2.ColRef(0);	// Can do this without copying to a Ref
                                // because a1 and a2 are diff arrays, hence
                                // different internal Refs used
    TEST("2.39", a1[1][0]==101 && a1[1][1]==103);
    // Row-to-col copying
    a2.Fill(11);
    a2.ColRef(1) = a1[1];
    TEST("2.40", a2[0][1]==101 && a1[1][1]==103);

    // RowRef too
    re.CloneRef(a2[1]);
    probRef r2(a2.RowRef(1));
    TEST("2.41", re.Data()==r2.Data() && re.Len()==r2.Len() && re.Step()==r2.Step());

    // RowRef, ColRef out of bounds
    probRef *rup;
    a2.Resize(2,2);
    // Array now just aborts for reference rFequests out of bounds - don't test
    // also, ColRef and RowRef now return by value - don't take address!
//    rup = &a2.ColRef(0);
//    TEST("2.42b", rup!=0);
//    rup = &a2.ColRef(1);
//    TEST("2.42c", rup!=0);
//    rup = &a2.RowRef(0);
//    TEST("2.43b", rup!=0);
//    rup = &a2.RowRef(1);
//    TEST("2.43c", rup!=0);

    // Print()
    a1.Print(cerr, "Test 2 done.");
}

void testProbArray(void)
{   /* test the probabilistic parts of the probArray */
    probArray a1(2, 3), a2(2, 3);
    floatArray fa(2, 3);
    int i,j;
    for(i=0; i<2; ++i) {
	for(j=0; j<3; ++j)  {
	    a1[i][j] = 10*i+j;
	    a2[i][j] = 1;
	    fa[i][j] = 10*i+j+1;
	    a1.PSDv()[i][j] = 1;
	    a2.PSDv()[i][j] = 1;
	    a1.NSDv()[i][j] = 1;
	    a2.NSDv()[i][j] = 1;
	}
    }
    a1.PSDv()[1][0] = 0;

    a1.Add(a2);
    TEST("3.1", a1[0][0]==1 && a1[1][0]==11 
	 && Feq(a1.PSDv()[0][0], sqrt(2)) && Feq(a1.PSDv()[1][0], 1));
    TEST("3.2", a1.Mahal(fa)==0);
    a1.PSDv()[1][1] = 0;
    fa[0][0] = 0;	// from a1[0][0]=N(1,sqrt(2)
    TEST("3.3", Feq(a1.Likelihood(fa), exp(-0.5*pow(1/sqrt(2),2))/(pow(sqrt(2),4)*pow(2*M_PI, 2.5))));
//    cout << "Lik:" << a1.Likelihood(fa) << ", Exp:" << exp(-0.5*pow(1/sqrt(2),2))/(pow(sqrt(2),4)*pow(2*M_PI, 2.5)) << endl;
    a1.PSDv().Print();
    // Test cast to floatArray
    // floatArray *f2 = (floatArray)a1);
    // SGI's DCC compiles this correctly (i.e. puts the return of a1.castToFAP
    // into f2) but g++ 2.6.3 doesn't - it makes a temporary copy of the 
    // floatVec and returns the address of that (and then deletes the 
    // temporary on the SGI but not on the Sun).  Thus to test, we have to 
    // do the following - grab the address directly.
    floatArray* f2 = a1.castToFAPtr();
    TEST("3.4", f2->Rows()==2 && f2->Cols()==3 && f2->El(0,0)==1 && f2->El(1,0)==11);
    f2->Print();

    floatArray *f3 = &(a1.giveExp());
    cout << "f2@" << (void*)f2 << ", f3@" << (void*)f3 <<  endl;
    TEST("3.5", ((void*)f2) == ((void*)f3));
    TEST("3.6", f3->Rows()==2);

}

void main(void)
{
    testProbVec();
    testArray();
    testProbArray();
}

#endif /* MAIN */

