//
// ProbVec2.H
//
// Second version of composite floatVec/floatArrays with range limits.
// This version seeks to encode the asymmetry of tolerances for +ve 
// and -ve deviations in various surfex conditions, so has a 
// a separate 'sd' param for each polarity around the expectation.
//
// dpwe@media.mit.edu 1995nov24
//

#ifndef _PROBVEC2_H_
#define _PROBVEC2_H_

#include "nSurface.H" 	// for floatVec base class
#include <math.h>	// M_PI used

class probArray;

class probRef : public floatRef {
    // Like a floatRef, but includes two extra floatRefs 
    // holding positive and negative excursion limits for each point
friend class probArray;	// allow access to _Init
protected:
    floatRef nsd;	// std deviation in the negative sense
    floatRef psd;	// psd is second sd-like value for positive devs

    void  _Swap(probRef& pr) 
             { floatRef::_Swap(pr); nsd._Swap(pr.NSDv()); psd._Swap(pr.PSDv()); }
    void _Init(int l, float *d, float *dps, float *dns, int s=1)
             { floatRef::_Init(l,d,s);psd._Init(l,dps,s);nsd._Init(l,dns,s);}
public:
    probRef(int l=0, float *d=0, float *dps=0, float *dns = 0, int s=1) 
	: floatRef(l,d,s)
        { psd._Init(l, dps, s); nsd._Init(l, dns, s); }
    probRef(const probRef& pv)
        { floatRef::CloneRef(pv); 
	  psd.CloneRef(pv.PSD()); nsd.CloneRef(pv.NSD()); }
    ~probRef(void)
        { }

    // Otcl accessors
    floatRef& giveExp(void)	{ return *(floatRef *)this; }
    floatRef& givePSD(void)		{ return psd; }
    floatRef& giveNSD(void)		{ return nsd; }
    const floatRef& PSD(void)	const	{ return psd; }
    const floatRef& NSD(void)	const	{ return nsd; }
    floatRef& PSDv(void)		{ return psd; }
    floatRef& NSDv(void)		{ return nsd; }

    // Basic ops
    probRef& operator= (const probRef& pr)	// copy the data
        { psd.operator= (pr.PSD()); 
	  nsd.operator= (pr.NSD()); 
	  floatRef::operator= (pr); return *this; }
    probRef& operator= (const floatRef& fr)	// copy the data
        { psd.Clear(); nsd.Clear(); floatRef::operator= (fr); return *this; }
    void CloneRef(const probRef& pr)
        { floatRef::CloneRef(pr); psd.CloneRef((const floatRef&)pr.PSD());
	  nsd.CloneRef(pr.NSD()); }
    void Copy(const probRef& pr)	{ *this = pr; }	// copy data
    void Clear(void)
        { floatRef::Clear();  psd.Clear(); nsd.Clear(); }

    // Basic arithmetic with std devs
    probRef& operator += (const probRef& pr)
        { Add(pr);  return *this; }
    probRef& operator -= (const probRef& pr)
        { Sub(pr);  return *this; }
    probRef& operator *= (const floatRef& fr)
        { Mul(fr);  return *this; }
    probRef& operator /= (const floatRef& fr)
        { Div(fr);  return *this; }
    void Add(const probRef& pr)	// Variances add, hence sd->Hypot
        { floatRef::Add(pr); psd.Hypot(pr.PSD()); nsd.Hypot(pr.NSD()); }
    void Sub(const probRef& pr)	// Variances still add even for subtract
        { floatRef::Sub(pr); psd.Hypot(pr.PSD()); nsd.Hypot(pr.NSD()); }
    void Scale(double f)	// std devs scale
        { floatRef::Scale(f); psd.Scale(f); psd.FWRectify(); 
	  nsd.Scale(f); nsd.FWRectify(); }
    void operator *= (double d)
        { Scale(d); }
    void operator /= (double d)
        { Scale(1.0/d); }
    // What happens when you multiply random vars?  No idea; just scale els
    void Mul(const floatRef &fr)
        { floatRef::Mul(fr);  psd.Mul(fr); psd.FWRectify(); 
	  nsd.Mul(fr); nsd.FWRectify(); }
    void Div(const floatRef &fr)
        { floatRef::Div(fr);  psd.Div(fr); psd.FWRectify(); 
	  nsd.Div(fr); nsd.FWRectify(); }
    // Hmm, want to be able to add incoherent probabilistic spectra
    void Hypot(const probRef& pr)	// Variances still add??
        { floatRef::Hypot(pr); psd.Hypot(pr.PSD()); nsd.Hypot(pr.NSD()); }

    // probVec-specific functions
    double Mahal(const floatRef &fr);	// return the mahalanobis distance
	                                // between the expected vector & this
	                                // Only dimensions with nonzero SD
	                                // are included.
    double Likelihood(const floatRef &fr); // score the likelihood that the 
	                                   // passed floatref came from this pv
    double AvgLogLikBits(const floatRef &fr);  // different likelihood fn
    double AvgLogLikBits1dB(const floatRef &fr, 
			    floatRef* pbrkdown = 0);  // a real probability

    // behavior flag
    static int avgAllChans;	// if 1, AvgLogLikBits1dB avgs over z chans too

};

class probVec : public probRef {
    // Like a floatVec, but includes a second, conformal floatVec
    // which is the std. dev. for each point in the vector.
public:
    probVec(int l=0) : probRef(l)
        { data = new float[len]; 
	  psd.SetData(new float[len]); nsd.SetData(new float[len]); Clear(); }
    probVec(const probVec& pv) : probRef(pv.Size()) 
        { data = new float[len]; 
	  psd.SetData(new float[len]); nsd.SetData(new float[len]);
	  probRef::Copy(pv); }
    probVec(const probRef& pr) : probRef(pr.Size()) 
        { data = new float[len]; 
	  psd.SetData(new float[len]); nsd.SetData(new float[len]);
	  probRef::Copy(pr); }
    probVec(const floatRef& fr) : probRef(fr.Size())
        { data = new float[len]; floatRef::Copy(fr);
	  psd.SetData(new float[len]); psd.Clear(); 
	  nsd.SetData(new float[len]); nsd.Clear(); }
    ~probVec(void)
        { delete [] data; delete [] psd.Data(); delete [] nsd.Data(); }

    probVec& operator = (const probVec& pv)
        { if(this != &pv) CloneVec(pv); return *this; }

//    operator floatRef& (void)
//        { return *(floatRef *)this; }

    // Basic ops
    void CloneVec(const probRef& pr);
    void Resize(int l)
        { probVec tmp(l); tmp.Clear(); tmp.Copy(*this); _Swap(tmp); }
    // no Resample?

};

probVec operator + (const probRef& a, const probRef& b);
probVec operator + (const probRef& a, double d);
probVec operator + (double d, const probRef& a);

probVec operator - (const probRef& a, const probRef& b);
probVec operator - (const probRef& a, double d);
probVec operator - (double d, const probRef& a);

probVec operator * (const probRef& a, const floatRef& b);
probVec operator * (const floatRef& a, const probRef& b);
probVec operator * (const probRef& a, double d);
probVec operator * (double d, const probRef& a);

probVec operator / (const probRef& a, const floatRef& b);
// probVec operator / (const floatRef& a, const probRef& b);
probVec operator / (const probRef& a, double d);
// probVec operator / (double d, const probRef& a);


class probArray : public probVec, public arrayRef {
protected:
    probRef pr;
    arrayRef psdr;
    arrayRef nsdr;
    floatArray *pfa;	// just for faking the cast.

    void     _Swap(probArray& pa) 
               { probRef::_Swap(pa);  arrayRef::_Swap(pa); }
public:
    probArray(int rs=0, int cs=0) : probVec(rs*cs), arrayRef(rs, cs)
                { pfa = 0; adata = data; psdr._Init(rows, cols, psd.Data());
		  nsdr._Init(rows, cols, nsd.Data());
                  DBG((cout << "probArray(" << rs << "," << cs << ")@" \
		       << (void *)this << endl)); }
    probArray(const probArray& pa) : probVec(pa.rows*pa.cols), 
                                     arrayRef(pa.rows, pa.cols)
                { pfa = 0; adata = data; psdr._Init(rows, cols, psd.Data()); 
		  nsdr._Init(rows, cols, nsd.Data()); Copy(pa); }
    ~probArray(void) 
                { if(pfa)	{ pfa->SetData(0); delete pfa; }  }
    probArray& operator = (const probArray &pa);

    operator floatArray&(void)	{ return *castToFAPtr(); }

    int operator == (const probArray& pa) 
                 { return arrayRef::operator == (pa); }

    floatArray* castToFAPtr(void)
                { if(pfa)	{ pfa->SetData(0); delete pfa; }
		  pfa = new floatArray(); pfa->_Init(rows, cols, data);
		  DBG((cout << "Casting prAr" << (void *)this << " to flAr" \
		       << (void *)pfa << endl));
		  return pfa; }

    // Otcl accessors
    floatArray& giveExp(void)	{ return *castToFAPtr(); }
    floatArray& givePSD(void)
                { if(pfa)	{ pfa->SetData(0); delete pfa; }
		  pfa = new floatArray(); pfa->_Init(rows, cols, psd.Data());
		  DBG((cout << "Casting prAr" << (void *)this << " to psflAr" \
		       << (void *)pfa << endl));
		  return *pfa; }
    floatArray& giveNSD(void)
                { if(pfa)	{ pfa->SetData(0); delete pfa; }
		  pfa = new floatArray(); pfa->_Init(rows, cols, nsd.Data());
		  DBG((cout << "Casting prAr" << (void *)this << " to nsflAr" \
		       << (void *)pfa << endl));
		  return *pfa; }
    arrayRef& PSDv(void)		{ return psdr; }
    arrayRef& NSDv(void)		{ return nsdr; }
    const arrayRef& PSD(void) const	{ return psdr; }
    const arrayRef& NSD(void) const	{ return nsdr; }
    void      Copy(const probArray& pa);
    void      Resize(int nr, int nc);
    void      Reshape(const probArray &pa)	// match the shape
                { Resize(pa.rows, pa.cols); }
    void      CloneArray(const probArray& pa)
                { Reshape(pa); Copy(pa); }

    float     Val(int r, int c) const { return arrayRef::Val(r, c); }
    float&    El(int r, int c)        { return arrayRef::El(r, c); }
    void      Set(int r, int c, float d)  { arrayRef::Set(r, c, d); }
    probRef  ColRef(int c) const
                { probRef _pr;
		  _pr._Init(rows, data+c, psd.Data()+c, nsd.Data()+c, cols);
		  return _pr; }
    probRef  RowRef(int r) const
                { probRef _pr;
		  _pr._Init(cols, data+r*cols, 
			    psd.Data()+r*cols, psd.Data()+r*cols, 1);
		  return _pr; }
    probRef  operator[] (int r)
                { return RowRef(r); }

    void      Transpose(void);
//    void      Resample(int nr, int nc);
    int       Load(const char *filename, int dataset=0);
    void      Print(ostream &os=cout, char *name=0)
                {  arrayRef::Print(os, name);  }

};

#endif /* _PROBVEC2_H_ */
