//
// moreFVops.C
//
// Implementation of various sophisticated operations on floatvecs
// modeled after my matlab functions
//
// dpwe@media.mit.edu 1996jan24
//

#include "moreFVops.H"

#ifdef __MWERKS__
#include <fp.h>   /* Metrowerks */
#endif /* __MWERKS __ */

void QuadFit3pt(double pm1, double p0, double pp1, 
		double *px, double *py, double cfs[])
{   /* Fit a quadratic peak to the three uniformly-spaced points p0-p2.
     * Return the local max/min x and y, and the coefs in cfs.
     * On return, p0 = cfs[0]; pm1 = cfs[0]-cfs[1]+cfs[2];
     * pp1 = cfs[0]+cfs[1]+cfs[2].
     */
    /* Y = k(x-x0)^2 + c, x = [-1 0 1] */
    double k, x0, c;
    k  = (pm1+pp1)/2-p0;
    x0 = (pm1-pp1)/(4*k);
    c  = p0 - k*(x0*x0);
    if(px)    *px = x0;
    if(py)    *py = c;
    if(cfs)  {
	cfs[0] = c+k*x0*x0;  cfs[1] = -2*k*x0; cfs[2] = k;
    }
}

floatVec SubSmooth(const floatRef& fr, int subfact, int winlen)
{   /* smooth a floatref with a hanning window of <winlen> nonzero points 
       subsample the result by <subfact>; return a new floatVec with 
       the result. */
    // create hann window - must be odd len
    if( (winlen&1)==0)    ++winlen;
    double *win = new double[winlen];
    int i;
    double accum = 0;
    for(i=0; i<winlen; ++i) {
	win[i] = 0.5-0.5*cos( (i+1)/((double)winlen+1) *2*M_PI);
	accum += win[i];
    }
    // normalize so sum(win[i]) is 1 - smoothing preserves const vals
    for(i=0; i<winlen; ++i) {
	win[i] /= accum;
    }    
    int winhlen = (winlen-1)/2;     // how far back we have to look
    // figure size of output
    int nipts = fr.Size();
    int nopts = (nipts+subfact-1)/subfact;
    floatVec op(nopts);
    int curopt = 0;
    int cuript = 0;
    int iptx;
    while(curopt < nopts) {
	// when we're close to the start, repeat the first point
	accum = 0;
	int imin = 0, imax = winlen;
	if(cuript < winhlen) {
	    iptx = 0;
	    for(i = 0; i<winhlen-cuript; ++i) {
		accum += win[i] * fr[iptx];
	    }
	    imin = winhlen-cuript;
	} else {
	    iptx = cuript-winhlen;
	}
	if(cuript+winhlen >= nipts - 1) {
	    imax = winlen - 1 - (cuript+winhlen-nipts+1);
	}
	for(i = imin; i<imax; ++i) {
	    accum += win[i]* fr[iptx];
	    ++iptx;
	}
	if(imax<winlen) {
	    for(i = imax; i<winlen; ++i) {
		accum += win[i] * fr[iptx];
	    }
	}
	op.El(curopt++) = accum;
	cuript += subfact;
    }
    delete[] win;
    return op;
}

int FindLocMax(const floatRef& fr, int base, int sense, 
	       int maxnotmin /* =1 */, int saddleOK /* =0 */ )
{   /* Walk along a vector from the specified position looking for 
     * a local maximum.  Return its index if found, or -1 if none 
     * found.  sense==1 means search upwards, -1 (or <=0) means down.
     * If maxnotmin is 1, find a max, else find a min.
     * Now acts to be as early as possible - flat regions are reported 
     * as the first point the flatness encountered, although not 
     * until the far side is seen.
     * If saddleOK is not zero, will return on a saddle point as well 
     * as a true max i.e. two successive points the same counts.
     * If search is upward and saddleOK is set, the 'I didn't find one'
     * return code is fr.Size() i.e. there is an implied saddle 'beyond 
     * the end' (hence -1 is the 'true' saddle location for down searches
     * already).
     */
    int lim;
    int inflat = 0;
    int flatstart;
    float px, cx, nx;	// previous, current and next x's
    int sign;
    if(maxnotmin>0)	sign = 1;
    else		sign = -1;
    if(sense>0)	{
	sense = 1;
	lim = fr.Size()-1;
	if(base==0) {
	    cx = sign*fr[0]-sign;	// make a flat edge a max
	} else {
	    cx = sign*fr[base-1];
	}
    } else {
	sense = -1;
	lim = 0;
	if(base==fr.Size()-1) {
	    cx = sign*fr[fr.Size()-1]-sign;	// make a flat edge a max
	} else {
	    cx = sign*fr[base+1];
	}
    }
    nx = sign*fr[base];
    // actually do the walk
    while(base != lim) {
	px = cx; cx = nx; nx = sign*fr[base+sense];
	if(cx>px && cx>nx) {
	    return base;
	} else if(cx>px && cx==nx) {
	    if(saddleOK) {
		return base;
	    } else {
		inflat = 1;
		flatstart = base;
	    }
	} else if(inflat && cx > nx) {
	    // reached the end of a flat region
	    // return base;
	    return flatstart;
	} else if(cx != px) {
	    inflat = 0;
	}
	base += sense;
    }
    // Hit the limit.  But if the limit is at least larger than its 
    // neighbor, return that OK
    if(nx > cx)
	return base;
    else {
	if(saddleOK && sense>0) {
	    return fr.Size();
	} else {
	    return -1;
	}
    }
}

int NearestPeak(const floatRef& fr, double startix, double ixwin, 
		double *pos, double *ht)
{   /* Pick out a peak in the vector fr at or near the (fractional)
     * index ix.  To count, the found local maximum must be within 
     * ixwin of the specified position.  If a peak is found, it returns 
     * 1 and the quadratic-interpolated position of the peak is returned 
     * in *pos (and height in *ht).  Otherwise, return 0.
     * Specifying ixwin as zero disables limiting window.
     */

    int rc = 0;

    // Search in the direction of the local slope here
    double ym, yp, y0;
    int ixx = (int)rint(startix);
    int ixmax = fr.Size();
    if(ixwin == 0)          ixwin = ixmax;    // no limit

    if(ixx<=0)		{  ixx = 0;  ym = fr[ixx]; }
    else 		{  ym = fr[ixx-1];    }
    if(ixx>=ixmax-1)	{  ixx = ixmax-1; yp = fr[ixx]; }
    else 		{  yp = fr[ixx+1]; }
    if(yp>ym) {
	// looks like we're on the lower edge of a max
	ixx = FindLocMax(fr, ixx, 1);
    } else if(yp < ym) {
	// looks like we're above a max;
	ixx = FindLocMax(fr, ixx, -1);
    } else {
	// it's flat here.  Search both ways
	int ixxu = FindLocMax(fr, ixx, 1);
	int ixxd = FindLocMax(fr, ixx, -1);
	if(ixxu == -1 && ixxd == -1)  {
	    ixx = -1;	// none found
	} else if(ixxu == -1 
		  || (ixxd != -1 
		      && fabs(startix - ixxd) < fabs(startix - ixxu)) ) {
	    ixx = ixxd;
	} else {
	    ixx = ixxu;
	}
    }
    if(ixx!=-1) {
	// Found a point.  Now fit a quadratic to the three points
	// If local max was one of the end points, treat special - 
	// make it a symmetric peak at the end
	if(ixx<=0)		{  ixx = 0;  ym = fr[1]; }
	else	 		{  ym = fr[ixx-1];    }
	if(ixx>=ixmax-1)	{  ixx = ixmax-1; yp = fr[ixx-1]; }
	else 			{  yp = fr[ixx+1]; }
	double x, y;
	QuadFit3pt(ym, fr[ixx], yp, &x, &y, NULL);
	x += ixx;
	// Check the window
	if( fabs(x-startix) <= ixwin) {   // or <= ?
	    *pos = x;
	    *ht = y;
	    rc = 1;
	}
    }
    return rc;
}

double FracInterp(const floatRef& fr, double ix, int order /*=6*/)
{   /* Interpolate the values in the vector to return the value 
     * that is interpolated at the fractional index <ix> (from 0).
     * Uses truncated sinc interp at present, there's always 
     * the future for 4pt lagrangians 
     */
    // Calculate the sinc window
    double *win = new double[order];
    double flix, val;
    int intix = (int)ix;	// i.e. floor
    flix = ix-intix;		// 0<=flix<1, intix+flix = ix
    int hwin = (int)(order/2);	// how far back to step
    int i;
    for(i = 0; i<order; ++i) {
	val = M_PI*(i - (hwin-1) - flix);
	if(val==0)
	    win[i] = 1;
	else
	    win[i] = sin(val)/val;
    }
    int frlen = fr.Size();
    int indx;
    val = 0;
    for(i = 0; i<order; ++i) {
	indx = -(hwin-1)+intix + i;
	if(indx>=0 && indx < frlen) {
	    val += win[i]*fr[indx];
	}
    }    
    delete [] win;
    return val;
}

floatVec FracInterpVec(const floatRef& src, const floatRef& ixs, 
		       int order /*=6*/)
{   /* return a whole vector of values interpolated from <src> at the 
       indexes indicated in <ixs> */
    int ixlen = ixs.Size();
    int i;
    floatVec rslt(ixlen);
    for(i=0; i<ixlen; ++i) {
	rslt[i] = FracInterp(src, ixs[i], order);
    }
    return rslt;
}

int PeakWidth(const floatRef &src, int ix, int *pll /*=0*/, int *pul /*=0*/)
{   /* How wide is the gap between local minima occupied by this 
       peak?  Basically the same logic as ExtractPeak, but rather 
       than copying data, it simply returns a width. 
       Will fill in the actual indices of the upper and lower limit 
       into *pll and *pul if they are non-null. */
    // find upper and lower adjacent minima (-1 as last arg searches f min)
    int ul = FindLocMax(src, min(ix+1,src.Size()-1), 1, -1, 1);
    int ll = FindLocMax(src, max(0,ix-1), -1, -1, 1);	// -1 to search down
    int rslt;
    // Now FindLocMax(,,,saddleOK=1) will return fr.Size() (i.e. one 
    // beyond the end) if it runs out of an upwards search, so 
    // let these through - will always find a peak
//    if(ul == -1 || ll == -1) {
//	rslt = 0;
//	// make both limits return as -1?  No, leave the info there for now
//    } else {
	rslt = ul - ll;
//    }
    if(pll)	*pll = ll;
    if(pul)	*pul = ul;
    return rslt;
}


floatVec ExtractPeak(const floatRef& src, int ix)
{   /* Search both ways from indx for adjacent local minima, or the 
       edge of the array.  Copy everything between these two limits from 
       <src> into the result;  other points of result (same size as <src>)
       are zero. */
    floatVec rslt(src.Size());

    // find upper and lower adjacent minima (-1 as last arg searches f min)
    int ul = FindLocMax(src, ix, 1, -1, 1);
    if(ul == -1)	ul = src.Size()-1;	// didn't find, but must take
    int ll = FindLocMax(src, ix, -1, -1, 1);	// -1 to search down
    if(ll == -1)	ll = 0;			// didn't find, but must take
    int i;
    for(i=ll; i<=ul; ++i)	rslt[i] = src[i];

    return rslt;
}

#define LOG2BASEE  0.69314718

floatVec LogShift(const floatRef& src, double shift, double binsperoct, 
		  int order /*=6*/)
{   /* <src> is treated as a log-sampled signal, with <binsperoct>
       bins per octave.  The output is the same size, but with the 
       contents delayed by <shift> units, where one unit is equivalent 
       to the lowest sample point's linear value. 
       <order> is the order used for FracInterp. */
    floatVec rslt(src.Size());
    int i, sz = src.Size();
    double it, ot, ix;

    // optimization : only apply this for the range of Y corresponding to 
    // the nonzero elements in x.
    int minx, miny, maxx, maxy;
    double fw = LOG2BASEE/binsperoct;
    double bw = 1.0/fw;
    for(i = 0; src[i] == 0 && i<sz; ++i)
	;
    minx = i;
    for(i = sz-1; src[i] == 0 && i>minx; --i)
	;
    maxx = i;
    miny = (int)max(0.0, (int)bw*log(exp(fw*minx)+shift)-1);
    maxy = (int)min(sz-1.0, (int)bw*log(exp(fw*maxx)+shift)+1);

//    for(i = miny; i<=maxy; ++i) {
//	// Figure linear indexes for both input and output
//	it = exp(fw*i) - shift;
//	if(it >= 1) {	// only for indexes actually in x
//	    // convert back to log-sampling index
//	    ix = bw * log(it);
//	    if(ix < src.Size()) {
//		rslt._El(i) = FracInterp(src, ix, order);
//	    }
//	}
//    }
//    return rslt;
    if(maxy>=miny) {
	float *pf = &rslt[miny];
	for(i = miny; i<=maxy; ++i) {
	    ix = bw * log(exp(fw*i) - shift);
	    // rslt._El(i) = FracInterp(src, ix, order);
	    *pf++ = FracInterp(src, ix, order);
	}
    }
    return rslt;
}
floatArray LogShift(const arrayRef& ar, double shift, double binsperoct, 
		    int order/*=6*/)
{   /* LogShift for arrays */
    floatArray rslt(ar.Rows(), ar.Cols());
    int r; 
    for(r=0; r<rslt.Rows(); ++r) {
	rslt.RowRef(r) = LogShift(ar.RowRef(r), shift, binsperoct, order);
    }
    return rslt;
}
void LogShiftAdd(floatRef& accum, const floatRef& src, double shift, 
		 double binsperoct, double scale /*=1.0*/, int order /*=6*/)
{   /* Optimization for the way LogShift is used in Weft::Update - 
     * shift src by the logshift, then *add* it into accum, rather 
     * than generating and copying a new floatVec.  Efficient 
     * because often the shift and the nonzero region of src
     * mean that no additions need be performed
     */
    // Guts all copied from LogShift
    int i, sz = src.Size();
    double it, ot, ix;

    int minx, miny, maxx, maxy;
    double fw = LOG2BASEE/binsperoct;
    double bw = 1.0/fw;
    for(i = 0; src[i] == 0 && i<sz; ++i)
	;
    minx = i;
    for(i = sz-1; src[i] == 0 && i>minx; --i)
	;
    maxx = i;
    miny = (int)max(0.0, (int)bw*log(exp(fw*minx)+shift)-1);
    maxy = (int)min(sz-1.0, (int)bw*log(exp(fw*maxx)+shift)+1);

    for(i = miny; i<=maxy; ++i) {
	ix = bw * log(exp(fw*i) - shift);
	accum[i] += scale * FracInterp(src, ix, order);
    }
}

int FindAllMaxes(const floatRef& fr, int ixs[], int ixsz, double thresh /*=0*/)
{   /* Go through <fr> looking for local maxes.  Store the integer 
     * locations of the local max points in the ixs[] array, up to 
     * a maximum of <ixsz> elements.  Return the actual number of 
     * elements written into <ixs>.
     * Local maxima smaller than <thresh> won't be counted.
     */
    int nfound = 0;
//    double yp, yz, ym;
//    yz = fr[1]; yp = fr[0];
    int ix = -1;
    int frsz = fr.Size();
    while(nfound < ixsz && (ix < frsz-1)
	  && (ix = FindLocMax(fr, ix+1, 1, 1))>=0 ) {
//    while(ix >=0 && ix<frsz - 1 && nfound < ixsz) {
//	ym = yz; yz = yp; yp = fr[ix+1];
//	if(yz > thresh && yz > ym && yz > yp) {
//	if(ix > 0) {
	if(fr[ix]>=thresh)
	    ixs[nfound++] = ix;
//	}
    }
    return nfound;
}

int FindAllPeaks(const floatRef& fr, floatVec** pPosRtn, floatVec** pHtRtn, 
		 double thresh /*=0*/)
{   /* Go through a vector selecting all the peaks above a threshold;
     * return them in two new floatVecs, giving the position and 
     * corresponging height of each quad-interpolated peak found.
     * Return the size of these two new arrays.
     * If addresses for returns are null, don't fill them in.
     */
    int npks = 0;

    int *ixs = new int[fr.Size()];
    npks = FindAllMaxes(fr, ixs, fr.Size(), thresh);
    if(pPosRtn) 	*pPosRtn = new floatVec(npks);
    if(pHtRtn)		*pHtRtn  = new floatVec(npks);
    int i; 
    for(i=0; i<npks; ++i) {
	double a,b;
	NearestPeak(fr, ixs[i], 0, &a, &b);
	if(pPosRtn)	(*(*pPosRtn))[i] = a;
	if(pHtRtn)	(*(*pHtRtn))[i] = b;
    }
    delete [] ixs;
    return npks;
}

static int compareFloats(const void* a, const void* b)
{   /* qsort comparison fn for sorting lists of float-headed elements */
    float fa = *(float*)a;
    float fb = *(float*)b;

    if(fa>fb) 		return 1;
    else if(fa<fb)	return -1;
    else		return 0;
}

static int ncompareFloats(const void* a, const void* b)
{   /* qsort comparison fn for sorting lists of float-headed elements */
    float fa = *(float*)a;
    float fb = *(float*)b;

    if(fa<fb) 		return 1;
    else if(fa>fb)	return -1;
    else		return 0;
}

void SortedIndices(const floatRef& fr, int ixs[], int ixssz, int sense/*=1*/)
{   /* Without changing fr, fill ixs (which has the same number of elements) 
     * with the indexes of the elements if fr, which, presented in that order, 
     * would be an ascending list.  This array, passed to IndexFloatVec, 
     * will result in a correctly sorted floatVec.
     * <ixssz> is the size of the ixs array; only the first so many 
     * of the sorted elements will be returned. 
     * <sense> > 0 implies ascending sort; <=0 means largest first.
     */
    typedef struct _duple { float f; int i; } DUPLE;
    int frsz = fr.Size();
    struct _duple *dix = new struct _duple [frsz];
    int i;
    for(i=0; i<frsz; ++i) {
	dix[i].f = fr[i];
	dix[i].i = i;
    }
    if(sense>0)
	qsort(dix, frsz, sizeof(struct _duple), compareFloats);
    else
	qsort(dix, frsz, sizeof(struct _duple), ncompareFloats);
    for(i=0; i<frsz && i<ixssz; ++i) {
	ixs[i] = dix[i].i;
    }
    delete [] dix;
}

floatVec IndexFloatVec(const floatRef& fr, int ixs[], int nix)
{   /* Construct a new floatVec with <nix> elements, each of which's value 
     * is taken from the element of <fr> enumerated by the corresponding 
     * element of <ixs>. 
     */
    floatVec rslt(nix);
    int i;
    for(i=0; i<nix; ++i) {
	rslt[i] = fr[ixs[i]];
    }
    return rslt;
}

void ReverseIntVec(int ixs[], int nix)
{   /* Reverse the order of elements in a simple 1D vector of ints, 
     * <ixs>, of length nix. 
     */
    int i,t; 
    for(i=0; i<nix/2; ++i) {
	t = ixs[i]; ixs[i] = ixs[nix-1-i]; ixs[nix-1-i] = t;
    }
}

floatVec WindowedExtract(const floatRef& fr, double winCtr, double winWidth)
{   /* Extract a portion of a floatvec into a second floatvec of the 
     * same size by applying a raised-cosine-edged flat window.
     */
    floatVec rslt(fr.Size());
    int i;
    double w, dc;
    double wqwid = winWidth/4;
    for(i=0; i<rslt.Size(); ++i) {
	dc = fabs(i-winCtr);
	if(dc < wqwid) {
	    w = 1;
	} else if(dc < 2*wqwid) {
	    w = 0.5*(1+cos((dc-wqwid)/wqwid*M_PI));
	} else {
	    w = 0;
	}
	rslt[i] = w*fr[i];
    }
    return rslt;
}

floatArray WindowedExtract(const arrayRef& ar, double winCtr, double winWidth)
{   /* WindowedExtract for arrays */
    floatArray rslt(ar.Rows(), ar.Cols());
    int r; 
    for(r=0; r<rslt.Rows(); ++r) {
	rslt.RowRef(r) = WindowedExtract(ar.RowRef(r), winCtr, winWidth);
    }
    return rslt;
}


#ifdef MAIN

#define ___quote(ARG)      #ARG
#define ___QUOTE(ARG)	___quote(ARG)
#define TEST(num,clause,elseclause)					\
          if(!(clause))	{ cerr << "FAIL Test " << num << ": "	        \
			       << ___QUOTE(clause) << endl;  elseclause ; }

int fEq(float f1, float f2)
{   // Approximately equal for the purposes of the test
    if(fabs(f1-f2) < 0.0001)
	return 1;
    else
	return 0;
}

void testSubSmooth(void)
{
    floatVec fv1(10);

    fv1.Fill(0);
    fv1[3] = 2;
    fv1[4] = 3;
    fv1[5] = 4;

    floatVec p;

    p = SubSmooth(fv1, 1, 3);
    TEST("1.1", fEq(p.Val(0),0) && fEq(p.Val(1),0) && fEq(p.Val(2),0.5) && fEq(p.Val(3),1.75) && fEq(p.Val(4),3) && fEq(p.Val(5),2.75) && fEq(p.Val(6),1) && fEq(p.Val(7),0) && fEq(p.Val(8),0) && fEq(p.Val(9),0),p.Print());

    p = SubSmooth(fv1, 3, 21);
    TEST("1.2", fEq(p.Val(0),0.551147) && fEq(p.Val(1),0.783795) && fEq(p.Val(2),0.757793) && fEq(p.Val(3), 0.49109),p.Print());

    fv1.Fill(10);
    p = SubSmooth(fv1, 1, 5);
    TEST("1.3", Sum(p == 10)==10,p.Print());

    // Check access to things not spaced unity apart
    floatArray fa(6,10);
    floatRef fr1(fa.RowRef(3));
//    cout << "Row Size=" << fr1.Size() << endl;
    fr1[3] = 2;
    fr1[4] = 3;
    fr1[5] = 4;
    p = SubSmooth(fr1, 1, 3);
    TEST("1.4", fEq(p.Val(0),0) && fEq(p.Val(1),0) && fEq(p.Val(2),0.5) && fEq(p.Val(3),1.75) && fEq(p.Val(4),3) && fEq(p.Val(5),2.75) && fEq(p.Val(6),1) && fEq(p.Val(7),0) && fEq(p.Val(8),0) && fEq(p.Val(9),0),p.Print());

    fr1.CloneRef(fa.ColRef(9));
//    cout << "Col Size=" << fr1.Size() << endl;
    fr1[3] = 2;
    fr1[4] = 3;
    fr1[5] = 4;
    p = SubSmooth(fr1, 1, 3);
    TEST("1.5", fEq(p.Val(0),0) && fEq(p.Val(1),0) && fEq(p.Val(2),0.5) && fEq(p.Val(3),1.75) && fEq(p.Val(4),3) && fEq(p.Val(5),3.75),p.Print());

    // Test rounding-up of even window length - so 2 should be the same as 3
    p = SubSmooth(fr1, 1, 2);
    TEST("1.6", fEq(p.Val(0),0) && fEq(p.Val(1),0) && fEq(p.Val(2),0.5) && fEq(p.Val(3),1.75) && fEq(p.Val(4),3) && fEq(p.Val(5),3.75),p.Print());

    cout << "OK" << endl;
}

void testQFit(void)
{
    double a, b, c, d, e;
    double cfs[3];

    QuadFit3pt(1, 0, 1, &a, &b, cfs);
    TEST("2.1", fEq(a,0) && fEq(b,0) && fEq(cfs[0],0) && fEq(cfs[1], 0) && fEq(cfs[2], 1), cout << a << " " << b << " " << cfs[0] << " " << cfs[1] << " " << cfs[2] << endl);
    QuadFit3pt(2, 3, 6, &a, &b, NULL);
    TEST("2.2", fEq(a,-1) && fEq(b,2), cout << a << " " << b << endl);
    QuadFit3pt(-3.5, -1, 0.5, NULL, NULL, cfs);
    TEST("2.3", fEq(cfs[0],-1) && fEq(cfs[1],2) && fEq(cfs[2],-0.5), cout << cfs[0] << " " << cfs[1] << " " << cfs[2] << endl);

    cout << "OK" << endl;
}

void testLocMax(void)
{
    floatVec fv(10);
    int i;
    // unambigous local max
    for(i = 0; i < 10; ++i) {
	fv[i] = exp(-0.5*pow((i-3.0)/4,2));
    }
//    fv.Print();
    floatRef fr(fv);
    // search up from below
    TEST("3.1", FindLocMax(fr, 0, 1)==3, cout << FindLocMax(fr, 0, 1) << endl);
    // search down from above
    TEST("3.2", FindLocMax(fr, 9, -1)==3, cout << FindLocMax(fr, 9, -1) << endl);
    // up and down from closer
    TEST("3.3", FindLocMax(fr, 2, 1)==3, cout << FindLocMax(fr, 2, 1) << endl);
    TEST("3.4", FindLocMax(fr, 4, -1)==3, cout << FindLocMax(fr, 4, -1) << endl);
    // up and down from the point itself - should still find it
    TEST("3.5", FindLocMax(fr, 3, 1)==3, cout << FindLocMax(fr, 3, 1) << endl);
    TEST("3.6", FindLocMax(fr, 3, -1)==3, cout << FindLocMax(fr, 3, -1) << endl);
    // flat case - make three pts the same
    fr[3] = fr[2];
    fr[4] = fr[2];
    // returns the trailing edge of the plateau with respect to search dir
//    TEST("3.7", FindLocMax(fr, 1, 1)==4, cout << FindLocMax(fr, 1, 1) << endl);
//    TEST("3.8", FindLocMax(fr, 8, -1)==2, cout << FindLocMax(fr, 8, -1) << endl);
    // no, now returns the *leading* edge
    TEST("3.7", FindLocMax(fr, 1, 1)==2, cout << FindLocMax(fr, 1, 1) << endl);
    TEST("3.8", FindLocMax(fr, 8, -1)==4, cout << FindLocMax(fr, 8, -1) << endl);
    // Searching up or down from within the plateau fails..
    TEST("3.9", FindLocMax(fr, 3, 1)==-1, cout << FindLocMax(fr, 3, 1) << endl);
    TEST("3.10", FindLocMax(fr, 3, -1)==-1, cout << FindLocMax(fr, 3, -1) << endl);
    // .. but starting just at the edge of it works
//    TEST("3.11", FindLocMax(fr, 2, 1)==4, cout << FindLocMax(fr, 2, 1) << endl);
//    TEST("3.12", FindLocMax(fr, 4, -1)==2, cout << FindLocMax(fr, 4, -1) << endl);
    TEST("3.11", FindLocMax(fr, 2, 1)==2, cout << FindLocMax(fr, 2, 1) << endl);
    TEST("3.12", FindLocMax(fr, 4, -1)==4, cout << FindLocMax(fr, 4, -1) << endl);
    // test treatment of edges
    for(i=0; i<10; ++i) {
	fr[i] = i;
    }
    // Search up to half-local-max at end should work
    TEST("3.13", FindLocMax(fr, 4, 1)==9, cout << FindLocMax(fr,4,1));
    // Search down should fail
    TEST("3.14", FindLocMax(fr, 4, -1)==-1, cout << FindLocMax(fr,4,-1));
    // Starting at the edge should be the same
    TEST("3.15", FindLocMax(fr, 9, 1)==9, cout << FindLocMax(fr,9,1));
    // .. but starting *on* a local-max-treated edge should find it
    TEST("3.16", FindLocMax(fr, 9, -1)==9, cout << FindLocMax(fr,9,-1));
    // .. and at the other edge
    TEST("3.15a", FindLocMax(fr, 0, 1)==9, cout << FindLocMax(fr,0,1));
    TEST("3.16a", FindLocMax(fr, 0, -1)==-1, cout << FindLocMax(fr,0,-1));
    // Whole thing with slope the other way
    for(i=0; i<10; ++i) {
	fr[i] = -i;
    }
    TEST("3.17", FindLocMax(fr, 4, 1)==-1, cout << FindLocMax(fr,4,1));
    TEST("3.18", FindLocMax(fr, 4, -1)==0, cout << FindLocMax(fr,4,-1));
    TEST("3.19", FindLocMax(fr, 9, 1)==-1, cout << FindLocMax(fr,9,1));
    TEST("3.20", FindLocMax(fr, 9, -1)==0, cout << FindLocMax(fr,9,-1));
    TEST("3.21", FindLocMax(fr, 0, 1)==0, cout << FindLocMax(fr,0,1));
    TEST("3.22", FindLocMax(fr, 0, -1)==0, cout << FindLocMax(fr,0,-1));

    // Test saddleOK function - returns at saddle points
    fr.Fill(0);
    fr[1] = 1;
    fr[2] = 2;
    fr[3] = 3;
    fr[4] = 3;
    fr[5] = 3;
    fr[6] = 4;
    fr[7] = 4;
    fr[8] = 3;
    int m;
    TEST("3.23", (m=FindLocMax(fr, 0, 1, 1, 0))==6, cout << m << endl);
    TEST("3.24", (m=FindLocMax(fr, 0, 1, 1, 1))==3, cout << m << endl);
    // Test interaction of saddles and edges
    fr.Fill(0);
    fr[8] = 1;
    fr[9] = 2;
    // -1 in 4th arg is searching for min, not max
    TEST("3.25", (m=FindLocMax(fr, 9, -1, -1, 1))==7, cout << m << endl); 
    TEST("3.26", (m=FindLocMax(fr, 9, 1, -1, 1))==10, cout << m << endl); 

    cout << "OK" << endl;
}

void testNearestPk(void)
{
    floatVec fv(50);
    int i;
    // unambigous local max
    double a, b, c;
    for(i = 0; i < 50; ++i) {
	a = -4*pow(i-17.7,2);
	b = -0.1*pow(i-33.0,2);
	c = -2-0.2*pow(i-40.5,2);
	fv[i] = 100+max(max(a,b),c);
    }
    int rc;
    rc = NearestPeak(fv,0,0,&a,&b);
    TEST("4.1", rc==1 && fEq(a,17.7) && fEq(b,100), cout << rc << " " << a << " " << b << endl);
    rc = NearestPeak(fv,30,0,&a,&b);
    TEST("4.2", rc==1 && fEq(a,33.0) && fEq(b,100), cout << rc << " " << a << " " << b << endl);
    rc = NearestPeak(fv,40.5,0,&a,&b);
    TEST("4.3", rc==1 && fEq(a,40.5) && fEq(b,98), cout << rc << " " << a << " " << b << endl);

    // Test the window limit 
    rc = NearestPeak(fv,45,5,&a,&b);
    TEST("4.4", rc==1 && fEq(a,40.5) && fEq(b,98), cout << rc << " " << a << " " << b << endl);
    rc = NearestPeak(fv,45,4,&a,&b);
    TEST("4.5", rc==0, cout << rc << " " << a << " " << b << endl);
    rc = NearestPeak(fv,45,4.5,&a,&b);
    TEST("4.6", rc==1 && fEq(a,40.5) && fEq(b,98), cout << rc << " " << a << " " << b << endl);
   
    // Test sloping
    for(i = 0; i < 50; ++i) {
	// Sharp peak at 10.1 and much shallower one at 30.9
	fv[i] = exp(-0.5*pow((i-10.1)/2, 2)) + 0.1*exp(-0.5*pow((i-30.9)/20, 2));
    }
    // Now slope for sharp peak should die out way before shallower one
    rc = NearestPeak(fv,18,0,&a,&b);	//  closer to sharp, but slope for shallow
    TEST("4.7", rc==1 && fEq(a,30.9) && fEq(b,0.1), cout << rc << " " << a << " " << b << endl);
    rc = NearestPeak(fv,12,0,&a,&b);	// close enough for sharp to capture
    TEST("4.8", rc==1 && fEq(a,10.107) && fEq(b,1.058), cout << rc << " " << a << " " << b << endl);
    
    // Test in flat
    fv.Fill(2);	// all flat - nothing found
    rc = NearestPeak(fv,5,0,&a,&b);
    TEST("4.9", rc==0, cout << rc << " " << a << " " << b << endl);
    // flat with one bump
    for(i = 0; i < 50; ++i) {
	fv[i] = max(0.0,10-0.1*pow(i-20.0,2));   // zero except 10..30
    }
    rc = NearestPeak(fv,12,0,&a,&b);  // start in flat region - search bth ways
    TEST("4.10", rc==1 && fEq(a,20) && fEq(b,10), cout << rc << " " << a << " " << b << endl);
    rc = NearestPeak(fv,5,0,&a,&b);  // start in flat region - search bth ways
    TEST("4.11", rc==1 && fEq(a,20) && fEq(b,10), cout << rc << " " << a << " " << b << endl);
    rc = NearestPeak(fv,40.3,0,&a,&b);
    TEST("4.12", rc==1 && fEq(a,20) && fEq(b,10), cout << rc << " " << a << " " << b << endl);

    // Add a second slope upwards right at the edge
    fv[49] = 4;
    fv[48] = 3;
    fv[47] = 2;
    fv[46] = 1;
    // Peaks at 20 and 49 - has to choose one or other
    rc = NearestPeak(fv,34,0,&a,&b);
    TEST("4.13", rc==1 && fEq(a,20) && fEq(b,10), cout << rc << " " << a << " " << b << endl);
    // testing high-side also tests pk at end
    rc = NearestPeak(fv,35,0,&a,&b);
    TEST("4.14", rc==1 && fEq(a,49) && fEq(b,4), cout << rc << " " << a << " " << b << endl);

    cout << "OK" << endl;
}

void testFindAllMaxes(void)
{
    floatVec fv(48);
    int ixs[20];
    int i;
    for(i=0; i<48; ++i) {
	fv[i] = cos(2*M_PI*i/8)*exp(-i/16.0);
    }
    int nm = FindAllMaxes(fv, ixs, 20, 0);
    TEST("5.1", nm==7 && ixs[0]==0 && ixs[1]==8 && ixs[2]==16, 
	 cout << nm << " " << ixs[0] << " " << ixs[1] << " " << ixs[2] << endl);
    // test that threshold works 
    nm = FindAllMaxes(fv, ixs, 20, 0.3);
    TEST("5.2", nm==3 && ixs[0]==0 && ixs[1]==8 && ixs[2]==16, 
	 cout << nm << " " << ixs[0] << " " << ixs[1] << " " << ixs[2] << endl);
    cout << "OK" << endl;
}

int main()
{
    cout << "Testing SubSmooth.." << endl;
    testSubSmooth();
    cout << "Testing QuadFit3pt.." << endl;
    testQFit();
    cout << "Testing findLocMac.." << endl;
    testLocMax();
    cout << "Testing NearestPeak.." << endl;
    testNearestPk();
    cout << "Testing FindAllMaxes.." << endl;
    testFindAllMaxes();
}

#endif /* MAIN */


