//
// moreFVops.H
//
// Implementation of various sophisticated operations on floatvecs
// modeled after my matlab functions
//
// dpwe@media.mit.edu 1996jan24
//

#ifndef _MOREFVOPS_H
#define _MOREFVOPS_H

#include "nSurface.H"   // "floatVec.H"

#ifndef M_PI
#define M_PI 3.1415926535
#endif /* !M_PI */

void QuadFit3pt(double pm1, double p0, double pp1, 
		double *px, double *py, double cfs[]);
    /* Fit a quadratic peak to the three uniformly-spaced points p0-p2.
     * Return the local max/min x and y, and the coefs in cfs.
     * On return, p0 = cfs[0]; pm1 = cfs[0]-cfs[1]+cfs[2];
     * pp1 = cfs[0]+cfs[1]+cfs[2].
     */

floatVec SubSmooth(const floatRef& fr, int subfact, int winlen);
    /* Smooth a floatref with a hanning window of <winlen> nonzero points 
     * subsample the result by <subfact>; return a new floatVec with 
     * the result. 
     */

int FindLocMax(const floatRef& fr, int base, int sense, 
	       int maxnotmin = 1, int saddleOK = 0);
    /* Walk along a vector from the specified position looking for 
     * a local maximum.  Return its index if found, or -1 if none 
     * found.  sense==1 means search upwards, -1 (or <=0) means down.
     * If maxnotmin is 1, find a max, else find a min.
     * Now acts to be as early as possible - flat regions are reported 
     * as the first point the flatness encountered, although not 
     * until the far side is seen.
     * If saddleOK is not zero, will return on a saddle point as well 
     * as a true max i.e. two successive points the same counts.
     * If search is upward and saddleOK is set, the 'I didn't find one'
     * return code is fr.Size() i.e. there is an implied saddle 'beyond 
     * the end' (hence -1 is the 'true' saddle location for down searches
     * already).
     */

int NearestPeak(const floatRef& fr, double ix, double ixwin, 
		double *pos, double *ht);
    /* Pick out a peak in the vector fr at or near the (fractional)
     * index ix.  To count, the found local maximum must be within 
     * ixwin of the specified position.  If a peak is found, it returns 
     * 1 and the quadratic-interpolated position of the peak is returned 
     * in *pos (and height in *ht).  Otherwise, return 0.
     * Specifying ixwin as zero disables limited range.
     */

double FracInterp(const floatRef& fr, double ix, int order=6);
    /* Interpolate the values in the vector to return the value 
     * that is interpolated at the fractional index <ix> (from 0).
     * Uses truncated sinc interp at present, there's always 
     * the future for 4pt lagrangians 
     */

floatVec FracInterpVec(const floatRef& src, const floatRef& ixs, int order=6);
    /* return a whole vector of values interpolated from <src> at the 
       indexes indicated in <ixs> */

int PeakWidth(const floatRef& src, int ix, int *pll=0, int *pul=0);
   /* How wide is the gap between local minima occupied by this 
       peak?  Basically the same logic as ExtractPeak, but rather 
       than copying data, it simply returns a width. 
       Will fill in the actual indices of the upper and lower limit 
       into *pll and *pul if they are non-null. */

floatVec ExtractPeak(const floatRef& src, int indx);
    /* find a local max near indx; track out to the local minima on 
       both sides, copy the entire intervening region to the output 
       vector (same size as src); rest of output is zero. */

floatVec LogShift(const floatRef& src, double shift, double binsperoct, 
		  int order=6);
    /* <src> is treated as a log-sampled signal, with <binsperoct>
       bins per octave.  The output is the same size, but with the 
       contents delayed by <shift> units, where one unit is equivalent 
       to the lowest sample point's linear value. 
       <order> is the order used for FracInterp. */
floatArray LogShift(const arrayRef& ar, double shift, double binsperoct, 
		    int order=6);
void LogShiftAdd(floatRef& accum, const floatRef& src, double shift, 
		 double binsperoct, double scale = 1.0, int order=6);
    /* Optimization for the way LogShift is used in Weft::Update - 
     * shift src by the logshift, scale it by scale, then *add* it 
     * into accum, rather than generating and copying a new floatVec.  
     * Efficient because often the shift and the nonzero region of src
     * mean that no additions need be performed
     */

int FindAllMaxes(const floatRef& fr, int ixs[], int ixsz, double thresh=0);
    /* Go through <fr> looking for local maxes.  Store the integer 
     * locations of the local max points in the ixs[] array, up to 
     * a maximum of <ixsz> elements.  Return the actual number of 
     * elements written into <ixs>.
     * Local maxima smaller than <thresh> won't be counted.
     */

int FindAllPeaks(const floatRef& fr, floatVec** pPosRtn, floatVec** pHtRtn, 
		 double thresh=0);
    /* Go through a vector selecting all the peaks above a threshold;
     * return them in two new floatVecs, giving the position and 
     * corresponging height of each quad-interpolated peak found.
     * Return the size of these two new arrays.
     */

void SortedIndices(const floatRef& fr, int ixs[], int ixssz, int sense=1);
    /* Without changing fr, fill ixs (which has the same number of elements) 
     * with the indexes of the elements if fr, which, presented in that order, 
     * would be an ascending list.  This array, passed to IndexFloatVec, 
     * will result in a correctly sorted floatVec.
     * <ixssz> is the size of the ixs array; only the first so many 
     * of the sorted elements will be returned. 
     * <sense> > 0 implies ascending sort; <=0 means largest first.
     */

floatVec IndexFloatVec(const floatRef& fr, int ixs[], int nix);
    /* Construct a new floatVec with <nix> elements, each of which's value 
     * is taken from the element of <fr> enumerated by the corresponding 
     * element of <ixs>. 
     */

void ReverseIntVec(int ixs[], int nix);
    /* Reverse the order of elements in a simple 1D vector of ints, 
     * <ixs>, of length nix. 
     */

floatVec WindowedExtract(const floatRef& fr, double winCtr, double winWidth);
    /* Extract a portion of a floatvec into a second floatvec of the 
     * same size by applying a raised-cosine-edged flat window whose 
     * total nozero width is winwidth
     */
floatArray WindowedExtract(const arrayRef& ar, double winCtr, double winWidth);

#include <stdio.h>  // for sprintf, for faking multiple results into a tcl str
#include <string.h>	// for strlen

class Glob {	// dummy class to provide otcl access to global methods
public:
    Glob(void) {};
    ~Glob(void) {};

    floatVec SubSmooth(floatRef &fr, int subfact, int winlen)
        { return ::SubSmooth(fr, subfact, winlen); }
    int FindLocMax(floatRef &fr, int base, int sense)
        { return ::FindLocMax(fr, base, sense); }
    char* NearestPeak(floatRef &fr, double ix, double ixwin)
        { char s[256]; double p,h;
	  if(::NearestPeak(fr, ix, ixwin, &p, &h)) {
	      sprintf(s, "%f %f", p, h);
	  } else {
	      s[0] = '\0';
	  } return s; }
    double FracInterp(floatRef& fr, double ix, int order=6)
        { return ::FracInterp(fr, ix, order); }
    floatVec FracInterpVec(floatRef& fr, floatRef& ixs, int order=6)
        { return ::FracInterpVec(fr, ixs, order); }
    floatVec ExtractPeak(floatRef& src, int ix)
        { return ::ExtractPeak(src, ix); }
    floatVec LogShift(floatRef& src, double shift, double binsperoct, 
		      int order =6)
        { return ::LogShift(src, shift, binsperoct, order); }
    char* FindAllMaxes(floatRef& fr, double thresh=0)
        { int i,ixs[1024];
	  int ni = ::FindAllMaxes(fr, ixs, 256, thresh);
	  return IntsToStr(ixs, ni); }
    floatVec* FindAllPeaks(floatRef& fr, double thresh=0)
        {  floatVec *poss, *hts; ::FindAllPeaks(fr, &poss, &hts, thresh);
           delete hts; return poss; }
    floatVec* FindAllPeakHts(floatRef& fr, double thresh=0)
        {  floatVec *poss, *hts; ::FindAllPeaks(fr, &poss, &hts, thresh);
           delete poss; return hts; }
    char *SortedIndices(floatRef& fr)
        { int ixs[1024]; ::SortedIndices(fr, ixs, 1024, 1);
 	  return IntsToStr(ixs, fr.Size()); }
    floatVec SortFloatVec(floatRef& src, floatRef &key, int sense=1)
        { int ixs[1024]; ::SortedIndices(key, ixs, 1024, 1);
	  if(sense <= 0) ::ReverseIntVec(ixs, src.Size());
	  return ::IndexFloatVec(src, ixs, src.Size()); }
    floatVec IndexFloatVec(floatRef& src, char *ixss)  
        { int *pixs; int ni = StrToInts(ixss, &pixs);
	  floatVec fv = ::IndexFloatVec(src, pixs, ni); delete [] pixs;
	  return fv; }
    floatVec WindowedExtract(floatRef& fr, double winCtr, double winWidth)
        { return ::WindowedExtract(fr, winCtr, winWidth); }

private:
    char* IntsToStr(int ixs[], int nix)
      { char s[2048], *c; int i; c = s; for(i=0; i<nix; ++i) {
	      if(i>0) *c++ = ' ';
	      sprintf(c, "%d", ixs[i]); c += strlen(c); }  return s; }
    int NToks(char *s) 
      { int tok=0, havesome=0; char c;
	while(c = *s++) { if(havesome && (c==' ')) { ++tok; havesome = 0; }
	else if(c != ' ') {havesome = 1; }} if(havesome) ++tok; return tok;}
    int StrToInts(char *s, int **pix)
      { int nt = NToks(s); *pix = new int[nt]; int tt = 0;
        while(*s) { while(*s==' ') ++s; (*pix)[tt++] = atoi(s); 
        while(*s!=' ' && *s!='\0') ++s; } return nt; }
};

#endif /* _MOREFVOPS_H */
