/*
 * pfif.C
 *
 * C language interface to QN pfile fns for Tcl/floatArray
 * 
 * 1997apr08 dpwe@icsi.berkeley.edu
 * $Header: /u/drspeech/src/guitools/pfif_otcl/RCS/pfif.C,v 1.5 1999/05/22 00:42:50 dpwe Exp $
 */

#include "pfif.H"
#include <QN_filters.h>

// for the mode field
#define MODE_CLOSED 0
#define MODE_READ 1
#define MODE_WRITE 2

pfileif::pfileif(int a_normutts /* = 0 */) : normutts(a_normutts) {
    // constructor 
    outpfile = NULL;
    inpfile = NULL;
    normipf = NULL;
    file = NULL;
    mode = MODE_CLOSED;

    // ask QN to report debug messages
    debug = 0;
    // ask QN to give random-access to pfiles
    indexed = 1;

    // default to 0 labels (it's ok)
    nlabels = 0;

    nfeatures = -1;
    nutts = -1;
    totalframes = -1;

    // tracking state for writes
    current_utt = -1;
    current_frame = 0;
}

pfileif::~pfileif(void) {
    close();
}

int pfileif::open(char *filename, char *fmode) {
    // open a pfile.  Mode is "r" for read or "w" for write
    // returns 1 on success or 0 on failure
    if (mode != MODE_CLOSED) {
	fprintf(stderr, "pfileif::open: object is already open\n");
	return 0;
    }
    if (strcmp(fmode, "r") == 0) {
	return openRead(filename);
    } else if (strcmp(fmode, "w") == 0) {
	return openWrite(filename);
    } else {
	fprintf(stderr, "pfileif::open: mode of '%s' is not 'r' or 'w'\n", 
		fmode);
	return 0;
    }
}

int pfileif::openRead(char *filename) {
    // open a named pfile for input
    // returns 1 on success or 0 for failure
    file = fopen(filename, "r");
    if(file == NULL) {
	fprintf(stderr, "pfileif::openRead: couldn't open '%s'\n", filename);
	return 0;
    }
    inpfile = new QN_InFtrLabStream_PFile(debug, filename, file, indexed);
    if(inpfile == NULL) {
	fprintf(stderr, "pfileif::openRead: QN open returned null for '%s'\n", 
		filename);
	fclose(file);
	file = NULL;
	return 0;
    }
    nfeatures = inpfile->num_ftrs();
    nlabels = inpfile->num_labs();
    nutts = inpfile->num_segs();
    totalframes = inpfile->num_frames();	// defaults to whole file

    if (normutts) {
	normipf = new QN_InFtrStream_NormUtts(debug, filename, *inpfile);
    }

    mode = MODE_READ;
    return 1;
}

int pfileif::openWrite(char *filename) {
    // open a named pfile for output
    // returns 1 on success or 0 for failure
    if(nfeatures < 0 || nlabels < 0) {
	fprintf(stderr, "pfileif::openWrite: not enough features and labels (need 0+1, have %d+%d)\n", nfeatures, nlabels);
	return 0;
    }
    file = fopen(filename, "w");
    if(file == NULL) {
	fprintf(stderr, "pfileif::openWrite: couldn't open '%s'\n", filename);
	return 0;
    }
    outpfile = new QN_OutFtrLabStream_PFile(debug, filename, file, nfeatures,
					    nlabels, indexed);
    if(outpfile == NULL) {
	fprintf(stderr, "pfileif::openWrite: QNopen returned null for '%s'\n", 
		filename);
	fclose(file);
	file = NULL;
	return 0;
    }

    mode = MODE_WRITE;
    current_utt = 0;
    current_frame = 0;
    return 1;
}

void pfileif::close(void) {
    if (normipf != NULL) {
	delete normipf;
	normipf = NULL;
    }
    if (mode == MODE_READ) {
	delete inpfile;
	inpfile = NULL;
    } else if (mode == MODE_WRITE) {
	delete outpfile;
	outpfile = NULL;
    }
    if (file != NULL) {
	fclose(file);
	file = NULL;
    }
    mode = MODE_CLOSED;
}

int pfileif::setNumFeatures(int nfeats) {
    // set the number of features prior to opening an output file
    // returns 1 on success, 0 on failure
    if (mode != MODE_CLOSED) {
	fprintf(stderr, "pfileif::setNumFeatures: cannot set an open file!\n");
	return 0;
    } else {
	nfeatures = nfeats;
	return 1;
    }
}

int pfileif::numFramesForUtt(int utt) {
    // read the index for the number of frames in a given utt
    // returns -1 on error
    if (mode != MODE_READ) {
	fprintf(stderr, "pfileif::numFramesForUtt: only for input files!\n");
	return -1;
    } else {
	int frames = inpfile->num_frames(utt);
	if(frames == QN_SIZET_BAD) {
	    return -1;
	} else {
	    return frames;
	}
    }
}

int pfileif::seek(int utt, int frame) {
    // position next read to be the named frame of the named utt
    // return 1 on success, 0 on error
    if (mode != MODE_READ) {
	fprintf(stderr, "pfileif::seek: only for input files!\n");
	return 0;
    } else {
	int seg;
	if (normutts) {
	    seg = normipf->set_pos(utt, frame);
	} else {
	    seg = inpfile->set_pos(utt, frame);
	}
	if(seg == QN_SEGID_BAD) {
	    return 0;
	} else {
	    return 1;
	}
    }
}

int pfileif::tell(int* putt, int* pframe) {
    // return the address of the next frame to be returned
    // return 1 on success, 0 on error
    if (mode == MODE_READ) {
	int ok;
	size_t utt, frame;
	if (normutts) {
	    ok = normipf->get_pos(&utt, &frame);
	} else {
	    ok = inpfile->get_pos(&utt, &frame);
	}
	if(ok == QN_OK) {
	    *putt = utt;
	    *pframe = frame;
	    return 1;
	} else {
	    return 0;
	}
    } else if (mode == MODE_WRITE) {
	if (current_utt < 0) {
	    return 0;
	} else {
	    *putt = current_utt;
	    *pframe = current_frame;
	    return 1;
	}
    } else {
	return 0;
    }
}

int pfileif::readFramesSub(int nframes, float *buf) {
    // private bottom-level read function
    if (mode != MODE_READ) {
	fprintf(stderr, "pfileif::readFramesSub: only for input files!\n");
	return 0;
    }
    int nred;
    if (normutts) {
	nred = normipf->read_ftrs(nframes, buf);
    } else {
	nred = inpfile->read_ftrs(nframes, buf);
    }
    if(nred < nframes) {
	fprintf(stderr, "pfileif::readFramesSub: wanted %dx%d, got %d\n", 
		nframes, nfeatures, nred);
	return 0;
    } else {
	return nred;
    }
}

floatVec pfileif::readFrame(void) {
    // read one frame into a floatVec
    floatVec rtn(nfeatures);
    if(readFramesSub(1, &(rtn.El(0))) > 0) {
	return rtn;
    } else {
	return 0;
    }
}

floatArray pfileif::readFrames(int nframes) {
    // read the next several frames of features
    floatArray rtn(nframes, nfeatures);
    if(readFramesSub(nframes, &(rtn.El(0,0))) > 0) {
	return rtn;
    } else {
	return 0;
    }
}

int pfileif::readLabels(int nframes, int *labels) {
    // read the next several frames of labels; fill in pre-allocated array
    if (mode != MODE_READ) {
	fprintf(stderr, "pfileif::readLabels: only for input files!\n");
	return 0;
    }
    int nred = inpfile->read_labs(nframes, (QNUInt32*)labels);
    return nred;
}

int pfileif::writeFramesSub(int nframes, const float *buf) {
    // private bottom-level write function
    if (mode != MODE_WRITE) {
	fprintf(stderr, "pfileif::writeFramesSub: only for output files!\n");
	return 0;
    }
    int nwrit;
    // write_ftrs returns void??
    outpfile->write_ftrs(nframes, buf);
    nwrit = nframes;
//    nwrit = outpfile->write_ftrs(nframes, buf);

    if(nwrit < nframes) {
	fprintf(stderr, "pfileif::writeFramesSub: tried %dx%d, managed %d\n", 
		nframes, nfeatures, nwrit);
	return 0;
    } else {
	current_frame += nwrit;
	return nwrit;
    }
}

int pfileif::writeFrame(floatRef& fr) {
    // write one frame from a floatRef
    // return 1 on success, 0 on failure
    if(fr.Size() != nfeatures) {
	fprintf(stderr, "pfileif::writeFrame: floatRef.Size=%d but nfeatures=%d\n", fr.Size(), nfeatures);
	return 0;
    }
    if(writeFramesSub(1, &(fr.El(0))) > 0) {
	return 1;
    } else {
	return 0;
    }
}

int pfileif::writeFrames(arrayRef& ar) {
    // write several frames from a arrayRef
    // returns the actual number of frames read (0 on failure)
    if(ar.Cols() != nfeatures) {
	fprintf(stderr, "pfileif::writeFrames: arrayRef.Cols=%d but nfeatures=%d\n", ar.Cols(), nfeatures);
	return 0;
    }
    int rows;
    if((rows = writeFramesSub(ar.Rows(), &(ar.El(0,0)))) > 0) {
	return rows;
    } else {
	return 0;
    }
}

int pfileif::writeLabels(int nframes, int *labels) {
    // write out some frames as just labels
    if (mode != MODE_WRITE) {
	fprintf(stderr, "pfileif::writeLabels: only for output files!\n");
	return 0;
    }
/*    int nwrit = outpfile->write_labs(nframes, (QNUInt32*)labels); */
    int nwrit = nframes;
    outpfile->write_labs(nframes, (QNUInt32*)labels);
    return nwrit;
}

int pfileif::writeUttEnd(void) {
    // indicate that the current utterence has ended
    // return 1 on success, 0 on failure
    if (mode != MODE_WRITE) {
	fprintf(stderr, "pfileif::writeUtttEnd: only for output files!\n");
	return 0;
    }
    outpfile->doneseg(current_utt);
    current_utt++;
    // reset the frame counter
    current_frame = 0;
    return 1;
}

int pfileif::isOpen(void) {
    // return 1 if the file is open
    if (mode == MODE_CLOSED) {
	return 0;
    } else {
	return 1;
    }
}
