/*
   $Header: /u/spert/src/intvec/testsuite/RCS/vecele_vi32vi32_vi32_test.c,v 1.2 1996/01/03 03:46:32 davidj Exp $

   Test element wise vector routines of the form op_vi32vi32_vi32.
*/

#include <stddef.h>
#include <stdlib.h>

#include "intvec_test.h"


typedef void (*rtn_t)(size_t len, const intv_int32_t* a, const intv_int32_t* b,
		      intv_int32_t* res);

/* Performs one test given input values and expected output. */

void
onetest_normal(rtn_t rtn, size_t len,
	       const intv_int32_t* a, const intv_int32_t* b,
	       const intv_int32_t* correct)
{
    /* Create padded copies of inputs. */
    intv_int32_t* c_a = rtst_padvec_clone_vi32(len, a);
    intv_int32_t* c_b = rtst_padvec_clone_vi32(len, b);
    intv_int32_t* c_res = rtst_padvec_new_vi32(len);

    /* Run routine. */
    rtn(len, c_a, c_b, c_res);

    /* Check result. */
    rtst_checkeq_vi32vi32(len, c_res, correct);

    /* Check input not corrupted. */
    rtst_checkeq_vi32vi32(len, a, c_a);
    rtst_checkeq_vi32vi32(len, b, c_b);

    /* Free and check padding. */
    rtst_padvec_del_vi32(c_a);
    rtst_padvec_del_vi32(c_b);
    rtst_padvec_del_vi32(c_res);
}

void
add1(rtn_t rtn)
{
    const intv_int32_t a[] = { 0, -1, 1, -1, 1, 0x7fffffff, -0x7fffffff-1,
				   123 };
    const intv_int32_t b[] = { 0, -1, -1, 1, 1, 1, -1, -456 };
    const intv_int32_t correct[] = { 0, -2, 0, 0, 2, 0x80000000, 0x7fffffff,
					 -333 };
    
    onetest_normal(rtn, sizeof(a)/sizeof(intv_int32_t), a, b, correct);
}

void
sub1(rtn_t rtn)
{
    const intv_int32_t a[] = { 0, -1, 1, -1, 1, -0x7fffffff-1, 0x7fffffff,
				   123 };
    const intv_int32_t b[] = { 0, -1, -1, 1, 1, 1, -1,  -456 };
    const intv_int32_t correct[] = { 0, 0, 2, -2, 0, 0x7fffffff, 0x80000000,
					 579 };
    
    onetest_normal(rtn, sizeof(a)/sizeof(intv_int32_t), a, b, correct);
}

void
mul1(rtn_t rtn)
{
    const intv_int32_t a[] =
    { 0, 0, 1, -1, 0x7fffffff, 0x80000000, 123, 0x7fffffff};
    const intv_int32_t b[] =
    { 0, 1, 1, -1, 1,          -1,         -456, 0x7fffffff};
    const intv_int32_t correct[] =
    { 0, 0, 1, 1, 0x7fffffff,  0x80000000, -56088, 0x00000001}; 
    
    onetest_normal(rtn, sizeof(a)/sizeof(intv_int32_t), a, b, correct);
}

void
satadd1(rtn_t rtn)
{
    const intv_int32_t a[] = { 0, -1, 1, -1, 1, 0x7fffffff, -0x7fffffff-1,
				   123 };
    const intv_int32_t b[] = { 0, -1, -1, 1, 1, 1, -1, -456 };
    const intv_int32_t correct[] = { 0, -2, 0, 0, 2, 0x7fffffff, 0x80000000,
					 -333 };
    
    onetest_normal(rtn, sizeof(a)/sizeof(intv_int32_t), a, b, correct);
}

void
satsub1(rtn_t rtn)
{
    const intv_int32_t a[] = { 0, -1, 1, -1, 1, INTV_INT32_MIN,
			       INTV_INT32_MAX, 123 };
    const intv_int32_t b[] = { 0, -1, -1, 1, 1, 1, -1,  -456 };
    const intv_int32_t correct[] = { 0, 0, 2, -2, 0, INTV_INT32_MIN,
				     INTV_INT32_MAX, 579 };
    
    onetest_normal(rtn, sizeof(a)/sizeof(intv_int32_t), a, b, correct);
}

/* Compare the results of two vector routines */

void
random_test(rtn_t nv_rtn, rtn_t test_rtn)
{
    const int numtests = rtst_numtests;
    const size_t maxlen = rtst_sizetests;
    intv_int32_t *a = rtst_padvec_new_vi32(maxlen);
    intv_int32_t *b = rtst_padvec_new_vi32(maxlen);
    intv_int32_t *res = rtst_padvec_new_vi32(maxlen);

    int i;
    for (i=0; i<numtests; i++)
    {
        size_t len = (size_t) rtst_urand_i32i32_i32(1, maxlen);

        rtst_badrand_vi32(len, a);
        rtst_badrand_vi32(len, b);
        nv_rtn(len, a, b, res);

        rtst_log("len=%d\n", len);

        onetest_normal(test_rtn, len, a, b, res);
    }

    rtst_padvec_del_vi32(a);
    rtst_padvec_del_vi32(b);
    rtst_padvec_del_vi32(res);
}

void
test_add(void)
{
    rtst_start("add_vi32vi32_vi32");
    rtst_log("testing nv.\n");
    add1(nv_add_vi32vi32_vi32);
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    add1(tor_add_vi32vi32_vi32);
    random_test(nv_add_vi32vi32_vi32, tor_add_vi32vi32_vi32);
#endif
    rtst_passed();
}

void
test_sub(void)
{
    rtst_start("sub_vi32vi32_vi32");
    rtst_log("testing nv.\n");
    sub1(nv_sub_vi32vi32_vi32);
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    sub1(tor_sub_vi32vi32_vi32);
    random_test(nv_sub_vi32vi32_vi32, tor_sub_vi32vi32_vi32);
#endif
    rtst_passed();
}

void
test_mul(void)
{
    rtst_start("mul_vi32vi32_vi32");
    rtst_log("testing nv.\n");
    mul1(nv_mul_vi32vi32_vi32);
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    mul1(tor_mul_vi32vi32_vi32);
    random_test(nv_mul_vi32vi32_vi32, tor_mul_vi32vi32_vi32);
#endif
    rtst_passed();
}

void
test_satadd(void)
{
    rtst_start("satadd_vi32vi32_vi32");
    rtst_log("testing nv.\n");
    satadd1(nv_satadd_vi32vi32_vi32);
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    satadd1(tor_satadd_vi32vi32_vi32);
    random_test(nv_satadd_vi32vi32_vi32, tor_satadd_vi32vi32_vi32);
#endif
    rtst_passed();
}

void
test_satsub(void)
{
    rtst_start("satsub_vi32vi32_vi32");
    rtst_log("testing nv.\n");
    satsub1(nv_satsub_vi32vi32_vi32);
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    satsub1(tor_satsub_vi32vi32_vi32);
    random_test(nv_satsub_vi32vi32_vi32, tor_satsub_vi32vi32_vi32);
#endif
    rtst_passed();
}

int
main(int argc, char *argv[])
{
    rtst_args(argc, argv);

    test_add();
    test_sub();
    test_mul();

    test_satadd();
    test_satsub();

    rtst_exit();

    return 0;
}
